#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../unix/unix.h"
#include "freebsd.h"

char osversion[] = "x86-linux-2012";

enum
{
	Maxinsn = 32,
	int3 = 0xcc,		/* breakpoint */
};

void
dead(Ctl *ctl)
{
	kqnotrace(ctl->id);
	_dead(ctl);
}

void
sctlinit()
{
	syminit();
	_sctlinit();
}

void
sctlfini()
{
	symfini();
	_sctlfini();
}

As*
mkas(Ctl *ctl)
{
	As *as;
	as = _mkas(ctl);
	return as;
}

void
freeas(As *as)
{
	symfreeas(as);
	_freeas(as);
}


void
xgetregs(Ctl *ctl, Reg *regs)
{
	if(0 > ptrace(PT_GETREGS, ctl->id, (caddr_t)regs, 0))
		xfatalno("PT_GETREGS");
}

void
xsetregs(Ctl *ctl, Reg *regs)
{
	if(0 > ptrace(PT_SETREGS, ctl->id, (caddr_t)regs, 0))
		xfatalno("PT_SETREGS");
}

static Event*
doevent(Ctl *ctl, int st)
{
	Brkref *br;
	char *err;
	Event *e;
	struct ptrace_lwpinfo lwpinfo;

	e = 0;
	if(WIFEXITED(st) || WIFSIGNALED(st)){
		/* FIXME: return status to client */
		if(ctl->tracemode&Eexit)
			e = eventexit(ctl);
		dead(ctl);
	}
	else if(!WIFSTOPPED(st))
		bug();
	else if(WSTOPSIG(st) == SIGTRAP){
		atbrkx86(ctl);
		br = atbrk(ctl);
		if(br){
			switch(br->bp->mode){
			case Msnap:
			default:
				bug();
				break;
			case Mbrk:
				e = eventtrap(ctl, br);
				break;
			case Mld:
				if(0 > ldevent(ctl, &err))
					xfatal("ldevent: %s", err);
				e = eventload(ctl, "look at stat", -1);
				if(0 > steptrap(ctl, &err))
					xfatalno("steptrap: %s", err);
				break;
			case Mentry:
				if(0 > steptrap(ctl, &err))
					xfatalno("steptrap: %s", err);
				ctlcont(ctl);
				break;
			}
		}else{
			if(ctl->state == Strap && 0 > steptrap(ctl, &err))
				xfatalno("cannot step over trap: %s", err);
			ctlcont(ctl);
		}
	}
	else{
		ctl->state = Sstop;
		if(ctl->ctx.cursig)
			xfatal("prctl: unimplemented: nested signals");
		xgetregs(ctl, &ctl->ctx.regs);
		ctl->ctx.cursig = WSTOPSIG(st);
		if(ctl->tracemode&Esignal){
			if(0 > ptrace(PT_LWPINFO, ctl->id, (caddr_t)&lwpinfo, 0))
				xfatalno("PT_LWPINFO");
			e = eventsignal(ctl, &lwpinfo.pl_siginfo);
		}else
			ctlcont(ctl);
	}

	return e;
}

void
ctlcont(Ctl *ctl)
{
	if(0 > ptrace(PT_CONTINUE, ctl->id, (caddr_t)1, ctl->ctx.cursig))
		xfatalno("PT_CONTINUE");
	ctl->ctx.cursig = 0;
	ctl->state = Srun;
}

static Event*
evqreverse(Event *p)
{
	Event *t, *q;

	q = 0;
	while(p){
		t = p->link;
		p->link = q;
		q = p;
		p = t;
	}
	return q;
}

Event*
ctlstop(Ctl *ctl, char **err)
{
	int st;
	Event *e, *p;

	p = 0;
	if(err)
		*err = 0;
	if(ctl->state != Srun)
		return 0;
	while(ctl->state == Srun){
		kill(ctl->id, SIGSTOP);
		if(0 > waitpid(ctl->id, &st, 0))
			xfatalno("waitpid");
		if(WIFSTOPPED(st) && WSTOPSIG(st) == SIGSTOP){
			xgetregs(ctl, &ctl->ctx.regs);
			ctl->state = Sstop;
			return evqreverse(p);
		}
		e = doevent(ctl, st);
		if(e){
			e->link = p;
			p = e;
		}
	}
	return evqreverse(p);
}

int
ctlstep(Ctl *ctl, char **err)
{
	int st;
	Event *e, *p;

	p = 0;
	if(err)
		*err = 0;
	if(0 > ptrace(PT_STEP, ctl->id, (caddr_t)1, 0))
		goto fail;
	while(1){
		if(0 > waitpid(ctl->id, &st, 0))
			goto fail;
		if(WIFSTOPPED(st) || WSTOPSIG(st) == SIGTRAP){
			if(ctl->tracemode&Estepctx)
				xgetregs(ctl, &ctl->ctx.regs);
			ctl->state = Sstop;
			if(p)
				xfatal("ctlstep: event propagation unimplemented");
			return 0;
		}
		e = doevent(ctl, st);
		if(e){
			e->link = p;
			p = e;
		}
	}
fail:
	if(err)
		*err = strerror(errno);
	return -1;
}

int
ctldetach(Ctl *ctl, char **err)
{
	u64 id;
	id = ctl->id;
	switch(ctl->state){
	case Srun:
		ctlstop(ctl, err);
		/* ignore errors, ignore events */
		/* fall through */
	case Strap:
	case Sstop:
		dead(ctl);
		if(0 > ptrace(PT_DETACH, id, (caddr_t)1, 0))
			xfatalno("PT_DETACH");
		break;
	case Sdead:
		return -1;
	default:
		bug();
	}
	return 0;
}

int
ctlkill(Ctl *ctl, char **err)
{
	if(ctl->state == Sdead)
		return -1;
	if(0 > ptrace(PT_KILL, ctl->id, 0, 0))
		xfatalno("PT_KILL");
	if(0 > waitpid(ctl->id, 0, 0))
		xfatalno("waitpid");
	dead(ctl);
	return 0;
}

static int
editbrk(Ctl *ctl, Brk *bp, int set, char **err)
{
	Event *e;
	char *lerr;
	int wasrunning;
	char txt;

	if(!err)
		err = &lerr;
	*err = 0;
	wasrunning = 0;
	if(ctl->state == Srun){
		wasrunning = 1;
		e = ctlstop(ctl, err);
		if(*err)
			return 0;
		if(e)
			xfatal("editbrk: event propagation unimplemented");
	}

	if(set){
		if(0 > ctlread(ctl, bp->addr, &txt, 1, err))
			goto fail;
		bp->text = txt;
		txt = int3;
	}else
		txt = bp->text;
	if(0 > ctlwrite(ctl, bp->addr, &txt, 1, err))
		goto fail;
	if(wasrunning)
		ctlcont(ctl);
	return 0;
fail:
	if(wasrunning)
		ctlcont(ctl);
	*err = strerror(errno);
	return -1;
}

int
clrbrk(Ctl *ctl, Brk *bp, char **err)
{
	/* FIXME: some thread may be stopped in this BP */
	return editbrk(ctl, bp, 0, err);
}

int
setbrk(Ctl *ctl, Brk *bp, char **err)
{
	return editbrk(ctl, bp, 1, err);
}

int
ctltrace(Ctl *ctl, Trace *t, char **err)
{
	int rv;
	rv = _ctltrace(ctl, t, err);
	return rv;
}

static int
ctlio(Ctl *ctl, int op, unsigned long addr, char *buf,
      unsigned long cnt, char **err)
{
	struct ptrace_io_desc io;

	if(err)
		*err = 0;
	io.piod_op = op;
	io.piod_offs = (void*)addr;
	io.piod_addr = buf;
	io.piod_len = cnt;
	if(0 > ptrace(PT_IO, ctl->id, (caddr_t)&io, 0)){
		if(err)
			*err = strerror(errno);
		return -1;
	}
	return 0;
}

int
ctlread(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	return ctlio(ctl, PIOD_READ_D, addr, buf, cnt, err);
}

int
ctlwrite(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	return ctlio(ctl, PIOD_WRITE_D, addr, buf, cnt, err);
}

int
steptrap(Ctl *ctl, char **err)
{
	/* FIXME: handle multi-thread case (scratch) */
	if(0 > clrbrk(ctl, &ctl->curbrk, err) ||
	   0 > ctlstep(ctl, err) ||
	   0 > setbrk(ctl, &ctl->curbrk, err))
		return -1;
	return 0;
}

Event*
ctlwait(Ctl *maybectl)
{
	Ctl *ctl;
	int pid, st;

	if(maybectl){
		pid = waitpid(maybectl->id, &st, 0);
		if(pid == -1)
			return 0; /* no more procs */
		if(pid != maybectl->id)
			bug();
		return doevent(maybectl, st);
	}

	/* wait for any ctl */
	pid = waitpid(-1, &st, 0);
	if(pid == -1)
		return 0; /* no more procs */
	if(pid == 0)
		bug();
	ctl = lookctl(pid);
	if(ctl == 0)
		bug();
	return doevent(ctl, st);
}

Event*
ctlpoll(Ctl *maybectl)
{
	Ctl *ctl;
	int st, pid;

	if(maybectl){
		pid = waitpid(maybectl->id, &st, WNOHANG);
		if(pid == -1)
			return 0; /* no more procs */
		if(pid == 0)
			return 0; /* WNOHANG -> nothing waiting */
		if(pid != maybectl->id)
			bug();
		return doevent(maybectl, st);
	}
	pid = waitpid(-1, &st, WNOHANG);
	if(pid == -1)
		return 0; /* no more procs */
	if(pid == 0)
		return 0; /* WNOHANG -> nothing waiting */
	ctl = lookctl(pid);
	if(ctl == 0)
		bug();
	return doevent(ctl, st);
}
