#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../unix/unix.h"

enum
{
	Maxpath = 128,
};

static int
parseregion(char *line, Region *r, char **err)
{
	char *p;
	char perm[4];
	char path[Maxpath];
	unsigned long endaddr;
	int rv;

	rv = sscanf(line,
		    "%lx %lx %*d %*d %*x %4s %*d %*d %*x %*s %*s %*s %128s",
		    &r->addr, &endaddr, perm, path);
	if(rv != 4){
		*err = Estrange;
		return -1;
	}
	r->len = endaddr-r->addr;
	r->flags = 0;
	p = perm;
	while(*p)
		switch(*p++){
		case 'r':
			r->flags |= PROT_READ;
			break;
		case 'w':
			r->flags |= PROT_WRITE;
			break;
		case 'x':
			r->flags |= PROT_EXEC;
			break;
		case '-':
			/* Do nothing */
			break;
		default:
			if(err)
				*err = Estrange;
			return -1;
		}
	r->file = xstrdup(path);
	return 0;
}

/* override the readfile in unix/util.c with one that
   recovers from EFBIG */
static char*
bsdreadfile(char *name, u64 *plen)
{
	char *rv;
	u64 len, max, munch;
	int fd;
	int m;

	len = 0;
	max = 1024;
again:
	rv = xmalloc(max+1);

	fd = open(name, O_RDONLY);
	if(0 > fd)
		goto fail;

	while(1){
		if(len >= max){
			rv = xrealloc(rv, max+1, max*2);
			max *= 2;
		}
		munch = max-len;
		m = read(fd, rv+len, munch);
		if(0 > m && (errno == EINTR || errno == EAGAIN))
			continue;
		if(0 > m && errno == EFBIG){
			close(fd);
			free(rv);
			max *= 2;
			goto again;
		}
		if(0 > m)
			goto fail;
		len += m;
		if(m < munch)
			break;
	}
	close(fd);
	rv[len] = 0;
	if(plen)
		*plen = len;
	return rv;
fail:
	close(fd);
	free(rv);
	return 0;
}

int
parsemaps(Stat *s, char **err)
{
	u64 n;
	char path[Maxpath];
	char *p, *m, *e;

	snprintf(path, sizeof(path), "/proc/%d/map", s->pid);
	m = bsdreadfile(path, 0);
	if(m == 0){
		/* system probably doesn't mount /proc.
		   user will figure it out. */
		s->nr = 0;
		return 0;
	}
	n = 0;
	p = m;
	while(*p){
		if(*p == '\n'){
			n++;
			*p = 0;
		}
		p++;
	}
	e = p;

	s->nr = 0;
	s->r = xmalloc(n*sizeof(Region));
	p = m;
	while(p < e){
		if(0 > parseregion(p, &s->r[s->nr++], err))
			goto fail;
		p += strlen(p)+1;
	}
	free(m);
	return 0;
fail:
	free(m);
	return -1;
}
