#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../inc/elf.h"
#include "../unix/unix.h"


// amd64 calling convention: %rdi, %rsi, %rdx, %rcx, %r8 and %r9

typedef struct {	/* Auxiliary vector entry on initial stack */
	int	a_type;			/* Entry type. */
	union {
		int	a_val;		/* Integer value. */
	} a_un;
} Elf32_Auxinfo;


typedef struct {	/* Auxiliary vector entry on initial stack */
	long	a_type;			/* Entry type. */
	union {
		long	a_val;		/* Integer value. */
		void	*a_ptr;		/* Address. */
		void	(*a_fcn)(void);	/* Function pointer (not used). */
	} a_un;
} Elf64_Auxinfo;

/* Values for a_type. */
#define	AT_NULL		0	/* Terminates the vector. */
#define	AT_IGNORE	1	/* Ignored entry. */
#define	AT_EXECFD	2	/* File descriptor of program to load. */
#define	AT_PHDR		3	/* Program header of program already loaded. */
#define	AT_PHENT	4	/* Size of each program header entry. */
#define	AT_PHNUM	5	/* Number of program header entries. */
#define	AT_PAGESZ	6	/* Page size in bytes. */
#define	AT_BASE		7	/* Interpreter's base address. */
#define	AT_FLAGS	8	/* Flags (unused for i386). */
#define	AT_ENTRY	9	/* Where interpreter should transfer control. */
#define	AT_NOTELF	10	/* Program is not ELF ?? */
#define	AT_UID		11	/* Real uid. */
#define	AT_EUID		12	/* Effective uid. */
#define	AT_GID		13	/* Real gid. */
#define	AT_EGID		14	/* Effective gid. */
#define	AT_EXECPATH	15	/* Path to the executable. */

#define	AT_COUNT	16	/* Count of defined aux entry types. */

struct link_map32
{
	u32	l_addr;
	u32	l_name;
	u32	l_ld;
	u32	l_next, l_prev;
};

struct link_map64
{
	u64	l_addr;
	u64	l_name;
	u64	l_ld;
	u64	l_next, l_prev;
};

static int
ldevent32(Ctl *ctl, char **err)
{
	u32 lmp, rd, sp;
	char *path;
	struct link_map32 lm;
	Bin *b;

	sp = ctl->ctx.regs.SP;


	if(0 > ctlread(ctl, sp+4, &rd, sizeof(rd), err))    /* r_debug */
		return -1;
	if(0 > ctlread(ctl, sp+8, &lmp, sizeof(lmp), err))    /* *link_map */
		return -1;

	while(lmp){
		if(0 > ctlread(ctl, lmp, &lm, sizeof(lm), err))
			return -1;
		path = ctlreads(ctl, lm.l_name, err);
		if(path == 0)
			return -1;
		// rtld re-writes its own name in 32-bit compatibility mode
		// assume any "/libexec/ld-elf" is the already seen loader
		if(strncmp(path, "/libexec/ld-elf", 15) != 0
		   && !lookbinpath(ctl, path)){
			b = mkbin(path, lm.l_addr, err);
			if(!b)
				return -1;
			addbin(ctl, b);
		}
		free(path);
		lmp = lm.l_next;
	}
	return 0;
}

static int
ldevent64(Ctl *ctl, char **err)
{
	u64 lmp, rd;
	char *path;
	struct link_map64 lm;
	Bin *b;

	rd = ctl->ctx.regs.DI;  /* r_debug */
	lmp = ctl->ctx.regs.SI; /* *link_map */
	while(lmp){
		if(0 > ctlread(ctl, lmp, &lm, sizeof(lm), err))
			return -1;
		path = ctlreads(ctl, lm.l_name, err);
		if(path == 0)
			return -1;
		if(!lookbinpath(ctl, path)){
			b = mkbin(path, lm.l_addr, err);
			if(!b)
				return -1;
			addbin(ctl, b);
		}
		free(path);
		lmp = lm.l_next;
	}
	return 0;
}

int
ldevent(Ctl *ctl, char **err)
{
	if(ctl->as->x.ld == 0)
		return 0; /* static binary */
	switch(ctl->as->x.mode){
	case M32:
		return ldevent32(ctl, err);
	case M64:
		return ldevent64(ctl, err);
	default:
		bug();
		return -1;
	}
}

static int
initaux32(Ctl *ctl, char **err)
{
	int i, argc;
	u32 x;
	u32 p;
	Elf32_Auxinfo *aux, a;
	unsigned sz;

	p = ctl->ctx.regs.SP;
	sz = sizeof(u32);

	if(0 > ctlread(ctl, p, &x, sz, err))
		return -1;
	argc = x;

	for(i = 0; i < argc+1; i++)
		if(0 > ctlread(ctl, p+(1+i)*sz, &x, sizeof(x), err))
			return -1;

	p += (1+argc+1)*sz;  /* skip argc and argv */

	/* skip environment */
	i = 0;
	while(1){
		if(0 > ctlread(ctl, p, &x, sz, err))
			return -1;
		p += sz;
		i++;
		if(x == 0)
			break;
	}

	/* read aux vector */
	i = 0;
	aux = xmalloc(AT_COUNT*sizeof(*aux));
	while(1){
		if(0 > ctlread(ctl, p, &a, sizeof(a), err))
			return -1;
		if(a.a_type == AT_NULL)
			break;
		if(a.a_type >= AT_COUNT){
			xlog("out of bounds auxiliary vector index: %d",
			     a.a_type);
			return -1;
		}
		aux[a.a_type] = a;
		p += sizeof(a);
	}
	ctl->as->x.aux = aux;
	return 0;
}

/* contrary to both the x86_64 ABI and the informal
   comments in the freebsd rtld, the kernel passes
   the pointer to argc in %rdi.  the actual stack
   pointer is 16-byte aligned down from %rdi. */
static int
initaux64(Ctl *ctl, char **err)
{
	int i, argc;
	u64 x;
	u64 p;
	Elf64_Auxinfo *aux, a;
	unsigned sz;

	p = ctl->ctx.regs.DI;
	sz = sizeof(u64);

	if(0 > ctlread(ctl, p, &x, sz, err))
		return -1;
	argc = x;

	for(i = 0; i < argc+1; i++)
		if(0 > ctlread(ctl, p+(1+i)*sz, &x, sizeof(x), err))
			return -1;

	p += (1+argc+1)*sz;  /* skip argc and argv */

	/* skip environment */
	i = 0;
	while(1){
		if(0 > ctlread(ctl, p, &x, sz, err))
			return -1;
		p += sz;
		i++;
		if(x == 0)
			break;
	}

	/* read aux vector */
	aux = xmalloc(AT_COUNT*sizeof(*aux));
	while(1){
		if(0 > ctlread(ctl, p, &a, sizeof(a), err))
			return -1;
		if(a.a_type == AT_NULL)
			break;
		if(a.a_type >= AT_COUNT){
			xlog("out of bounds auxiliary vector index: %d",
			     a.a_type);
			return -1;
		}
		aux[a.a_type] = a;
		p += sizeof(a);
	}
	ctl->as->x.aux = aux;
	return 0;
}

static int
initaux(Ctl *ctl, char **err)
{
	switch(ctl->as->x.mode){
	case M32:
		return initaux32(ctl, err);
		break;
	case M64:
		return initaux64(ctl, err);
		break;
	default:
		bug();
		return -1;
	}
}

static u64
lookaux(Ctl *ctl, unsigned idx)
{
	Elf32_Auxinfo *a32;
	Elf64_Auxinfo *a64;

	if(idx > AT_COUNT)
		bug();
	switch(ctl->as->x.mode){
	case M32:
		a32 = (Elf32_Auxinfo*)ctl->as->x.aux;
		// don't sign extend signed a_val
		return (u64)(unsigned)a32[idx].a_un.a_val;
		break;
	case M64:
		a64 = (Elf64_Auxinfo*)ctl->as->x.aux;
		return (u64)(uintptr_t)a64[idx].a_un.a_ptr;
		break;
	default:
		bug();
		return -1;
	}
}

static char*
ctlexe(Ctl *ctl, char **err)
{
	char buf[32];
	char exe[256];
	ssize_t rv;

	snprintf(buf, sizeof(buf), "/proc/%" PRIu64 "/file", ctl->id);
	rv = readlink(buf, exe, sizeof(exe)-1);
	if(0 > rv){
		*err = Ebin;
		return 0;
	}
	exe[rv] = 0;
	return xstrdup(exe);
}

static u64
mapaddr(Ctl *ctl, char *exepath, char **err)
{
	Stat s;
	unsigned i;

	s.pid = (int)ctl->id;
	if(0 > parsemaps(&s, err))
		return -1;
	for(i = 0; i < s.nr; i++)
		if(!strcmp(exepath, s.r[i].file))
			return s.r[i].addr;
	return -1;
}

int
syminitexec(Ctl *ctl, char **err)
{
	char *ldpath, *exepath;
	Bin *ld, *exe;
	u64 ldbase, exevbase, exeabase;
	u8 m;
	Sym *s;
	Trap t;
	Elf32_Ehdr *ehdr;

	exepath = ldpath = 0;
	exe = ld = 0;

	exepath = ctlexe(ctl, err);
	if(exepath == 0)
		goto fail;
	exevbase = loadaddr(exepath, err);
	if(exevbase == (u64)-1)
		goto fail;
	exeabase = mapaddr(ctl, exepath, err);
	if(exeabase == (u64)-1)
		goto fail;
	exe = mkbin(exepath, exeabase-exevbase, err);
	if(exe == 0)
		goto fail;
	exe->isexe = 1;
	addbin(ctl, exe);
	ctl->as->x.mode = exe->mode;
	if(0 > initaux(ctl, err))
		return -1;

	ldpath = interp(exepath, &m, err);
	if(ldpath == 0)
		return 0; /* static executable */
	ldbase = lookaux(ctl, AT_BASE);
	if(ldbase == (u64)-1)
		goto fail;
	ld = mkbin(ldpath, ldbase, err);
	if(ld == 0)
		goto fail;
	if(ctl->as->x.mode == M32){
			/*
			 * on amd64, 32-bit binaries that request
			 * /libexec/ld-elf.so.1 are actually loaded
			 * with /libexec/ld-elf32.so.1.  see usage
			 * of interp_newpath in sys/kern/imgact_elf.c
			 */
			ehdr = (Elf32_Ehdr*)ld->b;
			if(ehdr->e_machine != EM_386
			   && !strcmp(ldpath, "/libexec/ld-elf.so.1")){
				freebin(ld);
				ld = mkbin("/libexec/ld-elf32.so.1",
					   ldbase, err);
				if(ld == 0)
					goto fail;
			}
	}
	s = looksym(ld, "r_debug_state");
	if(s == 0)
		goto fail;
	t.addr = s->val;
	t.mode = Mld;
	if(!ctlsettrap(ctl, &t, err))
		goto fail;
#if 0
	t.addr = lookaux(ctl, AT_ENTRY);
	t.mode = Mld;
	if(!ctlsettrap(ctl, &t, err))
		goto fail;
#endif
	t.addr = lookaux(ctl, AT_ENTRY);
	t.mode = Mentry;
	if(!ctlsettrap(ctl, &t, err))
		goto fail;
	ctl->as->x.ld = ld;
	addbin(ctl, ld);

	return 0;
fail:
	/* FIXME: unlink ld/exe from ctl? */
	if(ld)
		freebin(ld);
	if(exe)
		freebin(exe);
	free(exepath);
	free(ldpath);
	if(*err == 0)
		*err = Ebin;
	return -1;
}

int
syminitattach(Ctl *ctl, char **err)
{
	*err = Eimpl;
	return -1;
}

int
symresetexec(Ctl *ctl, char **err)
{
	*err = Eimpl;
	return -1;
}
