#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../unix/unix.h"
#include "linux.h"

char osversion[] = "x86-linux-2012";

enum
{
	Maxinsn = 32,
	int3 = 0xcc,		/* breakpoint */
};

struct Unknown
{
	int pid;
	int status;
	struct Unknown *link;
};

typedef struct Unknown Unknown;
static Unknown *unknown;

void
dead(Ctl *ctl)
{
	_dead(ctl);
}

void
sctlinit()
{
	syminit();
	_sctlinit();
}

void
sctlfini()
{
	symfini();
	_sctlfini();
}

void*
ctlscratch(Ctl *ctl, char **err)
{
	Stat *s;
	u64 i;
	void *rv;

	if(ctl->as->x.scratch != 0)
		return ctl->as->x.scratch;

	s = ctlstat(ctl, err);
	if(s == 0)
		return 0;
	rv = 0;
	for(i = 0; i < s->nr; i++)
		if(s->r[i].flags&PROT_EXEC){
			rv = (void*)(uintptr_t)s->r[i].addr;
			break;
		}
	freestat(s);
	ctl->as->x.scratch = rv;
	return rv;
}

As*
mkas(Ctl *ctl)
{
	As *as;
	as = _mkas(ctl);
	as->x.mem = -1;
	return as;
}

void
freeas(As *as)
{
	symfreeas(as);
	_freeas(as);
}

static unsigned long
eventmsg(Ctl *ctl)
{
	unsigned long rv;
	if(0 > ptrace(PTRACE_GETEVENTMSG, (pid_t)ctl->id, 0, &rv))
		xfatalno("PTRACE_GETEVENTMSG");
	return rv;
}

static void
newunknown(int pid, int status)
{
	Unknown *u;
	u = xmalloc(sizeof(*u));
	u->pid = pid;
	u->status = status;
	u->link = unknown;
	unknown = u;
}

static int
wasunknown(int pid, int *status)
{
	Unknown **p, *u;
	p = &unknown;
	u = unknown;
	while(u){
		if(u->pid == pid){
			*status = u->status;
			*p = u->link;
			free(u);
			return 1;
		}
		p = &u->link;
		u = u->link;
	}
	return 0;
}

void
xgetregs(Ctl *ctl, Reg *regs)
{
	if(0 > ptrace(PTRACE_GETREGS, (pid_t)ctl->id, 0, regs))
		xfatalno("PTRACE_GETREGS");
}

void
xsetregs(Ctl *ctl, Reg *regs)
{
	if(0 > ptrace(PTRACE_SETREGS, (pid_t)ctl->id, 0, regs))
		xfatalno("PTRACE_SETREGS");
}

static void
getsiginfo(Ctl *ctl, siginfo_t *info)
{
	if(0 > ptrace(PTRACE_GETSIGINFO, (pid_t)ctl->id, 0, info))
		xfatalno("PTRACE_GETSIGINFO");
}

static Event*
stopevent(Ctl *ctl, int st)
{
	Ctl *spwn;
	int w, cpid, ev, est;
	Event *e;
	char *err;

	e = 0;
	ev = WSTOPEVENT(st);
	switch(ev){
	case PTRACE_EVENT_EXIT:
		est = eventmsg(ctl);
		ctl->state = Sstop;
		if(ctl->tracemode&Eexit){
			xgetregs(ctl, &ctl->ctx.regs);
			e = eventexit(ctl, est);
		}else
			ctlcont(ctl);
		break;
	case PTRACE_EVENT_CLONE:
		cpid = eventmsg(ctl);
		xdebug(Daction, "%d cloned %d", ctl->id, cpid);
		if(!wasunknown(cpid, &st)){
			w = waitpid(cpid, &st, __WALL);
			if(cpid != w)
				bug();
		}
		ctl->state = Sstop;
		spwn = mkctl(cpid);
		spwn->state = Sstop;
		spwn->whence = ctl->whence;
		xgetregs(spwn, &spwn->ctx.regs);
		linkspawn(ctl, spwn, 0);
		if(ctl->tracemode&Eclone)
			e = eventclone(ctl, spwn);
		else{
			ctlcont(ctl);
			ctlcont(spwn);
		}
		break;
	case PTRACE_EVENT_FORK:
	case PTRACE_EVENT_VFORK:
		cpid = eventmsg(ctl);
		xdebug(Daction, "%d forked %d", ctl->id, cpid);
		if(!wasunknown(cpid, &st)){
			w = waitpid(cpid, &st, __WALL);
			if(cpid != w)
				bug();
		}
		ctl->state = Sstop;
		spwn = mkctl(cpid);
		spwn->state = Sstop;
		spwn->whence = ctl->whence;
		xgetregs(spwn, &spwn->ctx.regs);
		linkspawn(ctl, spwn, 1);
		if(ctl->tracemode&Efork)
			e = eventfork(ctl, spwn);
		else{
			ctlcont(ctl);
			ctlcont(spwn);
		}
		break;
	case PTRACE_EVENT_EXEC:
		/* FIXME: clear traps and other as-specific state */
		xdebug(Daction, "%d called exec", ctl->id);
		ctl->state = Sstop;
		xgetregs(ctl, &ctl->ctx.regs);
		if(0 > resetexec(ctl, &err)){
			/* nowhere for this error to go */
			printf("we lost: %s\n", err);
			bug();
		}
		if(ctl->tracemode&Eexec)
			e = eventexec(ctl);
		else
			ctlcont(ctl);
		break;
	default:
		bug();
	}

	return e;
}

static Event*
doevent(Ctl *ctl, int st)
{
	Ctl *spwn;
	Brkref *br;
	char *err;
	Event *e;
	siginfo_t info;
	int m;
	Bin **bs;

	e = 0;
	if(WIFEXITED(st) || WIFSIGNALED(st))
		dead(ctl);
	else if(!WIFSTOPPED(st))
		bug();
	else if(WSTOPSIG(st) == SIGTRAP && WSTOPEVENT(st))
		e = stopevent(ctl, st);
	else if(WSTOPSIG(st) == (SIGTRAP|0x80)){
		if(ctl->tracemode&Esyscall){
			ctl->state = Sstop;
			xgetregs(ctl, &ctl->ctx.regs);
			e = eventsyscall(ctl);
		}else
			/* ignore spurious syscall trap */
			ctlcont(ctl);
	}
	else if(WSTOPSIG(st) == SIGTRAP){
		atbrkx86(ctl);
		br = atbrk(ctl);
		if(br){
			switch(br->bp->mode){
			case Msnap:
				spwn = ctlsnap(ctl);
				if(0 > steptrap(ctl, &err))
					xfatalno("steptrap: %s", err);
				ctlcont(ctl);
				e = eventsnap(ctl, br, spwn);
				break;
			case Mbrk:
				e = eventtrap(ctl, br);
				break;
			case Mld:
				m = ldevent(ctl, &bs, &err);
				if(0 > m)
					xfatal("ldevent: %s", err);
				if(0 > steptrap(ctl, &err))
					xfatalno("steptrap: %s", err);
				if(m == Eload)
					e = eventload(ctl, bs);
				else
					ctlcont(ctl);
				break;
			case Mentry:
				/* FIXME: what are we doing with this event? */
				if(0 > steptrap(ctl, &err))
					xfatalno("steptrap: %s", err);
				break;
			default:
				bug();
			}
		}else{
			if(ctl->state == Strap && 0 > steptrap(ctl, &err))
				xfatalno("cannot step over trap: %s", err);
			ctlcont(ctl);
		}
	}
	else{
		ctl->state = Sstop;
		if(ctl->ctx.cursig)
			xfatal("prctl: unimplemented: nested signals");
		xgetregs(ctl, &ctl->ctx.regs);
		ctl->ctx.cursig = WSTOPSIG(st);
		if(ctl->tracemode&Esignal){
			getsiginfo(ctl, &info);
			e = eventsignal(ctl, &info);
		}
		else
			ctlcont(ctl);
	}

	return e;
}

static Fd
ctlmem(Ctl *ctl)
{
	char buf[32];
	Fd fd;
	if(ctl->as->x.id == ctl->id)
		return ctl->as->x.mem;
	close(ctl->as->x.mem);
	snprintf(buf, sizeof(buf), "/proc/%" PRIu64 "/mem", ctl->id);
	fd = open(buf, O_RDONLY);
	if(0 > fd)
		xfatalno("open %s", buf);
	ctl->as->x.id = ctl->id;
	ctl->as->x.mem = fd;
	return fd;
}

void
ctlcont(Ctl *ctl)
{
	int how;
	if(ctl->tracemode&Esyscall)
		how = PTRACE_SYSCALL;
	else
		how = PTRACE_CONT;
	if(0 > ptrace(how, (pid_t)ctl->id, 0, ctl->ctx.cursig))
		xfatalno("PTRACE_CONT");
	ctl->ctx.cursig = 0;
	ctl->state = Srun;
}

static Event*
evqreverse(Event *p)
{
	Event *t, *q;

	q = 0;
	while(p){
		t = p->link;
		p->link = q;
		q = p;
		p = t;
	}
	return q;
}

Event*
ctlstop(Ctl *ctl, char **err)
{
	int st;
	Event *e, *p;

	p = 0;
	if(err)
		*err = 0;
	if(ctl->state != Srun)
		return 0;
	while(ctl->state == Srun){
		kill(ctl->id, SIGSTOP);
		if(0 > waitpid(ctl->id, &st, 0))
			xfatalno("waitpid");
		if(WIFSTOPPED(st) && WSTOPSIG(st) == SIGSTOP){
			xgetregs(ctl, &ctl->ctx.regs);
			ctl->state = Sstop;
			return evqreverse(p);
		}
		e = doevent(ctl, st);
		if(e){
			e->link = p;
			p = e;
		}
	}

	return evqreverse(p);
}

int
ctlstep(Ctl *ctl, char **err)
{
	int st;
	Event *e, *p;

	p = 0;
	if(err)
		*err = 0;
	if(0 > ptrace(PTRACE_SINGLESTEP, (pid_t)ctl->id, 0, 0))
		goto fail;
	while(1){
		if(0 > waitpid(ctl->id, &st, __WALL))
			goto fail;
		if(WIFSTOPPED(st) || WSTOPSIG(st) == SIGTRAP){
			if(ctl->tracemode&Estepctx)
				xgetregs(ctl, &ctl->ctx.regs);
			ctl->state = Sstop;
			if(p)
				bug(); /* FIXME */
			return 0;
		}
		e = doevent(ctl, st);
		if(e){
			e->link = p;
			p = e;
		}
	}
fail:
	if(err)
		*err = strerror(errno);
	return -1;
}

int
ctldetach(Ctl *ctl, char **err)
{
	u64 id;
	id = ctl->id;
	switch(ctl->state){
	case Srun:
		ctlstop(ctl, err);
		/* ignore errors, ignore events */
		/* fall through */
	case Strap:
	case Sstop:
		dead(ctl);
		if(0 > ptrace(PTRACE_DETACH, (pid_t)id, 0, 0))
			xfatalno("PTRACE_DETACH");
		break;
	case Sdead:
		return -1;
	default:
		bug();
	}
	return 0;
}

int
ctlkill(Ctl *ctl, char **err)
{
	int st;

	switch(ctl->state){
	case Srun:
		kill(ctl->id, SIGKILL);
		if(0 > waitpid(ctl->id, &st, 0))
			xfatalno("waitpid");
		break;
	case Strap:
	case Sstop:
		if(0 > ptrace(PTRACE_CONT, (pid_t)ctl->id, 0, SIGKILL))
			xfatalno("PTRACE_CONT");
		if(0 > waitpid(ctl->id, &st, 0))
			xfatalno("waitpid");
		break;
	case Sdead:
		return -1;
	default:
		bug();
	}
	dead(ctl);
	return 0;
}

static int
editbrk(Ctl *ctl, Brk *bp, int set, char **err)
{
	unsigned char *text;
	unsigned long idx, rv;
	int wasrunning;

	wasrunning = 0;
	if(ctl->state == Srun){
		wasrunning = 1;
		ctlstop(ctl, err);
		if(err && *err)
			return 0;
	}
	idx = bp->addr%4;
	errno = 0;
	rv = ptrace(PTRACE_PEEKTEXT, (pid_t)ctl->id,
		    (unsigned long)(bp->addr&(~3)), 0);
	if(errno)
		goto fail;
	text = (unsigned char*)&rv; /* FIXME: little endian assumption */
	if(set){
		bp->text = text[idx];
		text[idx] = int3;
	}else
		text[idx] = bp->text;
	if(0 > ptrace(PTRACE_POKETEXT, (pid_t)ctl->id,
		      (unsigned long)(bp->addr&(~3)), rv))
		goto fail;
	if(wasrunning)
		ctlcont(ctl);
	return 0;
fail:
	if(wasrunning)
		ctlcont(ctl);
	if(err)
		*err = strerror(errno);
	return -1;
}

int
clrbrk(Ctl *ctl, Brk *bp, char **err)
{
	/* FIXME: some thread may be stopped in this BP */
	return editbrk(ctl, bp, 0, err);
}

int
setbrk(Ctl *ctl, Brk *bp, char **err)
{
	return editbrk(ctl, bp, 1, err);
}
int defaultevents = (PTRACE_O_TRACESYSGOOD
		     |PTRACE_O_TRACECLONE
		     |PTRACE_O_TRACEFORK
		     |PTRACE_O_TRACEVFORK
		     |PTRACE_O_TRACEEXEC
		     |PTRACE_O_TRACEEXIT);

void
xtrace(Ctl *ctl, int events)
{
	if(0 > ptrace(PTRACE_SETOPTIONS, (pid_t)ctl->id, 0, events))
		xfatalno("PTRACE_SETOPTIONS");
}

int
ctltrace(Ctl *ctl, Trace *t, char **err)
{
	int rv;
	rv = _ctltrace(ctl, t, err);
	/* FIXME: call xtrace */
	if(rv == 0 && ctl->state == Srun){
		/* toggle system call tracing if necessary */
		ctlstop(ctl, err);
		if(err && *err)
			return -1;
		ctlcont(ctl);
	}
	return rv;
}

int
ctlread(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	Fd mem;
	mem = ctlmem(ctl);
	if((off64_t)-1 == lseek64(mem, addr, SEEK_SET)
	   || 0 > xread(mem, buf, cnt)){
		if(err)
			*err = strerror(errno);
		return -1;
	}
	return 0;
}

static int
ctlwritepart(Ctl *ctl, unsigned long addr, char *q, unsigned long cnt)
{
	unsigned long off, dat;
	char *p;
	unsigned i;
	Fd mem;

	mem = ctlmem(ctl);

	off = sizeof(long)*(addr/sizeof(long));
	if((off_t)-1 == lseek64(mem, off, SEEK_SET)
	   || 0 > xread(mem, &dat, sizeof(dat)))
		return -1;
	p = (char*)&dat + addr%sizeof(long);
	for(i = 0; i < cnt; i++)
		*p++ = *q++;
	if(0 > ptrace(PTRACE_POKEDATA, (pid_t)ctl->id, off, dat))
		return -1;
	return 0;
}

int
ctlwrite(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	unsigned long pcnt;
	unsigned long unit;
	unsigned long *wp;

	if(cnt == 0)
		return 0;

	unit = sizeof(long);

	if(addr%unit != 0){
		pcnt = sizeof(long) - (addr%unit);
		if(cnt < pcnt)
			pcnt = cnt;
		if(0 > ctlwritepart(ctl, addr, buf, pcnt))
			goto fail;
		addr += pcnt;
		cnt -= pcnt;
		buf += pcnt;
	}

	while(cnt >= unit){
		wp = (unsigned long*)buf;
		if(0 > ptrace(PTRACE_POKEDATA, (pid_t)ctl->id, addr, *wp))
			goto fail;
		buf += unit;
		cnt -= unit;
		addr += unit;
	}

	if(cnt > 0)
		if(0 > ctlwritepart(ctl, addr, buf, cnt))
			goto fail;

	return 0;
fail:
	if(err)
		*err = strerror(errno);
	return -1;
}

/*
  FIXME:
  Two methods of stepping over a trap are implemented here:

  1. clear, step, reset
  2. relocate the clobbered instruction

  #1 is unsafe in the presence of threads
  #2 does not work for calls, pc-sensitive insns, and surely others

  both bugs should be fixed: it would be nice to do #1 (which should
  be faster) unless the target spawns threads

*/
int
steptrap(Ctl *ctl, char **err)
{
#if 1
	if(0 > clrbrk(ctl, &ctl->curbrk, err) ||
	   0 > ctlstep(ctl, err) ||
	   0 > setbrk(ctl, &ctl->curbrk, err))
		return -1;
	return 0;
#else
	unsigned int idx;
	Ctl *cp;
	char buf[Maxinsn];
	u64 opc, npc, dpc;
	Reg regs;
	unsigned long saddr;
	void *scratch;

	scratch = ctlscratch(ctl, err);
	if(scratch == 0)
		return -1;
	cp = ctl->as->ctl;
	idx = 0;
	while(cp){
		if(cp == ctl)
			break;
		cp = cp->link;
	}
	if(idx*Maxinsn >= 4096)
		xfatal("steptrap: scratch exhausted");
	saddr = (unsigned long)ctl->as->x.scratch+idx*Maxinsn;
	if(0 > ctlread(ctl, ctl->curbrk.addr, buf, sizeof(buf), err))
		xfatalno("procread");
	buf[0] = ctl->curbrk.text;
	if(0 > ctlwrite(ctl, saddr, buf, sizeof(buf), err))
		xfatalno("procwrite");

	xgetregs(ctl, &regs);
	opc = regs.PC;
	npc = saddr;
	regs.PC = npc;
	xsetregs(ctl, &regs);
	if(0 > ctlstep(ctl, err))
		xfatal("step failed");
	xgetregs(ctl, &regs);
	dpc = regs.PC-npc;
	/* FIXME: detect control transfer instructions */
	if(dpc <= 16){
		regs.PC = opc + dpc;
		xsetregs(ctl, &regs);
	}
	return 0;
#endif
}

Event*
ctlwait(Ctl *maybectl)
{
	Ctl *ctl;
	int pid, st;

	if(maybectl){
		pid = waitpid(maybectl->id, &st, __WALL);
		if(pid == -1)
			return 0; /* no more procs */
		if(pid != maybectl->id)
			bug();
		return doevent(maybectl, st);
	}

	/* wait for any ctl */
again:
	pid = waitpid(-1, &st, __WALL);
	if(pid == -1)
		return 0; /* no more procs */
	if(pid == 0)
		bug();
	ctl = lookctl(pid);
	if(ctl == 0){
		/* unknown pid; assume it is a new thread, stopped.
		   we expect PTRACE_EVENT_CLONE on its parent.
		   remember that we have already seen the stop.
		   FIXME: happens often on 2.6.18, never on 2.6.24.
		*/
		newunknown(pid, st);
		goto again;
	}
	return doevent(ctl, st);
}

Event*
ctlpoll(Ctl *maybectl)
{
	Ctl *ctl;
	int st, pid;

	if(maybectl){
		pid = waitpid(maybectl->id, &st, WNOHANG|__WALL);
		if(pid == -1)
			return 0; /* no more procs */
		if(pid == 0)
			return 0; /* WNOHANG -> nothing waiting */
		if(pid != maybectl->id)
			bug();
		return doevent(maybectl, st);
	}
again:
	pid = waitpid(-1, &st, WNOHANG|__WALL);
	if(pid == -1)
		return 0; /* no more procs */
	if(pid == 0)
		return 0; /* WNOHANG -> nothing waiting */
	ctl = lookctl(pid);
	if(ctl == 0){
		/* unknown pid; assume it is a new thread, stopped.
		   we expect PTRACE_EVENT_CLONE on its parent.
		   remember that we have already seen the stop.
		   FIXME: happens often on 2.6.18, never on 2.6.24.
		*/
		newunknown(pid, st);
		goto again;
	}
	return doevent(ctl, st);
}
