#include "../inc/sctl.h"

typedef struct Hent Hent;
struct Hent {
	uptr key;
	void *val;
	Hent *next;
};

enum {
	INITSZ = 128,
};

struct HT {
	unsigned long sz;
	unsigned long nent;
	Hent **ht;		/* table */
	Hent **hent;		/* vector of allocated Hents */
	unsigned long (*hash)(uptr k);
	int (*eq)(uptr a, uptr b);
};

/* one-at-a-time by jenkins */
static unsigned long
shash(char *s, unsigned len)
{
	unsigned char *p = (unsigned char*)s;
	u32 h;

	h = 0;
	while(len != 0){
		h += *p;
		h += h<<10;
		h ^= h>>6;
		p++;
		len--;
	}
	h += h<<3;
	h ^= h>>11;
	h += h<<15;
	return (unsigned long)h;
}

static unsigned long
nhash(uptr k)
{
	return shash((char*)&k, sizeof(k));
}

static unsigned long
strhash(uptr k)
{
	return shash((char*)(uintptr_t)k, strlen((char*)(uintptr_t)k));
}

static int
neq(uptr a, uptr b)
{
	return a == b;
}

static int
seq(uptr a, uptr b)
{
	return !strcmp((char*)(uintptr_t)a, (char*)(uintptr_t)b);
}

static HT*
mkhtsz(unsigned long sz, unsigned long (*h)(uptr k), int (*eq)(uptr a, uptr b))
{
	HT *ht;
	ht = xmalloc(sizeof(HT));
	ht->sz = sz;
	ht->ht = xmalloc(ht->sz*sizeof(Hent*));
	ht->hent = xmalloc(ht->sz*sizeof(Hent*));
	ht->hash = h;
	ht->eq = eq;
	return ht;
}

HT*
mkht()
{
	return mkhtsz(INITSZ, nhash, neq);
}

HT*
mksht()
{
	return mkhtsz(INITSZ, strhash, seq);
}

void
freeht(HT* ht)
{
	Hent *hp, *np;
	unsigned long i;

	for(i = 0; i < ht->sz; i++){
		hp = ht->ht[i];
		while(hp){
			np = hp->next;
			free(hp);
			hp = np;
		}
	}
	free(ht->ht);
	free(ht->hent);
	free(ht);
}

void
hdel(HT *ht, uptr k)
{
	Hent *hp, **q;

	q = &ht->ht[ht->hash(k)%ht->sz];
	hp = *q;
	while(hp){
		if(ht->eq(k, hp->key)){
			*q = hp->next;
			free(hp);
			break;
		}
		q = &hp->next;
		hp = hp->next;
	}
}

static Hent*
_hget(HT *ht, uptr k)
{
	Hent *hp;

	hp = ht->ht[ht->hash(k)%ht->sz];
	while(hp){
		if(ht->eq(k, hp->key))
			return hp;
		hp = hp->next;
	}
	return 0;
}

void*
hget(HT *ht, uptr k)
{
	Hent *hp;

	hp = _hget(ht, k);
	if(hp)
		return hp->val;
	return 0;
}

/* double hash table size and re-hash entries */
static void
hexpand(HT *ht)
{
	Hent **nht, *hp, *nxt;
	unsigned nsz;
	unsigned i;
	unsigned long idx;

	nsz = ht->sz*2;
	nht = xmalloc(nsz*sizeof(sizeof(Hent*)));
	for(i = 0; i < ht->sz; i++){
		hp = ht->ht[i];
		while(hp){
			nxt = hp->next;			
			idx = ht->hash(hp->key)%nsz;
			hp->next = nht[idx];
			nht[idx] = hp;
			hp = nxt;
		}
	}
	free(ht->ht);
	ht->ht = nht;
	ht->hent = xrealloc(ht->hent,
			    ht->sz*sizeof(sizeof(Hent*)),
			    nsz*sizeof(sizeof(Hent*)));
	ht->sz = nsz;
}

void
hput(HT *ht, uptr k, void *v)
{
	Hent *hp;
	unsigned long idx;

	hp = _hget(ht, k);
	if(hp){
		hp->val = v;
		return;
	}

	if(3*ht->nent > 2*ht->sz)
		hexpand(ht);

	hp = xmalloc(sizeof(Hent));
	hp->key = k;
	hp->val = v;
	idx = ht->hash(k)%ht->sz;
	hp->next = ht->ht[idx];
	ht->ht[idx] = hp;
	ht->hent[ht->nent] = hp;
	ht->nent++;
}

void
hforeach(HT *ht, void (*f)(void *u, uptr k, void *v), void *u)
{
	Hent *hp;
	unsigned long i;

	for(i = 0; i < ht->sz; i++){
		hp = ht->ht[i];
		while(hp){
			f(u, hp->key, hp->val);
			hp = hp->next;
		}
	}
}

unsigned long
hnent(HT *ht)
{
	return ht->nent;
}

void*
hrefval(HT *ht, unsigned long idx)
{
	if(idx >= ht->nent)
		return 0;
	return ht->hent[idx]->val;
}

struct arg{
	HT *ht;
	unsigned long n;
};

static void
cmp1(void *u, uptr k, void *v)
{
	struct arg *a;
	a = u;
	if(hget(a->ht, k) != v)
		a->n++;		/* not found */
}

int
heq(HT *ha, HT *hb)
{
	struct arg a;

	if(hnent(ha) != hnent(hb))
		return 0;

	a.ht = hb;
	a.n = 0;
	hforeach(ha, cmp1, &a);
	if(a.n)
		return 0;
	a.ht = ha;
	a.n = 0;
	hforeach(hb, cmp1, &a);
	if(a.n)
		return 0;
	return 1;
}
