/* 
   C test driver program for prctl

   usage: ./test
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>


#include "sys.h"
#include "util.h"


#define PRCTL_IP      "127.0.0.1"
#define PRCTL_PORT    30000
#define TEST1_PROG    HOME_DIR "/test1"
#define TEST1_FOO     0x08048344

#define TIMETEST_PROG HOME_DIR "/timetest"
#define TIMETEST_FOO  0x08048394

#define NUM_ITERS     1000000

#define TIMER_DATA double

inline void Timer_Mark(TIMER_DATA* data) {
  struct timeval mark;
  gettimeofday(&mark, NULL);
  
  *data = ((double) mark.tv_sec) + (((double) mark.tv_usec) / 1000000.0);
}

inline double Timer_Elapsed(TIMER_DATA* start) {
  struct timeval mark;
  double end;
  gettimeofday(&mark, NULL);
  end = ((double) mark.tv_sec) + (((double) mark.tv_usec) / 1000000.0);
  
  return (end - (*start));
}



int 
connect_prctl()
{
	int fd;
	struct sockaddr_in sa;

	memset(&sa, 0, sizeof(sa));

	sa.sin_family = AF_INET;
	sa.sin_port = htons(PRCTL_PORT);	

	if (!inet_aton(PRCTL_IP, &sa.sin_addr)) {
		xfatal("Failed to translate " PRCTL_IP " into a valid IP address\n");
	}

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(0 > fd)
		xfatalno("socket");

	nodelay(fd);
	

	if (0 > connect(fd, (struct sockaddr*)&sa, sizeof(sa))) 
		xfatalno("connect");

	return fd;
}


uint64_t 
launch_prog(int fd, char *pname)
{
	Msg m, *r;
	uint64_t progid;
	char *argv[] = {pname, NULL};
	
	m.op = Tlaunch;
	m.argv = argv;
	
	writemsg(fd, &m);

	r = readmsg(fd);
	assert(r);
	if (r->op == Rerror) {
		printf("Error on launch: %s\n", r->err);
	}
	
	assert(r->op == Rlaunch);
	progid = r->id;

	freemsg(r);
	
	return progid;
}

uint64_t
bpset(int fd, uint64_t progid, uint64_t addr)
{
	Msg m, *r;
	uint64_t bpid;
	
	m.op = Tsetbrk;
	m.id = progid;
	m.addr = addr;

	writemsg(fd, &m);

	r = readmsg(fd);
	
	assert(r);
	assert(r->op == Rsetbrk);
	assert(r->id == progid);

	bpid = r->bp;

	freemsg(r);

	return bpid;
}

void
bpclear(int fd, uint64_t progid, uint64_t bpid)
{
	Msg m, *r;
	
	m.op = Tclrbrk;
	m.id = progid;
	m.bp = bpid;

	writemsg(fd, &m);

	r = readmsg(fd);
	
	assert(r);
	assert(r->op == Rclrbrk);
	assert(r->id == progid);
	freemsg(r);
}

void
step(int fd, uint64_t progid)
{
	Msg m, *r;
	
	m.op = Tstep;
	m.id = progid;

	writemsg(fd, &m);

	r = readmsg(fd);
	
	assert(r);
	assert(r->op == Rstep);
	assert(r->id == progid);

	freemsg(r);

}

void
run_until_break(int fd, uint64_t progid, uint64_t bpid)
{
	Msg m, *r;
	
	m.op = Tcont;
	m.id = progid;

	writemsg(fd, &m);

	r = readmsg(fd);

	assert(r);
	assert(r->op == Rcont);
	assert(r->id == progid);
	freemsg(r);

	r = readmsg(fd);
	assert(r);
	assert(r->op == Abrk);
	assert(r->id == progid);
	assert(r->bp == bpid);

	freemsg(r);

	m.op = Aack;
	m.id = progid;

	writemsg(fd, &m);	
}

void
run_until_end(int fd, uint64_t progid)
{
	Msg m, *r;
	
	m.op = Tcont;
	m.id = progid;

	writemsg(fd, &m);

	r = readmsg(fd);

	assert(r);
	assert(r->op == Rcont);
	assert(r->id == progid);
	freemsg(r);

	r = readmsg(fd);
	assert(r);
	assert(r->op == Aobit);
	assert(r->id == progid);

	freemsg(r);

	m.op = Aack;
	m.id = progid;

	writemsg(fd, &m);	
}

void send_recv_Ttest(int fd)
{
	Msg m, *r;

	m.op = Ttest;
	
	m.cnt = 0;
	m.bytes = NULL;

	writemsg(fd, &m);

	r = readmsg(fd);

	assert(r);
	assert(r->op == Rtest);
	assert(r->cnt == m.cnt);

	freemsg(r);
}

void
test0()
{
	int fd, i;

	fd = connect_prctl();
	
	for (i = 0; i < 10; i++) {
		printf("sending msg %d\n", i);
		send_recv_Ttest(fd);
	}

	close(fd);
}

void
msgtimetest()
{
	int fd, i;
	double before, after; 

	fd = connect_prctl();

	Timer_Mark(&before);
	for (i = 0; i < NUM_ITERS; i++)
		send_recv_Ttest(fd);
	after = Timer_Elapsed(&before);

	printf("Performed %d operations is %f seconds (%f ops/sec)\n", 
	       NUM_ITERS, after, ((double)NUM_ITERS) / (after));	

	close(fd);
}

void 
test1_a()
{
	int fd;
	uint64_t progid, bpid;

	fd = connect_prctl();

	progid = launch_prog(fd, TEST1_PROG);
	
	bpid = bpset(fd, progid, TEST1_FOO);

	printf("running until bp %ld\n", (unsigned long) bpid);

	run_until_break(fd, progid, bpid);
	printf("hit foo() once\n");

	run_until_break(fd, progid, bpid);
	printf("hit foo() twice\n");

	run_until_break(fd, progid, bpid);
	printf("hit foo() three\n");

	run_until_break(fd, progid, bpid);
	printf("hit foo() four\n");

	bpclear(fd, progid, bpid);

	run_until_end(fd, progid);
	printf("end of %s reached\n", TEST1_PROG);
	

	close(fd);
	
}

void 
timetest()
{
	int fd;
	uint64_t progid, bpid;
	double before, after; 
	int i;

	fd = connect_prctl();
	
	printf("trying to launch %s\n", TIMETEST_PROG);
	progid = launch_prog(fd, TIMETEST_PROG);
	
	printf("program launched\n");
	bpid = bpset(fd, progid, TIMETEST_FOO);

	printf("running until bp %ld\n", (unsigned long) bpid);

	run_until_break(fd, progid, bpid);
	printf("hit foo() once\n");
	bpclear(fd, progid, bpid);

	Timer_Mark(&before);
	for(i = 0; i < NUM_ITERS; i++) {
		step(fd, progid);
	}
	after = Timer_Elapsed(&before);

	printf("Performed %d operations is %f seconds (%f ops/sec)\n", 
	       NUM_ITERS, after, ((double)NUM_ITERS) / (after));
	close(fd);
	
}

int 
main(int argc, char *argv[])
{
	current_debug_level = Debug_None;
	
	// test0();
	// test1_a();
	// msgtimetest();
	timetest();
	
	return 0;
}

