#include "../inc/sctl.h"
#include "../inc/err.h"
#include "dwarf.h"
#include "dw.h"
#include "unix.h"
#include "ns.h"

#define ANONID "anonid_"
#define MIN(a,b) ((a)<(b)?(a):(b))

static Type* die2type(Bin *b, CU *cu, u32 off, u32 complete, char **err);

u64
uleb(char **p)
{
	char *q;
	u8 b;
	u64 rv;
	u32 s;

	q = *p;
	rv = 0;
	s = 0;
	while(1){
		b = (u8)*q++;
		rv |= (b&0x7f)<<s;
		s += 7;
		if((b&0x80) == 0)
			break;
	}
	*p = q;
	return rv;
}

s64
sleb32(char **p)
{
	char *q;
	u8 b;
	s64 rv;
	u32 s, m;

	m = 8*sizeof(s32);
	q = *p;
	rv = 0;
	s = 0;
	while(1){
		b = (u8)*q++;
		rv |= ((s64)(b&0x7f))<<s;
		s += 7;
		if((b&0x80) == 0)
			break;
	}
	if(s < m && (b&0x40))
		rv |= ~(((u64)1<<s)-1);
	*p = q;
	return rv;
}

static char*
getstr(char **p)
{
	char *q, *rv;
	rv = q = *p;
	while(*q++)
		;
	*p = q;
	return rv;
}

static Abbrevtab*
parseabbrev(char *p)
{
	u32 maxab, maxattr;
	Abbrev *ab;
	Attrspec *at;
	Abbrevtab *tab;
	u64 code, name, form;

	maxab = 128;
	tab = xmalloc(sizeof(Abbrevtab));
	tab->ab = xmalloc(maxab*sizeof(Abbrev));
	tab->nab = 1;   /* abbrev entries begin at one */
	while(1){
		code = uleb(&p);
		if(code == 0)
			break;
		if(tab->nab >= maxab){
			tab->ab = xrealloc(tab->ab,
					   maxab*sizeof(Abbrev),
					   2*maxab*sizeof(Abbrev));
			maxab *= 2;
		}
		ab = &tab->ab[tab->nab++];
		ab->tag = uleb(&p);
		XRD(&ab->haskids, p, sizeof(u8));
		maxattr = 8;
		ab->attr = xmalloc(maxattr*sizeof(Attrspec));
		while(1){
			name = uleb(&p);
			form = uleb(&p);
			if(name == 0 && form == 0)
				break;
			if(ab->nattr >= maxattr){
				ab->attr = xrealloc(ab->attr,
						    maxattr
						    *sizeof(Attrspec),
						    2*maxattr
						    *sizeof(Attrspec));
				maxattr *= 2;
			}
			at = &ab->attr[ab->nattr++];
			at->name = name;
			at->form = form;
		}
	}

	return tab;
}

#if 0
static void
dumpabbrevtab(Abbrevtab *tab)
{
	u32 i;
	Abbrev *ab;

	for(i = 0; i < tab->nab; i++){
		ab = &tab->ab[i];
		printf("%d\t%s\n", i, dwarftagname[ab->tag]);
	}
}
#endif

static void
freeabbrevtab(Abbrevtab *tab)
{
	u32 i;
	for(i = 0; i < tab->nab; i++)
		free(tab->ab[i].attr);
	free(tab->ab);
	free(tab);
}

Attr*
lookattr(Die *d, u32 n)
{
	u32 i;
	for(i = 0; i < d->nattr; i++)
		if(d->attr[i].name == n)
			return &d->attr[i];
	return 0;
}

static Attr*
lookattrwithspec(Die *d, u32 name)
{
	Attr *at;
	Die *sd;

	at = lookattr(d, name);
	if(at)
		return at;
	at = lookattr(d, DW_AT_specification);
	if(at == 0)
		return 0;
	sd = hget(d->cu->idx, at->x64);
	if(sd == 0)
		bug();
	return lookattr(sd, name);
}

enum
{
	Maxstack = 128,
};


static u64
dweval(Bin *b, char *p, u64 n)
{
	u8 op;
	char *ep;
	u8 x8;
	u16 x16;
	u32 x32;
	u64 x64;
	s8 sx8;
	s16 sx16;
	s32 sx32;
	s64 sx64;
	u64 stack[Maxstack], *sp;

	sp = stack+Maxstack-1;
	*sp = 0;
	ep = p+n;
	while(p < ep){
		if(sp < stack)
			xfatal("dwarf stack overflow");
		op = *p++;
		switch(op){
		case DW_OP_deref:
		case DW_OP_xderef:
		case DW_OP_deref_size:
		case DW_OP_xderef_size:
		case DW_OP_call2:
		case DW_OP_call4:
		case DW_OP_call_ref:
		case DW_OP_shra:
		case DW_OP_piece:
		case DW_OP_bit_piece:
		case DW_OP_push_object_address:
		case DW_OP_form_tls_address:
		case DW_OP_call_frame_cfa:
		case DW_OP_bregx:
			xfatal("dweval: %s unimplemented", dwarfopname[op]);
			break;
		default:
			xfatal("dweval: unknown op 0x%x", op);
			break;
		case DW_OP_nop:
			break;
		case DW_OP_addr:
			if(b->mode == M32){
				XRD(&x32, p, sizeof(x32));
				sp--;
				*sp = x32;
			}else{
				XRD(&x64, p, sizeof(x64));
				sp--;
				*sp = x64;
			}
			break;
		case DW_OP_const1u:
			XRD(&x8, p, sizeof(x8));
			sp--;
			*sp = x8;
			break;
		case DW_OP_const1s:
			XRD(&sx8, p, sizeof(sx8));
			sp--;
			*sp = sx8;
			break;
		case DW_OP_const2u:
			XRD(&x16, p, sizeof(x16));
			sp--;
			*sp = x16;
			break;
		case DW_OP_const2s:
			XRD(&sx16, p, sizeof(sx16));
			sp--;
			*sp = sx16;
			break;
		case DW_OP_const4u:
			XRD(&x32, p, sizeof(x32));
			sp--;
			*sp = x32;
			break;
		case DW_OP_const4s:
			XRD(&sx32, p, sizeof(sx32));
			sp--;
			*sp = sx32;
			break;
		case DW_OP_const8u:
			XRD(&x64, p, sizeof(x64));
			sp--;
			*sp = x64;
			break;
		case DW_OP_const8s:
			XRD(&sx64, p, sizeof(sx64));
			sp--;
			*sp = sx64;
			break;
		case DW_OP_constu:
			x64 = uleb(&p);
			sp--;
			*sp = x64;
			break;
		case DW_OP_consts:
			x64 = sleb32(&p);
			sp--;
			*sp = x64;
			break;
		case DW_OP_dup:
			sp--;
			*sp = sp[1];
			break;
		case DW_OP_drop:
			sp++;
			break;
		case DW_OP_over:
			sp--;
			*sp = sp[2];
			break;
		case DW_OP_pick:
			XRD(&x8, p, sizeof(x8));
			sp--;
			*sp = sp[x8+1];
			break;
		case DW_OP_swap:
			x64 = sp[1];
			sp[1] = sp[0];
			sp[0] = x64;
			break;
		case DW_OP_rot:
			x64 = sp[2];
			sp[2] = sp[1];
			sp[1] = sp[0];
			sp[0] = x64;
			break;
		case DW_OP_abs:
			sx64 = (s64)sp[0];
			if(sx64 < 0)
				sp[0] = -sp[0];
			break;
		case DW_OP_and:
			sp[0] = sp[0]&sp[1];
			break;
		case DW_OP_div:
			sp[1] = (s64)sp[1]/(s64)sp[0];
			sp++;
			break;
		case DW_OP_minus:
			sp[1] = sp[0]-sp[1];
			sp++;
			break;
		case DW_OP_mod:
			sp[1] = sp[1]%sp[0];
			sp++;
			break;
		case DW_OP_mul:
			sp[1] = sp[0]*sp[1];
			sp++;
			break;
		case DW_OP_neg:
			sp[0] = -sp[0];
			break;
		case DW_OP_not:
			sp[0] = ~sp[0];
			break;
		case DW_OP_or:
			sp[1] = sp[0]|sp[1];
			sp++;
			break;
		case DW_OP_plus:
			sp[1] = sp[0]+sp[1];
			sp++;
			break;
		case DW_OP_plus_uconst:
			x64 = uleb(&p);
			sp[0] = sp[0]+x64;
			break;
		case DW_OP_shl:
			sp[1] = sp[1]<<sp[0];
			sp++;
			break;
		case DW_OP_shr:
			sp[1] = sp[1]>>sp[0];
			sp++;
			break;
		case DW_OP_xor:
			sp[1] = sp[1]^sp[0];
			sp++;
			break;
		case DW_OP_eq:
			sp[1] = (s64)sp[1]==(s64)sp[0];
			sp++;
			break;
		case DW_OP_ge:
			sp[1] = (s64)sp[1]>=(s64)sp[0];
			sp++;
			break;
		case DW_OP_gt:
			sp[1] = (s64)sp[1]>(s64)sp[0];
			sp++;
			break;
		case DW_OP_le:
			sp[1] = (s64)sp[1]<=(s64)sp[0];
			sp++;
			break;
		case DW_OP_lt:
			sp[1] = (s64)sp[1]<(s64)sp[0];
			sp++;
			break;
		case DW_OP_ne:
			sp[1] = (s64)sp[1]!=(s64)sp[0];
			sp++;
			break;
		case DW_OP_bra:
			XRD(&sx16, p, sizeof(s16));
			if(sp[0])
				p += sx16;
			sp++;
			break;
		case DW_OP_skip:
			XRD(&sx16, p, sizeof(s16));
			p += sx16;
			break;
		case DW_OP_lit0:
		case DW_OP_lit1:
		case DW_OP_lit2:
		case DW_OP_lit3:
		case DW_OP_lit4:
		case DW_OP_lit5:
		case DW_OP_lit6:
		case DW_OP_lit7:
		case DW_OP_lit8:
		case DW_OP_lit9:
		case DW_OP_lit10:
		case DW_OP_lit11:
		case DW_OP_lit12:
		case DW_OP_lit13:
		case DW_OP_lit14:
		case DW_OP_lit15:
		case DW_OP_lit16:
		case DW_OP_lit17:
		case DW_OP_lit18:
		case DW_OP_lit19:
		case DW_OP_lit20:
		case DW_OP_lit21:
		case DW_OP_lit22:
		case DW_OP_lit23:
		case DW_OP_lit24:
		case DW_OP_lit25:
		case DW_OP_lit26:
		case DW_OP_lit27:
		case DW_OP_lit28:
		case DW_OP_lit29:
		case DW_OP_lit30:
		case DW_OP_lit31:
			sp--;
			*sp = op-DW_OP_lit0;
			break;
		case DW_OP_reg0:
		case DW_OP_reg1:
		case DW_OP_reg2:
		case DW_OP_reg3:
		case DW_OP_reg4:
		case DW_OP_reg5:
		case DW_OP_reg6:
		case DW_OP_reg7:
		case DW_OP_reg8:
		case DW_OP_reg9:
		case DW_OP_reg10:
		case DW_OP_reg11:
		case DW_OP_reg12:
		case DW_OP_reg13:
		case DW_OP_reg14:
		case DW_OP_reg15:
		case DW_OP_reg16:
		case DW_OP_reg17:
		case DW_OP_reg18:
		case DW_OP_reg19:
		case DW_OP_reg20:
		case DW_OP_reg21:
		case DW_OP_reg22:
		case DW_OP_reg23:
		case DW_OP_reg24:
		case DW_OP_reg25:
		case DW_OP_reg26:
		case DW_OP_reg27:
		case DW_OP_reg28:
		case DW_OP_reg29:
		case DW_OP_reg30:
		case DW_OP_reg31:
		case DW_OP_regx:
		case DW_OP_fbreg:
		case DW_OP_breg0:
		case DW_OP_breg1:
		case DW_OP_breg2:
		case DW_OP_breg3:
		case DW_OP_breg4:
		case DW_OP_breg5:
		case DW_OP_breg6:
		case DW_OP_breg7:
		case DW_OP_breg8:
		case DW_OP_breg9:
		case DW_OP_breg10:
		case DW_OP_breg11:
		case DW_OP_breg12:
		case DW_OP_breg13:
		case DW_OP_breg14:
		case DW_OP_breg15:
		case DW_OP_breg16:
		case DW_OP_breg17:
		case DW_OP_breg18:
		case DW_OP_breg19:
		case DW_OP_breg20:
		case DW_OP_breg21:
		case DW_OP_breg22:
		case DW_OP_breg23:
		case DW_OP_breg24:
		case DW_OP_breg25:
		case DW_OP_breg26:
		case DW_OP_breg27:
		case DW_OP_breg28:
		case DW_OP_breg29:
		case DW_OP_breg30:
		case DW_OP_breg31:
			xfatal("dwarf operation requires access to registers");
			break;
		}
	}
	if(sp < stack+Maxstack)
		return sp[0];
	return -1;
}

static u64
at2u64(Bin *b, Attr *at)
{
	switch(at->form){
	case DW_FORM_block:
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
		return dweval(b, at->block, at->x64);
	case DW_FORM_sdata:
		return (u64)at->sx64;
	default:
		xfatal("unexpected u64 form %s", dwarfformname[at->form]);
		return 0;
	}
}

static void
freedies(Die *d)
{
	if(d == 0)
		return;
	freedies(d->kid);
	freedies(d->sib);
	free(d->attr);
	free(d->id);
	free(d);
}

static void
freecu(CU *cu)
{
	freedies(cu->d);
	freeht(cu->idx);
	freesrc(cu->src);
	free(cu);
}

static CU*
mkcu()
{
	CU *cu;
	cu = xmalloc(sizeof(CU));
	cu->idx = mkht();
	return cu;
}

static void
freediet(Diet *t)
{
	free(t);
}

static Diet*
mkdiet(u32 cuoff, u32 dieoff)
{
	Diet *t;
	t = xmalloc(sizeof(Diet));
	t->cuoff = cuoff;
	t->dieoff = dieoff;
	return t;
}

static Diets*
mkdiets()
{
	Diets *s;
	s = xmalloc(sizeof(Diets));
	s->max = 1024;
	s->diet = xmalloc(s->max*sizeof(Diet*));
	return s;
}

static void
adddiet(Diet *d, Diets *s)
{
	if(s->n >= s->max){
		s->diet = xrealloc(s->diet, s->max*sizeof(Diet*),
				   2*s->max*sizeof(Diet*));
		s->max *= 2;
	}
	s->diet[s->n++] = d;
}

static void
freediets(Diets *s)
{
	free(s->diet);
	free(s);
}

static Alist*
mkalist(int (*eq)(void *, void*))
{
	Alist *al;
	al = xmalloc(sizeof(Alist));
	al->eq = eq;
	al->max = 1024;
	al->k = xmalloc(al->max*sizeof(void*));
	al->v = xmalloc(al->max*sizeof(void*));
	return al;
}

static void*
lookalist(void *k, Alist *al)
{
	u32 i;
	for(i = 0; i < al->n; i++)
		if(al->eq(k, al->k[i]))
			return al->v[i];
	return 0;
}

static void
insalist(void *k, void *v, Alist *al)
{
	if(al->n >= al->max){
		al->k = xrealloc(al->k, al->max*sizeof(void*),
				 2*al->max*sizeof(void*));
		al->v = xrealloc(al->v, al->max*sizeof(void*),
				 2*al->max*sizeof(void*));
		al->max *= 2;
	}
	al->k[al->n] = k;
	al->v[al->n] = v;
	al->n++;
}

static void
freealist(Alist *al)
{
	free(al->k);
	free(al->v);
	free(al);
}

static int
istypetag(u32 tag)
{
	switch(tag){
	case DW_TAG_array_type:
	case DW_TAG_class_type:
	case DW_TAG_enumeration_type:
	case DW_TAG_pointer_type:
	case DW_TAG_reference_type:
	case DW_TAG_string_type:
	case DW_TAG_structure_type:
	case DW_TAG_subroutine_type:
	case DW_TAG_union_type:
	case DW_TAG_typedef:
	case DW_TAG_ptr_to_member_type:
	case DW_TAG_set_type:
	case DW_TAG_subrange_type:
	case DW_TAG_base_type:
	case DW_TAG_const_type:
	case DW_TAG_file_type:
	case DW_TAG_packed_type:
	case DW_TAG_thrown_type:
	case DW_TAG_volatile_type:
	case DW_TAG_restrict_type:
	case DW_TAG_unspecified_type:
	case DW_TAG_shared_type:
	case DW_TAG_upc_shared_type:
	case DW_TAG_upc_strict_type:
	case DW_TAG_upc_relaxed_type:
		return 1;
	default:
		return 0;
	}
}

static int
issymtag(u32 tag)
{
	switch(tag){
	case DW_TAG_subprogram:
	case DW_TAG_variable:
		return 1;
	default:
		return 0;
	}
}

/* FIXME: rationalize with die2type interface and perhaps just
   fold into die2type since our only caller calls both. */
static int
typesize(HT *tt, u64 idx, u64 *sz)
{
	Die *t;
	Attr *at;
	u64 x;
	*sz = -1;
	t = hget(tt, idx);
	if(t == 0)
		return -1;
	switch(t->tag){
	case DW_TAG_base_type:
	case DW_TAG_class_type:
	case DW_TAG_structure_type:
	case DW_TAG_union_type:
	case DW_TAG_enumeration_type:
		at = lookattr(t, DW_AT_byte_size);
		if(at == 0)
			return -1;
		*sz = at->x64;
		return 0;
	case DW_TAG_array_type:
		at = lookattr(t, DW_AT_byte_size);
		if(at){
			*sz = at->x64;
			return 0;
		}
		at = lookattr(t, DW_AT_type);
		if(at == 0)
			return -1;
		if(0 > typesize(tt, at->x64, &x))
			return -1;
		if(t->kid){
			t = t->kid;
			// FIXME: here we expect subrange_type children.
			// what is the meaning of DW_AT_type attribute of these
			// children?
			// spec also allows enumeration_type children.
			// what do these mean?
			while(t){
				if(t->tag != DW_TAG_subrange_type){
					xlog("expected subrange but got %s",
					     dwarftagname[t->tag]);
					return -1;
				}
				// FIXME: what about DW_AT_count?
				at = lookattr(t, DW_AT_upper_bound);
				if(at)
					// if present, assume 0-base index;
					// if not, assume it has form "a[0]"
					x *= at->x64+1;
				t = t->sib;
			}
		}else
			/* assume zero-element array */
			x = 0;
		*sz = x;
		return 0;
	case DW_TAG_pointer_type:
	case DW_TAG_reference_type:
		at = lookattr(t, DW_AT_byte_size);
		if(at == 0)
			return -1;
		*sz = at->x64;
		return 0;
	case DW_TAG_typedef:
	case DW_TAG_const_type:
	case DW_TAG_packed_type:
	case DW_TAG_volatile_type:
	case DW_TAG_restrict_type:
	case DW_TAG_shared_type:
		at = lookattr(t, DW_AT_type);
		if(at == 0)
			return -1;
		return typesize(tt, at->x64, sz);
	case DW_TAG_subroutine_type:
	case DW_TAG_string_type:
	case DW_TAG_file_type:
	case DW_TAG_thrown_type:
	case DW_TAG_unspecified_type:
	case DW_TAG_upc_shared_type:
	case DW_TAG_upc_strict_type:
	case DW_TAG_upc_relaxed_type:
	case DW_TAG_ptr_to_member_type:
	case DW_TAG_set_type:
		xfatal("typesize: unimplemented type %s", dwarftagname[t->tag]);
	default:
		xfatal("typesize: unexpected type %s", dwarftagname[t->tag]);
	}
	return -1;
}

// base type name tokens
typedef
enum Btok
{
	Kunsigned,
	Ksigned,
	Kint,
	Kchar,
	Klong,
	Kshort,
	Kfloat,
	Kdouble,
	Kcomplex,
	Kbool,
	Kvoid,   /* not used */
	Kntok,
} Btok;

static struct btok {
	char *base;
	Btok k;
} btok[] = {
	{ "unsigned",	Kunsigned },
	{ "signed",	Ksigned },
	{ "bool",	Kbool },
	{ "_Bool",	Kbool },
	{ "char",	Kchar },
	{ "short",	Kshort },
	{ "int",	Kint },
	{ "wchar_t",	Kint }, /* FIXME */
	{ "long",	Klong },
	{ "float",	Kfloat },
	{ "double",	Kdouble },
	{ "void",	Kvoid },
	{ "complex",	Kcomplex },
	{ "_Complex",	Kcomplex },
	{ 0,		Kntok },
};

static Cbase basemod[Vnbase][Kntok] = {
	[Vchar][Kunsigned]      = Vuchar,
	[Vchar][Ksigned]        = Vchar,

	[Vshort][Kunsigned]     = Vushort,
	[Vshort][Ksigned]       = Vshort,
	[Vshort][Kint]          = Vshort,
	[Vushort][Kint]         = Vushort,

	[Vint][Kunsigned]       = Vuint,
	[Vint][Ksigned]         = Vint,
	[Vint][Kchar]	        = Vchar,
	[Vuint][Kchar]	        = Vuchar,
	[Vint][Klong]           = Vlong,
	[Vuint][Klong]          = Vulong,
	[Vint][Kshort]          = Vshort,
	[Vuint][Kshort]         = Vushort,
	[Vint][Kint]	        = Vint,
	[Vuint][Kint]	        = Vuint,

	[Vlong][Kunsigned]      = Vulong,
	[Vlong][Ksigned]        = Vlong,
	[Vlong][Klong]          = Vvlong,
	[Vlong][Kdouble]        = Vlongdouble,
	[Vulong][Klong]         = Vuvlong,
	[Vlong][Kint]	        = Vlong,
	[Vulong][Kint]	        = Vulong,

	[Vvlong][Kunsigned]     = Vuvlong,
	[Vvlong][Ksigned]       = Vvlong,
	[Vvlong][Kint]	        = Vvlong,
	[Vuvlong][Kint]	        = Vuvlong,

	[Vlong][Kcomplex]       = Vlongdoublex,
	[Vfloat][Kcomplex]      = Vcomplex,
	[Vdouble][Klong]        = Vlongdouble,
	[Vdouble][Kcomplex]     = Vdoublex,
	[Vcomplex][Kfloat]      = Vcomplex,
	[Vcomplex][Kdouble]     = Vdoublex,
	[Vcomplex][Klong]       = Vlongdoublex,
	[Vdoublex][Klong]       = Vlongdoublex,
	[Vlongdoublex][Kdouble] = Vlongdoublex,

	[Vundef][Kbool]         = Vbool,
	[Vundef][Kchar]         = Vchar,
	[Vundef][Kdouble]       = Vdouble,
	[Vundef][Kfloat]        = Vfloat,
	[Vundef][Kint]          = Vint,
	[Vundef][Klong]         = Vlong,
	[Vundef][Kshort]        = Vshort,
	[Vundef][Ksigned]       = Vint,
	[Vundef][Kunsigned]     = Vuint,
	[Vundef][Kcomplex]      = Vcomplex,
	[Vundef][Kvoid]         = Vvoid,
	/* the rest are Vundef, which we assume to be 0 */
};

static Btok
lookbtok(char *b)
{
	struct btok *p;
	p = btok;
	while(p->base){
		if(!strcmp(b, p->base))
			return p->k;
		p++;
	}
	return Kntok;
}

static Cbase
die2cbase(Die *d)
{
	Attr *at;
	char *p, *t, *q, *e;
	Btok k;
	Cbase rv;

	rv = Vundef;
	at = lookattrwithspec(d, DW_AT_name);
	if(at == 0)
		return rv;
	p = xstrdup(at->str);
	e = p+strlen(p)+1;
	t = p;
	while(t < e){
		q = strchr(t, ' ');
		if(q)
			*q = 0;
		else
			q = e;  /* last iteration */
		k = lookbtok(t);
		if(k == Kntok){
			rv = Vundef;
			goto out;
		}
		rv = basemod[rv][k];
		t = q+1;
		while(t < e && *t == ' ')
			t++;
	}
out:
	free(p);
	return rv;
}

enum
{
	Maxrepwidth = 24
};

static Rkind lereptab[Vnbase][Maxrepwidth+1] = {
	/* undefined combinations map to 0 (Rundef).
	   some undefined combinations are sensible but uncommon:
	   add definitions as needed */
	[Vbool][1]		= Ru08le,
	[Vchar][1]		= Rs08le,
	[Vuchar][1]		= Ru08le,
	[Vshort][2]		= Rs16le,
	[Vushort][2]		= Ru16le,
	[Vint][2]		= Rs16le,
	[Vint][4]		= Rs32le,
	[Vint][8]		= Rs64le,
	[Vuint][2]		= Ru16le,
	[Vuint][4]		= Ru32le,
	[Vuint][8]		= Ru64le,
	[Vlong][4]		= Rs32le,
	[Vlong][8]		= Rs64le,
	[Vulong][4]		= Ru32le,
	[Vulong][8]		= Ru64le,
	[Vvlong][8]		= Rs64le,
	[Vuvlong][8]		= Ru64le,
	[Vfloat][4]		= Rf32,
	[Vfloat][8]		= Rf64,
	[Vfloat][16]		= Rf128,
	[Vdouble][4]		= Rf32,
	[Vdouble][8]		= Rf64,
	[Vdouble][16]		= Rf128,
	[Vlongdouble][4]	= Rf32,
	[Vlongdouble][8]	= Rf64,
	[Vlongdouble][12]	= Rf96,
	[Vlongdouble][16]	= Rf128,
	[Vcomplex][8]		= Rx64,
	[Vdoublex][16]		= Rx128,
	[Vlongdoublex][24]	= Rx192,
};

static Rkind bereptab[Vnbase][Maxrepwidth+1] = {
	/* undefined combinations map to 0 (Rundef).
	   some undefined combinations are sensible but uncommon:
	   add definitions as needed */
	[Vbool][1]		= Ru08be,
	[Vchar][1]		= Rs08be,
	[Vuchar][1]		= Ru08be,
	[Vshort][2]		= Rs16be,
	[Vushort][2]		= Ru16be,
	[Vint][2]		= Rs16be,
	[Vint][4]		= Rs32be,
	[Vint][8]		= Rs64be,
	[Vuint][2]		= Ru16be,
	[Vuint][4]		= Ru32be,
	[Vuint][8]		= Ru64be,
	[Vlong][4]		= Rs32be,
	[Vlong][8]		= Rs64be,
	[Vulong][4]		= Ru32be,
	[Vulong][8]		= Ru64be,
	[Vvlong][8]		= Rs64be,
	[Vuvlong][8]		= Ru64be,
	[Vfloat][4]		= Rf32,
	[Vfloat][8]		= Rf64,
	[Vfloat][16]		= Rf128,
	[Vdouble][4]		= Rf32,
	[Vdouble][8]		= Rf64,
	[Vdouble][16]		= Rf128,
	[Vlongdouble][4]	= Rf32,
	[Vlongdouble][8]	= Rf64,
	[Vlongdouble][12]	= Rf96,
	[Vlongdouble][16]	= Rf128,
	[Vcomplex][8]		= Rx64,
	[Vdoublex][16]		= Rx128,
	[Vlongdoublex][24]	= Rx192,
};

static int
die2base(Bin *b, Die *d, Type *t)
{
	Cbase cb;
	Attr *at;
	Rkind rep;
	u64 sz;

	cb = die2cbase(d);
	if(cb == Vundef)
		return -1;
	at = lookattr(d, DW_AT_byte_size);
	if(at == 0)
		return -1;
	if(at->x64 > Maxrepwidth){
		xlog("bad DW_AT_byte_size: %lu", at->x64);
		return -1;
	}
	if(b->enc == Mle)
		rep = lereptab[cb][at->x64];
	else
		rep = bereptab[cb][at->x64];
	if(rep == Rundef){
		sz = at->x64;
		at = lookattrwithspec(d, DW_AT_name);
		xlog("warning: no representation for %lu-byte '%s'",
		     sz, at->str);
		rep = Rundef;
	}
	t->kind = Tbase;
	t->base = cb;
	t->rep = rep;
	return 0;
}

static void
notebase(Bin *b, Die *d)
{
	Type t;
	if(0 > die2base(b, d, &t))
		return;
	b->dw.base[t.base].rep = t.rep;
}

static void
noteptr(Bin *b, Die *d)
{
	Attr *at;
	at = lookattr(d, DW_AT_byte_size);
	if(at == 0){
		xlog("no DW_AT_byte_size for DW_TAG_pointer type!");
		return;
	}
	switch(at->x64){
	case 4:
		if(b->enc == Mle)
			b->dw.base[Vptr].rep = Ru32le;
		else
			b->dw.base[Vptr].rep = Ru32be;
		break;
	case 8:
		if(b->enc == Mle)
			b->dw.base[Vptr].rep = Ru64le;
		else
			b->dw.base[Vptr].rep = Ru64be;
		break;
	default:
		xlog("unexpected byte size for DW_TAG_pointer type: %lu",
		     at->x64);
	}
}

static char*
mkanontag(CU *cu, u32 dieoff)
{
	char buf[7+1+16+1+16+1];
	sprintf(buf, "anontag_%x_%x", cu->off, dieoff);
	return xstrdup(buf);
}

static char*
mkanonid(u32 cuoff, u32 dieoff)
{
	char buf[sizeof(ANONID)+16+1+16+1];
	sprintf(buf, ANONID "%x_%x", cuoff, dieoff);
	return xstrdup(buf);
}

static char*
mkanon(u32 n)
{
	char buf[4+1+9+1];
	sprintf(buf, "anon_%u", n);
	return xstrdup(buf);
}

static Die*
readdie(Bin *b, CU *cu, char **pp, Abbrevtab *tab, Die *parent)
{
	Abbrev *ab;
	Attr *at;
	Attrspec *as;
	char *p, *s, *id;
	Die *d;
	Diet *e;
	u8 x8;
	u16 x16;
	u32 x32, i;
	u64 x64, a;
	s64 sx64;
	int z;
	u32 off;

	z = 0;
	d = 0;
	p = *pp;
	a = uleb(&p);
	if(a == 0)
		goto out;
	if(a >= tab->nab)
		xfatal("bad abbrev: %ld", a);
	ab = &tab->ab[a];
	d = xmalloc(sizeof(Die));
	d->cu = cu;
	d->ab = a;
	d->tag = ab->tag;
	d->nattr = ab->nattr;
	d->attr = xmalloc(ab->nattr*sizeof(Attr));
	s = 0;
	for(i = 0; i < ab->nattr; i++){
		as = &ab->attr[i];
		at = &d->attr[i];
		at->name = as->name;
		at->form = as->form;
		switch(as->form){
		case DW_FORM_addr:
			switch(cu->sz){
			case 4:
				XRD(&x32, p, cu->sz);
				at->x64 = x32;
				break;
			case 8:
				XRD(&x64, p, cu->sz);
				at->x64 = x64;
				break;
			default:
				xfatal("unhandled DW_FORM_addr size %d",
				       cu->sz);
			}
			break;
		case DW_FORM_block:
			x64 = uleb(&p);
			at->block = p;
			at->x64 = x64;
			p += x64;
			break;
		case DW_FORM_block1:
			XRD(&x8, p, sizeof(u8));
			at->block = p;
			at->x64 = x8;
			p += x8;
			break;
		case DW_FORM_block2:
			XRD(&x16, p, sizeof(u16));
			at->block = p;
			at->x64 = x16;
			p += x16;
			break;
		case DW_FORM_block4:
			XRD(&x32, p, sizeof(u32));
			at->block = p;
			at->x64 = x32;
			p += x32;
			break;
		case DW_FORM_data1:
			XRD(&x8, p, sizeof(u8));
			at->x64 = x8;
			break;
		case DW_FORM_data2:
			XRD(&x16, p, sizeof(u16));
			at->x64 = x16;
			break;
		case DW_FORM_data4:
			XRD(&x32, p, sizeof(u32));
			at->x64 = x32;
			break;
		case DW_FORM_data8:
			XRD(&x64, p, sizeof(u64));
			at->x64 = x64;
			break;
		case DW_FORM_udata:
			x64 = uleb(&p);
			at->x64 = x64;
			break;
		case DW_FORM_sdata:
			sx64 = sleb32(&p);
			at->sx64 = sx64;
			break;
		case DW_FORM_flag:
			XRD(&x8, p, sizeof(u8));
			at->x64 = x8;
			break;
		case DW_FORM_string:
			s = getstr(&p);
			z++;
			at->str = s;
			if(s == 0)
				xfatal("null s");
			break;
		case DW_FORM_strp:
			XRD(&x32, p, sizeof(u32));
			z++;
			s = b->dw.str+x32;
			if(s == 0)
				xfatal("null s");
			at->str = s;
			break;
		case DW_FORM_ref_addr:
			/* assume 32-bit dwarf format */
			XRD(&x32, p, sizeof(u32));
			at->x64 = x32;
			break;
		case DW_FORM_ref1:
			XRD(&x8, p, sizeof(u8));
			at->x64 = x8;
			break;
		case DW_FORM_ref2:
			XRD(&x16, p, sizeof(u16));
			at->x64 = x16;
			break;
		case DW_FORM_ref4:
			XRD(&x32, p, sizeof(u32));
			at->x64 = x32;
			break;
		case DW_FORM_ref8:
			XRD(&x64, p, sizeof(u64));
			at->x64 = x64;
			break;
		case DW_FORM_ref_udata:
			x64 = uleb(&p);
			at->x64 = x64;
			break;
		case DW_FORM_indirect:
			xfatal("i don't speak DW_FORM_indirect");
			break;
		default:
			xfatal("unknown dwarf form %d", at->form);
			break;
		}
	}

	if(istypetag(d->tag)){
		// remember type at this offset
		off = *pp-cu->p;
		hput(cu->idx, off, d);
		switch(d->tag){
		case DW_TAG_enumeration_type:
		case DW_TAG_structure_type:
		case DW_TAG_union_type:
			at = lookattrwithspec(d, DW_AT_name);
			if(at){
				id = cidentifier(at->str);
				e = hget(b->dw.sue, (uptr)id);
				if(!e){
					hput(b->dw.sue, (uptr)xstrdup(id),
					     mkdiet(cu->off, off));
				}else{
					/* prefer non-zero-sized definitions */
					at = lookattr(d, DW_AT_byte_size);
					if(at != 0 && at->x64 != 0){
						freediet(e);
						hput(b->dw.sue, (uptr)id,
						     mkdiet(cu->off, off));
					}
				}
			}else{
				s = mkanontag(cu, off);
				if(!hget(b->dw.ant, (uptr)s))
					hput(b->dw.ant, (uptr)s,
					     mkdiet(cu->off, off));
				else
					free(s);
			}
			break;
		case DW_TAG_typedef:
			at = lookattrwithspec(d, DW_AT_name);
			if(at == 0)
				xfatal("unexpected anonymous typedef");
			id = cidentifier(at->str);
			if(!hget(b->dw.typ, (uptr)id))
				hput(b->dw.typ, (uptr)xstrdup(id),
				     mkdiet(cu->off, off));
			break;
		case DW_TAG_base_type:
			notebase(b, d);
			break;
		case DW_TAG_pointer_type:
			if(b->dw.base[Vptr].rep == Rundef)
				noteptr(b, d);
			break;
		}
	}
	if(issymtag(d->tag)){
		at = lookattrwithspec(d, DW_AT_name);
		if(at){
			/* remember all named symbol dies, in case there is
			   a later reference by specification */
			off = *pp-cu->p;
			hput(cu->idx, off, d);
			if(!parent){
				/* this is a top-level symbol */
				id = cidentifier(at->str);
				if(!hget(b->dw.sym, (uptr)id))
					hput(b->dw.sym, (uptr)xstrdup(id),
					     mkdiet(cu->off, off));
			}
		}
	}
out:
	*pp = p;
	return d;

}

static Die*
rec(Bin *b, CU *cu, char **p, Abbrevtab *tab, Die *r)
{
	Die *root, *d;

	root = d = readdie(b, cu, p, tab, r);
	while(d){
		if(tab->ab[d->ab].haskids)
			d->kid = rec(b, cu, p, tab, d);
		d->sib = readdie(b, cu, p, tab, r);
		d = d->sib;
	}
	return root;
}

static Die*
readdies(Bin *b, CU *cu, char **p, Abbrevtab *tab)
{
	Die *d;
	d = readdie(b, cu, p, tab, 0);
	if(tab->ab[d->ab].haskids)
		/* treat cu entries as series of unrooted trees */
		d->kid = rec(b, cu, p, tab, 0);
	return d;
}

static int
cusrcinfo(CU *cu)
{
	Attr *at;
	Die *d;

	d = cu->d;
	if(d->tag != DW_TAG_compile_unit){
		xlog("dwsrc: expected compile unit, got %s",
		     dwarftagname[cu->d->tag]);
		return -1;
	}
	at = lookattr(d, DW_AT_comp_dir);
	if(at == 0){
//		xlog("compile unit %p does not define compilation directory",
//		     cu->off);
		cu->dir = "unknown-dir";
	}else
		cu->dir = at->str;

	at = lookattr(d, DW_AT_stmt_list);
	if(at == 0){
		xlog("compile unit %p does not define stmt_list", cu->off);
		cu->soff = -1;
	}else
		cu->soff = at->x64;

	at = lookattr(d, DW_AT_name);
	if(at == 0){
		xlog("compile unit %p does not define name", cu->off);
		cu->file = "unknown-file";
	}else
		cu->file = at->str;
#if 0
	at = lookattr(d, DW_AT_low_pc);
	if(at == 0){
		xlog("compile unit %p does not define low_pc",
		     cu->off);
		lo = -1;
	}else
		lo = at->x64;
	
	at = lookattr(d, DW_AT_high_pc);
	if(at == 0){
		xlog("compile unit %p does not define high_pc",
		     cu->off);
		hi = -1;
	}else
		hi = at->x64;
#endif
	return 0;
}

static CU*
parsecu(Bin *b, u32 cuoff, char **pp)
{
	u8 sz;
	u32 ul, off;
	u16 v;
	Abbrevtab *tab;
	char *p, *q, *o;
	CU *cu;

	o = p = *pp;
	XRD(&ul, p, sizeof(u32));
	if(ul >= 0xffffff00){
		xlog("parsecu: skipping unsupported 64-bit dwarf format");
		return 0;
	}
	q = p+ul;
	XRD(&v, p, sizeof(u16));
	XRD(&off, p, sizeof(u32));
	XRD(&sz, p, sizeof(u8));
	if(!((sz == 4 && b->mode == M32) || (sz == 8 && b->mode == M64))){
		xlog("parsecu: inconsistent address sizes");
		return 0;
	}
	tab = parseabbrev(b->dw.abbrev+off);
	cu = mkcu();
	cu->off = cuoff;
	cu->p = o;
	cu->sz = sz;
	cu->d = readdies(b, cu, &p, tab);
	if(0 > cusrcinfo(cu)){
		xlog("parsecu: compilation unit is missing source information");
		return 0;
	}
	freeabbrevtab(tab);
	if(p != q)
		bug();
	*pp = q;
	return cu;
}

static char*
alignup(char *p, uintptr_t n)
{
	uintptr_t x;
	x = (uintptr_t)p;
	x = (x+n-1)&~(n-1);
	return (char*)x;
}

// return in OP offset of compilation unit in B containing ADDR
int
lookarange(Bin *b, u64 addr, u32 *op)
{
	u8 sz, ss;
	u32 ul, off;
	u32 a32, s32;
	u64 a64, s64;
	u16 v;
	char *p, *q, *e;

	if(b->dw.aranges == 0)
		return -1;
	p = b->dw.aranges;
	e = p+b->dw.naranges;
	while(p < e){
		XRD(&ul, p, sizeof(u32));
		if(ul >= 0xffffff00){
			xlog("doaranges: unsupported 64-bit dwarf format");
			return -1;
		}
		q = p+ul;
		if(q > e){
			xlog("dwarf: bad arange section");
			return -1;
		}
		XRD(&v, p, sizeof(u16));
		XRD(&off, p, sizeof(u32));
		XRD(&sz, p, sizeof(u8));
		XRD(&ss, p, sizeof(u8));

		// align to tuple-size boundary relative to start of aranges
		p -= (uintptr_t)b->dw.aranges;
		p = alignup(p, 2*sz);
		p += (uintptr_t)b->dw.aranges;

		if(sz == 4)
			while(p < q){
				XRD(&a32, p, sizeof(u32));
				XRD(&s32, p, sizeof(u32));
				if(a32 == 0 && s32 == 0)
					break;
				if(a32 <= addr && addr < a32+s32){
					*op = off;
					return 0;
				}
			}
		else if(sz == 8)
			while(p < q){
				XRD(&a64, p, sizeof(u64));
				XRD(&s64, p, sizeof(u64));
				if(a64 == 0 && s64 == 0)
					break;
				if(a64 <= addr && addr < a64+s64){
					*op = off;
					return 0;
				}
			}
		else{
			xlog("dwarf: bad arange section");
			return -1;
		}
		p = q;
	}
	return -1;
}

static Lexpr*
mkle()
{
	Lexpr *le;
	le = xmalloc(sizeof(Lexpr));
	return le;
}

static Lexpr*
mkleulit(u64 u)
{
	Lexpr *le;
	le = mkle();
	le->lkind = Lulit;
	le->u = u;
	return le;
}

static Lexpr*
mkleslit(s64 s)
{
	Lexpr *le;
	le = mkle();
	le->lkind = Lslit;
	le->s = s;
	return le;
}

static Lexpr*
mklereg(u32 no)
{
	Lexpr *le;
	le = mkle();
	le->lkind = Lreg;
	le->no = no;
	return le;
}

static Lexpr*
mklederef(Lexpr *o)
{
	Lexpr *le;
	le = mkle();
	le->lkind = Lderef;
	le->op1 = o;
	return le;
}

static Lexpr*
mklebin(Lkind op, Lexpr *op1, Lexpr *op2)
{
	Lexpr *le;
	le = mkle();
	le->lkind = op;
	le->op1 = op1;
	le->op2 = op2;
	return le;
}

static void
freele(Lexpr *le)
{
	if(le == 0)
		return;
	freele(le->op1);
	freele(le->op2);
	free(le);
}

static Lexpr*
copyle(Lexpr *le)
{
	Lexpr *nle;
	if(le == 0)
		return 0;
	nle = mkle();
	*nle = *le;
	nle->op1 = copyle(le->op1);
	nle->op2 = copyle(le->op2);
	return nle;
}

static Ls*
mkls()
{
	Ls *ls;
	ls = xmalloc(sizeof(Ls));
	ls->max = 128;
	ls->l = xmalloc(ls->max*sizeof(L));
	return ls;
}

static L*
nextl(Ls *ls)
{
	if(ls->n >= ls->max){
		ls->l = xrealloc(ls->l, ls->max*sizeof(L),
				 2*ls->max*sizeof(L));
		ls->max *= 2;
	}
	return &ls->l[ls->n++];
}

static void
freel(L *l)
{
	freele(l->le);
}

void
freels(Ls *ls)
{
	u32 i;
	for(i = 0; i < ls->n; i++)
		freel(&ls->l[i]);
	free(ls->l);
	free(ls);
}

static Locs*
mklocs()
{
	Locs *ls;
	ls = xmalloc(sizeof(Locs));
	ls->max = 128;
	ls->loc = xmalloc(ls->max*sizeof(Loc));
	return ls;
}

static Loc*
nextloc(Locs *ls)
{
	if(ls->n >= ls->max){
		ls->loc = xrealloc(ls->loc, ls->max*sizeof(Loc),
				   2*ls->max*sizeof(Loc));
		ls->max *= 2;
	}
	return &ls->loc[ls->n++];
}

static void
freeloc(Loc *l)
{
	freetype(l->type);
	freels(l->ls);
	free(l->id);
}

void
freelocs(Locs *ls)
{
	u32 i;
	for(i = 0; i < ls->n; i++)
		freeloc(&ls->loc[i]);
	free(ls->loc);
	free(ls);
}

static Die*
lookdie(Die *d, int (*pred)(Die *d, void *a), void *a)
{
	Die *rv;
	if(d == 0)
		return 0;
	if(pred(d, a))
		return d;
	rv = lookdie(d->kid, pred, a);
	if(rv)
		return rv;
	rv = lookdie(d->sib, pred, a);
	return rv;
}

static int
lookfnpred(Die *d, void *a)
{
	u64 pc, lo, hi;
	unsigned i;
	char *id;
	Attr *at;

	if(d->tag != DW_TAG_subprogram)
		return 0;
	pc = *(u64*)a;
	lo = hi = (u64)-1;
	for(i = 0; i < d->nattr; i++){
		if(d->attr[i].name == DW_AT_low_pc)
			lo = d->attr[i].x64;
		if(d->attr[i].name == DW_AT_high_pc)
			hi = d->attr[i].x64;
		if(d->attr[i].name == DW_AT_inline)
			/* FIXME: we assume this is inlined,
			   not what we are looking for. */
			return 0;
	}
	if(lo == (u64)-1 || hi == (u64)-1){
		for(i = 0; i < d->nattr; i++)
			if(d->attr[i].name == DW_AT_ranges){
				xlog("found DW_TAG_subprogram with ranges");
				return 0;
			}
		at = lookattrwithspec(d, DW_AT_name);
		if(at)
			id = at->str;
		else
			id = "UNKNOWN";
		// FIXME: handle DW_AT_ranges case (Dwarf v3 page 47)
		xlog("found DW_TAG_subprogram without lo/hi pcs: %s", id);
		return 0;
	}
	if(lo <= pc && pc < hi)
		return 1;
	return 0;
}

static Die*
lookfn(Die *d, u64 pc)
{
	return lookdie(d, lookfnpred, &pc);
}

static L*
lookl(Ls *ls, u64 a)
{
	u32 i;
	L *l;
	for(i = 0; i < ls->n; i++){
		l = &ls->l[i];
		if(l->lo <= a && a < l->hi)
			return l;
	}
	return 0;
}

static Lexpr*
dwseval(Bin *b, char *p, u64 n, Lexpr *fb)
{
	Lexpr *stack[Maxstack], **sp, *le;
	u8 op;
	char *ep;
	u8 x8;
	u16 x16;
	u32 x32;
	u64 x64;
	s8 sx8;
	s16 sx16;
	s32 sx32;
	s64 sx64;

	sp = stack+Maxstack-1;
	*sp = 0;
	ep = p+n;
	while(p < ep){
		if(sp < stack)
			xfatal("dwarf stack overflow");
		op = *p++;
		switch(op){
		case DW_OP_xderef:
		case DW_OP_deref_size:
		case DW_OP_xderef_size:
		case DW_OP_call2:
		case DW_OP_call4:
		case DW_OP_call_ref:
		case DW_OP_shra:
		case DW_OP_piece:
		case DW_OP_bit_piece:
		case DW_OP_push_object_address:
		case DW_OP_form_tls_address:
		case DW_OP_call_frame_cfa:
		case DW_OP_bregx:
			xfatal("dsweval: %s unimplemented", dwarfopname[op]);
			break;
		default:
			xfatal("dwseval: unknown op 0x%x", op);
			break;
		case DW_OP_nop:
			break;
		case DW_OP_addr:
			if(b->mode == M32){
				XRD(&x32, p, sizeof(x32));
				sp--;
				*sp = mkleulit(x32);
			}else{
				XRD(&x64, p, sizeof(x64));
				sp--;
				*sp = mkleulit(x64);
			}
			break;
		case DW_OP_const1u:
			XRD(&x8, p, sizeof(x8));
			sp--;
			*sp = mkleulit(x8);
			break;
		case DW_OP_const1s:
			XRD(&sx8, p, sizeof(sx8));
			sp--;
			*sp = mkleslit(sx8);
			break;
		case DW_OP_const2u:
			XRD(&x16, p, sizeof(x16));
			sp--;
			*sp = mkleulit(x16);
			break;
		case DW_OP_const2s:
			XRD(&sx16, p, sizeof(sx16));
			sp--;
			*sp = mkleslit(sx16);
			break;
		case DW_OP_const4u:
			XRD(&x32, p, sizeof(x32));
			sp--;
			*sp = mkleulit(x32);
			break;
		case DW_OP_const4s:
			XRD(&sx32, p, sizeof(sx32));
			sp--;
			*sp = mkleslit(sx32);
			break;
		case DW_OP_const8u:
			XRD(&x64, p, sizeof(x64));
			sp--;
			*sp = mkleulit(x64);
			break;
		case DW_OP_const8s:
			XRD(&sx64, p, sizeof(sx64));
			sp--;
			*sp = mkleslit(sx64);
			break;
		case DW_OP_constu:
			x64 = uleb(&p);
			sp--;
			*sp = mkleulit(x64);
			break;
		case DW_OP_consts:
			sx64 = sleb32(&p);
			sp--;
			*sp = mkleslit(sx64);
			break;
		case DW_OP_dup:
			sp--;
			*sp = copyle(sp[1]);
			break;
		case DW_OP_drop:
			sp++;
			break;
		case DW_OP_over:
			sp--;
			*sp = copyle(sp[2]);
			break;
		case DW_OP_pick:
			XRD(&x8, p, sizeof(x8));
			sp--;
			*sp = copyle(sp[x8+1]);
			break;
		case DW_OP_swap:
			le = sp[1];
			sp[1] = sp[0];
			sp[0] = le;
			break;
		case DW_OP_rot:
			le = sp[2];
			sp[2] = sp[1];
			sp[1] = sp[0];
			sp[0] = le;
			break;
		case DW_OP_abs:
		case DW_OP_and:
		case DW_OP_div:
		case DW_OP_mod:
		case DW_OP_mul:
		case DW_OP_neg:
		case DW_OP_not:
		case DW_OP_or:
		case DW_OP_shl:
		case DW_OP_shr:
		case DW_OP_xor:
		case DW_OP_eq:
		case DW_OP_ge:
		case DW_OP_gt:
		case DW_OP_le:
		case DW_OP_lt:
		case DW_OP_ne:
		case DW_OP_bra:
		case DW_OP_skip:
			xfatal("dwseval: %s unimplemented", dwarfopname[op]);
			break;
		case DW_OP_minus:
			sp[1] = mklebin(Lsub, sp[0], sp[1]);
			sp++;
			break;
		case DW_OP_plus:
			sp[1] = mklebin(Ladd, sp[0], sp[1]);
			sp++;
			break;
		case DW_OP_plus_uconst:
			x64 = uleb(&p);
			sp[0] = mklebin(Ladd, sp[0], mkleulit(x64));
			break;
		case DW_OP_deref:
			sp[0] = mklederef(sp[0]);
			break;
		case DW_OP_lit0:
		case DW_OP_lit1:
		case DW_OP_lit2:
		case DW_OP_lit3:
		case DW_OP_lit4:
		case DW_OP_lit5:
		case DW_OP_lit6:
		case DW_OP_lit7:
		case DW_OP_lit8:
		case DW_OP_lit9:
		case DW_OP_lit10:
		case DW_OP_lit11:
		case DW_OP_lit12:
		case DW_OP_lit13:
		case DW_OP_lit14:
		case DW_OP_lit15:
		case DW_OP_lit16:
		case DW_OP_lit17:
		case DW_OP_lit18:
		case DW_OP_lit19:
		case DW_OP_lit20:
		case DW_OP_lit21:
		case DW_OP_lit22:
		case DW_OP_lit23:
		case DW_OP_lit24:
		case DW_OP_lit25:
		case DW_OP_lit26:
		case DW_OP_lit27:
		case DW_OP_lit28:
		case DW_OP_lit29:
		case DW_OP_lit30:
		case DW_OP_lit31:
			sp--;
			*sp = mkleulit(op-DW_OP_lit0);
			break;
		case DW_OP_reg0:
		case DW_OP_reg1:
		case DW_OP_reg2:
		case DW_OP_reg3:
		case DW_OP_reg4:
		case DW_OP_reg5:
		case DW_OP_reg6:
		case DW_OP_reg7:
		case DW_OP_reg8:
		case DW_OP_reg9:
		case DW_OP_reg10:
		case DW_OP_reg11:
		case DW_OP_reg12:
		case DW_OP_reg13:
		case DW_OP_reg14:
		case DW_OP_reg15:
		case DW_OP_reg16:
		case DW_OP_reg17:
		case DW_OP_reg18:
		case DW_OP_reg19:
		case DW_OP_reg20:
		case DW_OP_reg21:
		case DW_OP_reg22:
		case DW_OP_reg23:
		case DW_OP_reg24:
		case DW_OP_reg25:
		case DW_OP_reg26:
		case DW_OP_reg27:
		case DW_OP_reg28:
		case DW_OP_reg29:
		case DW_OP_reg30:
		case DW_OP_reg31:
			sp--;
			*sp = mklereg(op-DW_OP_reg0);
			break;
		case DW_OP_regx:
			x64 = uleb(&p);
			sp--;
			*sp = mklereg(x64);
			break;
		case DW_OP_fbreg:
			sx64 = sleb32(&p);
			sp--;
			*sp = mklebin(Ladd, copyle(fb), mkleslit(sx64));
			break;
		case DW_OP_breg0:
		case DW_OP_breg1:
		case DW_OP_breg2:
		case DW_OP_breg3:
		case DW_OP_breg4:
		case DW_OP_breg5:
		case DW_OP_breg6:
		case DW_OP_breg7:
		case DW_OP_breg8:
		case DW_OP_breg9:
		case DW_OP_breg10:
		case DW_OP_breg11:
		case DW_OP_breg12:
		case DW_OP_breg13:
		case DW_OP_breg14:
		case DW_OP_breg15:
		case DW_OP_breg16:
		case DW_OP_breg17:
		case DW_OP_breg18:
		case DW_OP_breg19:
		case DW_OP_breg20:
		case DW_OP_breg21:
		case DW_OP_breg22:
		case DW_OP_breg23:
		case DW_OP_breg24:
		case DW_OP_breg25:
		case DW_OP_breg26:
		case DW_OP_breg27:
		case DW_OP_breg28:
		case DW_OP_breg29:
		case DW_OP_breg30:
		case DW_OP_breg31:
			sx64 = sleb32(&p);
			sp--;
			*sp = mklebin(Ladd,
				      mklereg(op-DW_OP_breg0),
				      mkleslit(sx64));
			break;
		}
	}
	if(sp < stack+Maxstack)
		return sp[0];
	xfatal("dwseval: no value");
	return 0;
}

static void
add1loc(Ls *ls, Bin *b, u64 lo, u64 hi, char *p, u64 n, Lexpr *fb)
{
	L *l;
	l = nextl(ls);
	l->lo = lo;
	l->hi = hi;
	if(p)
		l->le = dwseval(b, p, n, fb);
}

static void
addloc(Ls *ls, Bin *b, u64 lo, u64 hi, char *p, u64 n, Ls *fbs)
{
	L *lfb;
	u64 m;

	if(fbs){
		/* subdivide range against frame base ranges */
		while(lo < hi){
			lfb = lookl(fbs, lo);
			if(lfb == 0)
				xfatal("location range falls outside of "
				       "frame base ranges");
			m = MIN(hi, lfb->hi);
			add1loc(ls, b, lo, m, p, n, lfb->le);
			lo = m;
		}
	}else
		add1loc(ls, b, lo, hi, p, n, 0);
}

static void
addloclist(Ls *ls, Bin *b, CU *cu, u64 locoff, Ls *fbs)
{
	char *p, *ep;
	u64 base;
	u64 l64, h64;
	u32 l32, h32;
	u16 n;
	Attr *at;

	p = b->dw.loc;
	if(p == 0)
		xfatal("no location list section");
	ep = p+b->dw.nloc;
	p += locoff;
	if(p >= ep)
		xfatal("location list is out of bounds");
	at = lookattr(cu->d, DW_AT_low_pc);
	if(at == 0)
		at = lookattr(cu->d, DW_AT_entry_pc);
	if(at == 0)
		xfatal("compilation unit does not define base address");
	base = at->x64; /* compilation unit lowpc, not function/block/etc. */
	switch(cu->sz){
	case 4:
		while(1){
			XRD(&l32, p, sizeof(u32));
			XRD(&h32, p, sizeof(u32));
			if(l32 == 0 && h32 == 0)
				break;
			if(l32 == (u32)0xffffffff){
				base = h32;
				continue;
			}
			XRD(&n, p, sizeof(u16));
			addloc(ls, b, base+l32, base+h32, p, n, fbs);
			p += n;
		}
		break;
	case 8:
		while(1){
			XRD(&l64, p, sizeof(u64));
			XRD(&h64, p, sizeof(u64));
			if(l64 == 0 && h64 == 0)
				break;
			if(l64 == (u64)0xffffffff){
				base = h64;
				continue;
			}
			XRD(&n, p, sizeof(u16));
			addloc(ls, b, base+l64, base+h64, p, n, fbs);
			p += n;
		}
		break;
	default:
		xfatal("compilation unit has unsupported address size %d",
		       cu->sz);
	}
}

CU*
lookcu(Bin *b, u32 cuoff)
{
	CU *cu;

	cu = hget(b->dw.cu, cuoff);
	if(cu == 0)
		bug();
	return cu;
}

static Ls*
drange(Bin *b, CU *cu, Die *d)
{
	Attr *at, *lat, *hat, *rat;
	u64 base;
	u64 l64, h64;
	u32 l32, h32;
	char *p, *ep;
	Ls *rs;

	rs = mkls();
	lat = lookattr(d, DW_AT_low_pc);
	hat = lookattr(d, DW_AT_high_pc);
	if(lat && hat == 0)
		xfatal("drange: entity has low pc but not high pc");
	if(lat && hat){
		addloc(rs, b, lat->x64, hat->x64, 0, 0, 0);
		return rs;
	}
	rat = lookattr(d, DW_AT_ranges);
	if(rat == 0)
		xfatal("cannot determine range");

	/* consume range list */
	p = b->dw.ranges;
	if(p == 0)
		xfatal("no range list section");
	ep = p+b->dw.nloc;
	p += rat->x64;
	if(p >= ep)
		xfatal("range list is out of bounds");
	at = lookattr(cu->d, DW_AT_low_pc);
	if(at == 0)
		at = lookattr(cu->d, DW_AT_entry_pc);
	if(at == 0)
		xfatal("compilation unit does not define base address");
	base = at->x64; /* compilation unit lowpc, not function/block/etc. */
	switch(cu->sz){
	case 4:
		while(1){
			XRD(&l32, p, sizeof(u32));
			XRD(&h32, p, sizeof(u32));
			if(l32 == 0 && h32 == 0)
				break;
			if(l32 == (u32)0xffffffff){
				base = h32;
				continue;
			}
			addloc(rs, b, base+l32, base+h32, 0, 0, 0);
		}
		break;
	case 8:
		while(1){
			XRD(&l64, p, sizeof(u64));
			XRD(&h64, p, sizeof(u64));
			if(l64 == 0 && h64 == 0)
				break;
			if(l64 == (u64)0xffffffff){
				base = h64;
				continue;
			}
			addloc(rs, b, base+l64, base+h64, 0, 0, 0);
		}
		break;
	}
	return rs;
}

/* LAT is DW_AT_location attribute.  Return non-zero if LAT
   is a location list form. */
static int
islocexpr(Attr *lat)
{
	switch(lat->form){
	case DW_FORM_block:
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
		return 1;
	default:
		return 0;
	}
}

/* LAT is DW_AT_location attribute.  Return non-zero if LAT
   is a location list form. */
static int
isloclist(Attr *lat)
{
	switch(lat->form){
	case DW_FORM_data4:
	case DW_FORM_data8:
		return 1;
	default:
		return 0;
	}
}

static void
d0var(Bin *b, CU *cu, Die *d, Locs *locs, Ls *rs, Ls *fbs)
{
	Attr *at, *lat;
	char *id, *p;
	u64 sz;
	Type *t;
	char *err;
	Ls *ls;
	u32 i;
	Loc *nl;

	at = lookattrwithspec(d, DW_AT_name);
	if(at == 0)
		return;
	id = at->str;
	lat = lookattr(d, DW_AT_location);
	if(lat == 0)
		return;
	at = lookattr(d, DW_AT_type);
	if(at == 0)
		return;
	if(0 > typesize(cu->idx, at->x64, &sz))
		xfatal("cannot determine size of type");
	t = die2type(b, cu, at->x64, 0, &err);
	if(t == 0)
		return;
	p = fmtdecl(t, id);
	ls = mkls();
	if(isloclist(lat))
		addloclist(ls, b, cu, lat->x64, fbs);
	else if(islocexpr(lat))
		for(i = 0; i < rs->n; i++)
			addloc(ls, b, rs->l[i].lo, rs->l[i].hi,
			       lat->block, lat->x64, fbs);
	else
		xfatal("unexpected location form %s", dwarfformname[at->form]);
	nl = nextloc(locs);
	nl->ls = ls;
	nl->id = xstrdup(id);
	nl->sz = sz;
	nl->type = t;
	at = lookattr(d, DW_AT_decl_file);
	switch(d->tag){
	case DW_TAG_variable:
		nl->vkind = Vlocal;
		break;
	case DW_TAG_formal_parameter:
		nl->vkind = Vparam;
		break;
	default:
		xfatal("unexpected variable form %s", dwarftagname[d->tag]);
	}
	free(p);
}

static void
d0(Bin *b, CU *cu, Die *d, Locs *locs, Ls *rs, Ls *fbs)
{
	Ls *krs;
	if(d == 0)
		return;
	krs = rs;
	switch(d->tag){
	case DW_TAG_variable:
	case DW_TAG_formal_parameter:
		d0var(b, cu, d, locs, rs, fbs);
		break;
	case DW_TAG_lexical_block:
		krs = drange(b, cu, d);
		break;
	}
	d0(b, cu, d->sib, locs, rs, fbs);
	d0(b, cu, d->kid, locs, krs, fbs);
	if(krs != rs)
		freels(krs);
}

static Locs*
locs(Bin *b, u64 pc, char **err)
{
	Locs *locs;
	Die *d;
	CU *cu;
	u32 i, cuoff;
	Attr *at;
	Ls *rs, *fbs;

	pc -= b->base;
	if(0 > lookarange(b, pc, &cuoff)){
		*err = Enone;
		return 0;
	}
	cu = lookcu(b, cuoff);
	if(cu == 0){
		xlog("cannot read dwarf unit %x for pc %p", cuoff, pc);
		*err = Ecorrupt;
		return 0;
	}
	d = lookfn(cu->d, pc);
	if(d == 0){
		xlog("cannot find subprogram die for pc %p in %x",
		     pc, cuoff);
		*err = Ecorrupt;
		return 0;
	}
	at = lookattr(d, DW_AT_frame_base);
	if(at == 0){
		xlog("cannot find frame base for %pc %p in %x", pc, cuoff);
		*err = Ecorrupt;
		return 0;
	}
	rs = drange(b, cu, d);
	fbs = mkls();
	if(isloclist(at))
		addloclist(fbs, b, cu, at->x64, 0);
	else if(islocexpr(at))
		for(i = 0; i < rs->n; i++)
			addloc(fbs, b, rs->l[i].lo, rs->l[i].hi,
			       at->block, at->x64, 0);
	else
		xfatal("unexpected frame_base attribute form %s",
		       dwarfformname[at->form]);
	locs = mklocs();
	d0(b, cu, d->kid, locs, rs, fbs);
	freels(rs);
	freels(fbs);
	return locs;
}

Locs*
enumloc(Bin *b, u64 pc, char **err)
{
	if(ignoreg){
		*err = Enone;
		return 0;
	}
	return locs(b, pc, err);
}

static int
evalrow(u8 docie, Unwinder *uw, u64 pc, u8 mode, Rule *init, Rule *rule,
	char **err)
{
	char *p, *q;
	u8 op, x1;
	u16 x2;
	u32 x4;
	u64 x8, y8;
	s64 s8;
	char *blk;
	u32 sp, max;
	Rule *stk, *s;
	int rv;
	u64 loc;

	rv = -1;
	stk = 0;
	if(docie){
		p = uw->c;
		q = uw->ec;
	}else{
		p = uw->f;
		q = uw->ef;
		loc = uw->lo;
	}
	while(p < q){
		op = *p&0xc0;
		switch(op){
		case DW_CFA_advance_loc:
			if(docie){
				*err = Ecorrupt;
				goto fail;
			}
			x1 = *p&0x3f;
			p++;
			loc += x1*uw->caf;
//			xlog("\t%s delta %u (%p)", dwarfcfaname[op], x1, loc);
			if(loc > pc)
				goto out;
			break;
		case DW_CFA_offset:
			x1 = *p&0x3f;
			p++;
			x8 = uleb(&p);
			rule[x1].kind = Roff;
			rule[x1].n = x8*uw->daf;
//			xlog("\t%s reg %d %lu", dwarfcfaname[op], x1, x8);
			break;
		case DW_CFA_restore:
			if(docie){
				*err = Ecorrupt;
				goto fail;
			}
			x1 = *p&0x3f;
			p++;
			rule[x8] = init[x8];
//			xlog("\t%s reg %d", dwarfcfaname[op], x1);
			break;
		case DW_CFA_extended:
			op = *p&0x3f;
			p++;
			switch(op){
			case DW_CFA_nop:
//				xlog("\t%s", dwarfcfaname[op]);
				break;
			case DW_CFA_set_loc:
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				if(mode == M32){
					XRD(&x4, p, sizeof(u32));
					loc = x4;
//					xlog("\t%s %x (%p)", dwarfcfaname[op],
//					       x4, loc);
				}else{
					XRD(&x8, p, sizeof(u64));
					loc = x8;
//					xlog("\t%s %lx (%p)",
//					       dwarfcfaname[op], x8, loc);
				}
				if(loc > pc)
					goto out;
				break;
			case DW_CFA_advance_loc1:
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				XRD(&x1, p, sizeof(u8));
				loc += x1*uw->caf;
//				xlog("\t%s %d (%p)", dwarfcfaname[op], x1, loc);
				if(loc > pc)
					goto out;
				break;
			case DW_CFA_advance_loc2:
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				XRD(&x2, p, sizeof(u16));
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				loc += x2*uw->caf;
//				xlog("\t%s %d (%p)", dwarfcfaname[op], x2, loc);
				if(loc > pc)
					goto out;
				break;
			case DW_CFA_advance_loc4:
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				XRD(&x4, p, sizeof(u32));
				loc += x4*uw->caf;
//				xlog("\t%s %d (%p)", dwarfcfaname[op], x4, loc);
				if(loc > pc)
					goto out;
				break;
			case DW_CFA_def_cfa:
				x8 = uleb(&p);
				y8 = uleb(&p);
				rule[CFA].kind = Rcfaro;
				rule[CFA].r = x8;
				rule[CFA].n = y8;
//				xlog("\t%s %lu %lu", dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_def_cfa_register:
				x8 = uleb(&p);
				rule[CFA].kind = Rcfaro;
				rule[CFA].r = x8;
				// keep the old offset
//				xlog("\t%s %lu", dwarfcfaname[op], x8);
				break;
			case DW_CFA_def_cfa_offset:
				x8 = uleb(&p);
				rule[CFA].kind = Rcfaro;
				rule[CFA].n = x8;
				// keep the old register
//				xlog("\t%s %lu", dwarfcfaname[op], x8);
				break;
			case DW_CFA_def_cfa_expression:
				y8 = uleb(&p);
				blk = p;
				p += y8;
				rule[CFA].kind = Rcfae;
				rule[CFA].e = blk;
				rule[CFA].n = y8;
//				xlog("\t%s (%ld-byte block)",
//				     dwarfcfaname[op], y8);
				break;
			case DW_CFA_def_cfa_sf:
				x8 = uleb(&p);
				s8 = sleb32(&p);
				rule[CFA].kind = Rcfaro;
				rule[CFA].r = x8;
				rule[CFA].n = s8*uw->daf;
//				xlog("\t%s %lu %ld", dwarfcfaname[op], x8, s8);
				break;
			case DW_CFA_def_cfa_offset_sf:
				s8 = sleb32(&p);
				rule[CFA].kind = Rcfaro;
				rule[CFA].n = s8*uw->daf;
				// keep the old register
//				xlog("\t%s %ld", dwarfcfaname[op], s8);
				break;
			case DW_CFA_undefined:
				x8 = uleb(&p);
				rule[x8].kind = Rnone;
//				xlog("\t%s %lu", dwarfcfaname[op], x8);
				break;
			case DW_CFA_same_value:
				x8 = uleb(&p);
				rule[x8].kind = Rsame;
//				xlog("\t%s %lu %lu", dwarfcfaname[op], x8);
				break;
			case DW_CFA_offset_extended:
				x8 = uleb(&p);
				y8 = uleb(&p);
				rule[x8].kind = Roff;
				rule[x8].n = y8*uw->daf;
//				xlog("\t%s %lu %lu", dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_cfa_offset_extended_sf:
				x8 = uleb(&p);
				s8 = sleb32(&p);
				rule[x8].kind = Roff;
				rule[x8].n = s8*uw->daf;
//				xlog("\t%s %lu %ld", dwarfcfaname[op], x8, s8);
				break;
			case DW_CFA_val_offset:
				x8 = uleb(&p);
				y8 = uleb(&p);
				rule[x8].kind = Rvoff;
				rule[x8].n = y8*uw->daf;
//				xlog("\t%s %lu %lu", dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_val_offset_sf:
				x8 = uleb(&p);
				s8 = sleb32(&p);
				rule[x8].kind = Rvoff;
				rule[x8].n = s8*uw->daf;
//				xlog("\t%s %lu %ld", dwarfcfaname[op], x8, s8);
				break;
			case DW_CFA_register:
				x8 = uleb(&p);
				y8 = uleb(&p);
				rule[x8].kind = Rreg;
				rule[x8].r = y8;
//				xlog("\t%s %lu %lu", dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_expression:
				x8 = uleb(&p);
				y8 = uleb(&p);
				blk = p;
				p += y8;
				rule[x8].kind = Rexp;
				rule[x8].e = blk;
				rule[x8].n = y8;
//				xlog("\t%s %ld (%ld-byte block)",
//				     dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_val_expression:
				x8 = uleb(&p);
				y8 = uleb(&p);
				blk = p;
				p += y8;
				rule[x8].kind = Rvexp;
				rule[x8].e = blk;
				rule[x8].n = y8;
//				xlog("\t%s %ld (%ld-byte block)",
//				     dwarfcfaname[op], x8, y8);
				break;
			case DW_CFA_restore_extended:
				if(docie){
					*err = Ecorrupt;
					goto fail;
				}
				x8 = uleb(&p);
				rule[x8] = init[x8];
//				xlog("\t%s %lu", dwarfcfaname[op], x8);
				break;

			case DW_CFA_remember_state:
				if(stk == 0){
					max = 8;
					sp = 0;
					stk = xmalloc(max*Nrow*sizeof(Rule));
				}
				if(sp >= max){
					xlog("%s: overflow",
					dwarfcfaname[op]);
					*err = Ebug;
					goto fail;
				}
				s = stk+sp*Nrow;
				memcpy(s, rule, Nrow*sizeof(Rule));
				sp++;
//				xlog("\t%s", dwarfcfaname[op]);
				break;
			case DW_CFA_restore_state:
				if(stk == 0 || sp == 0){
					*err = Ecorrupt;
					goto fail;
				}
				sp--;
				s = stk+sp*Nrow;
				memcpy(rule, s, Nrow*sizeof(Rule));
//				xlog("\t%s", dwarfcfaname[op]);
				break;
			case DW_CFA_lo_user:
			case DW_CFA_high_user:
			case DW_CFA_MIPS_advance_loc8:
			case DW_CFA_GNU_window_save:
			case DW_CFA_GNU_negative_offset_extended:
				//xlog("\t%s unimplemented", dwarfcfaname[op]);
				*err = Eimpl;
				goto fail;
			case DW_CFA_GNU_args_size:
				uleb(&p);
				xlog("ignoring DW_CFA_GNU_args_size");
				break;
			default:
				//xlog("unrecognized dwarf CFA op %u", op);
				*err = Eimpl;
				goto fail;
			}
			break;
		}
	}

	/* if we exhaust the instructions before
	   location reaches pc, the final rule is the
	   desired one */
out:
	rv = 0;
fail:
	if(stk)
		free(stk);
	return rv;
}

static int
readehcie(char *p, char *e, Unwinder *uw, char **err)
{
	char *in, *q, *a, *aug;
	u64 len;
	u32 ptr, ul;
	u8 v, L, P, R, S;
	static char *es = "augmentation unimplemented";

	/* check cache of previously seen CIEs */
	in = p;
	if(uw->cie == in)
		return 0;
	uw->cie = (void*)-1;	/* clear cache */

	XRD(&ul, p, sizeof(u32));
	if(ul >= 0xffffff00){
		xlog("unwind: unsupported 64-bit dwarf format");
		*err = Eimpl;
		return -1;
	}
	q = p+ul;
	if(ul == 0 || q > e){
		xlog("unwind: bad frame section");
		*err = Ecorrupt;
		return -1;
	}
	XRD(&ptr, p, sizeof(u32));
	if(ptr != 0){
		xlog("unwind: not a CIE header");
		*err = Ecorrupt;
		return -1;
	}
	XRD(&v, p, sizeof(u8));
	if(v != 1 && v != 3){
		xlog("unwind: unsupported CIE version %d", v);
		*err = Eimpl;
		return -1;
	}

	aug = p;
	p += strlen(p)+1;  // FIXME (portably bound the scan)
	if(p <= aug || p >= q){
		xlog("unwind: bad frame section");
		*err = Ecorrupt;
		return -1;
	}
	a = aug;
	L = P = R = S = 0;
	uw->enc = uw->z = 0;
	while(*a){
		switch(*a++){
		case 'z':
			uw->z++;
			break;
		case 'e':
			if(*a == 'h'){
				xlog("unwind: 'eh' %s", es);
				*err = Eimpl;
				return -1;
			}else{
				*err = Ecorrupt;
				return -1;
			}
			break;
		// FIXME: do something with these, if necessary
		case 'L':
			L++;
			break;
		case 'P':
			P++;
			break;
		case 'R':
			R++;
			break;
		case 'S':
			S++;
			break;
		default:
			*err = Ecorrupt;
			return -1;
		}
	}
	uw->caf = uleb(&p);
	uw->daf = sleb32(&p);
	uw->rar = uleb(&p);
	uw->enc = 0;
	if(uw->z){
		len = uleb(&p);
		if(R)
			uw->enc = *p&0x7f;
		// ignore 'P' and 'L' encodings
		p += len;
		if(p > q){
			*err = Ecorrupt;
			return -1;
		}
	}
	uw->c = p;
	uw->ec = q;
	uw->cie = in;		/* commit cache */
	return 0;
}

static int
getunwindereh(Bin *b, u64 pc, Unwinder *uw, char **err)
{
	char *o, *p, *q, *e;
	u32 ptr, ul;
	u64 lo, hi, len;
	u64 x64;
	u32 x32;
	u16 x16;
	s64 s64;
	s32 s32;
	s16 s16;

//	xlog("getunwindereh looking for %p in %s (%p)", pc, b->path, b->base);
	o = p = b->dw.ehframe;
	e = p+b->dw.nehframe;
	while(p < e){
		XRD(&ul, p, sizeof(u32));
		if(ul >= 0xffffff00){
			xlog("unwind: unsupported 64-bit dwarf format");
			*err = Eimpl;
			return -1;
		}
		if(ul == 0)
			break;
		q = p+ul;
		if(q > e){
			xlog("unwind: bad frame section");
			*err = Ecorrupt;
			return -1;
		}
		XRD(&ptr, p, sizeof(u32));
		if(ptr == 0){
			/* skip CIE header */
			p = q;
			continue;
		}

		ptr += sizeof(u32);
		if(0 > readehcie(p-ptr, e, uw, err))
			return -1;

		switch(uw->enc&0x70){
		case 0:
			if(b->mode == M32){
				XRD(&x32, p, sizeof(u32));
				lo = x32;
				XRD(&x32, p, sizeof(u32));
				hi = lo+x32;
			}else{
				XRD(&lo, p, sizeof(u64));
				XRD(&hi, p, sizeof(u64));
				hi += lo;
			}
			break;
		case DW_EH_PE_pcrel:
			// see gdb-7.0/gdb/dwarf2-frame.c
			lo = b->dw.vehframe+p-b->dw.ehframe;
			switch(uw->enc&0x0f){
			case DW_EH_PE_uleb128:
				x64 = uleb(&p);
				lo += x64;
				x64 = uleb(&p);
				hi = lo+x64;
			case DW_EH_PE_sleb128:
				s64 = sleb32(&p);
				lo += s64;
				s64 = sleb32(&p);
				hi = lo+s64;
				break;
			case DW_EH_PE_udata2:
				XRD(&x16, p, sizeof(x16));
				lo += x16;
				XRD(&x16, p, sizeof(x16));
				hi = lo+x16;
				break;
			case DW_EH_PE_udata4:
				XRD(&x32, p, sizeof(x32));
				lo += x32;
				XRD(&x32, p, sizeof(x32));
				hi = lo+x32;
				break;
			case DW_EH_PE_udata8:
				XRD(&x64, p, sizeof(x64));
				lo += x64;
				XRD(&x64, p, sizeof(x64));
				hi = lo+x64;
				break;
			case DW_EH_PE_sdata2:
				XRD(&s16, p, sizeof(s16));
				lo += s16;
				XRD(&s16, p, sizeof(s16));
				hi = lo+s16;
				break;
			case DW_EH_PE_sdata4:
				XRD(&s32, p, sizeof(s32));
				lo += s32;
				XRD(&s32, p, sizeof(s32));
				hi = lo+s32;
				break;
			case DW_EH_PE_sdata8:
				XRD(&s64, p, sizeof(s64));
				lo += s64;
				XRD(&s64, p, sizeof(s64));
				hi = lo+s64;
				break;
			case DW_EH_PE_signed:
				xlog("unimplemented eh encoding %x", uw->enc);
				*err = Eimpl;
				return -1;
			default:
				xlog("unrecognized eh encoding %x", uw->enc);
				*err = Ecorrupt;
				return -1;
			}
			break;
		case DW_EH_PE_textrel:
		case DW_EH_PE_datarel:
		case DW_EH_PE_funcrel:
		case DW_EH_PE_aligned:
		case DW_EH_PE_indirect:
			xlog("unimplemented eh encoding %x", uw->enc);
			*err = Eimpl;
			return -1;
		default:
			xlog("unrecognized eh encoding %x", uw->enc);
			*err = Ecorrupt;
			return -1;
		}

//		xlog("look for %p in %p - %p of %s", pc, lo, hi, b->path);
		if(pc < lo || pc >= hi){
			/* skip this FDE */
			p = q;
			continue;
		}
//		xlog("found %p in %p - %p", pc, lo, hi);

		if(uw->z){
			len = uleb(&p);
			if(len != 0){
				xlog("i still don't understand aug!");
				*err = Eimpl;
				return -1;
			}
		}

		/* this is our FDE */
		uw->lo = lo;
		uw->hi = hi;
		uw->f = p;
		uw->ef = q;
		return 0;
	}
	*err = Enone;
	return -1;
}

// dwarf spec
static int
getunwinderdw(Bin *b, u64 pc, Unwinder *uw, char **err)
{
	char *o, *p, *q, *e;
	char *aug;
	u32 ptr, x32, ul;
	u64 lo, hi;
	u8 v;

	o = p = b->dw.frame;
	e = p+b->dw.nframe;
	while(p < e){
		XRD(&ul, p, sizeof(u32));
		if(ul >= 0xffffff00){
			xlog("unwind: unsupported 64-bit dwarf format");
			*err = Eimpl;
			return -1;
		}
		if(ul == 0)
			break;
		q = p+ul;
		if(q > e){
			xlog("unwind: bad frame section");
			*err = Ecorrupt;
			return -1;
		}
		XRD(&ptr, p, sizeof(u32));
		if(ptr == 0xffffffff){
			/* skip CIE header */
//			xlog("skipping cie %p %ld", ptr, ul);
			p = alignup(q, b->mode == M32 ? 4 : 8);
			continue;
		}
//		xlog("ul %lx ptr %lx", ul, ptr);
		if(b->mode == M32){
			XRD(&x32, p, sizeof(u32));
			lo = x32;
			XRD(&x32, p, sizeof(u32));
			hi = lo+x32;
		}else{
			XRD(&lo, p, sizeof(u64));
			XRD(&hi, p, sizeof(u64));
			hi += lo;
		}
//		xlog("getunwinderdw is %p in %p - %p", pc, lo, hi);
		if(pc < lo || pc >= hi){
			/* skip this FDE */
			p = alignup(q, b->mode == M32 ? 4 : 8);
			continue;
		}
//		xlog("getunwinderdw found fd %p - %p", lo, hi);

		/* this is our FDE */
		uw->lo = lo;
		uw->hi = hi;
		uw->f = p;
		uw->ef = q;

		/* read the associated CIE */
		p = o+ptr;
		if(p < o || p >= e){
			*err = Ecorrupt;
			return -1;
		}
		XRD(&ul, p, sizeof(u32));
		if(ul >= 0xffffff00){
			xlog("unwind: unsupported 64-bit dwarf format");
			*err = Eimpl;
			return -1;
		}
		q = p+ul;
		if(q <= p || q > e){
			xlog("unwind: bad frame section");
			*err = Ecorrupt;
			return -1;
		}
		XRD(&ptr, p, sizeof(u32));
		if(ptr != 0xffffffff){
			xlog("unwind: bad CIE");
			*err = Ecorrupt;
			return -1;
		}
		XRD(&v, p, sizeof(u8));
		if(v != 1 && v != 3){
			xlog("unwind: unsupported CIE version %d", v);
			*err = Eimpl;
			return -1;
		}
		aug = p;
		p += strlen(p)+1;  // FIXME (portably bound the scan)
		if(p <= aug || p >= q){
			xlog("unwind: bad frame section");
			*err = Ecorrupt;
			return -1;
		}
		if(aug[0]){
			xlog("unwind: unsupported augmentation %s", aug);
			*err = Eimpl;
			return -1;
		}
		uw->cie = (void*)-1;
		uw->enc = uw->z = 0;
		uw->caf = uleb(&p);
		uw->daf = sleb32(&p);
		uw->rar = uleb(&p);
		uw->c = p;
		uw->ec = q;
		return 0;
	}
	*err = Enone;
	return -1;
}

static int
getunwinder(Bin *b, u64 pc, Unwinder *uw, char **err)
{
	int rv;
	rv = -1;
	if(b->dw.frame){
		rv = getunwinderdw(b, pc, uw, err);
		if(rv == 0 || *err != Enone)
			return rv;
	}
	if(b->dw.ehframe){
		rv = getunwindereh(b, pc, uw, err);
		if(rv == 0 || *err != Enone)
			return rv;
	}
	*err = Enone;
	return -1;
}

int
unwind1(Bin *b, u64 pc, Rule *rule, char **err)
{
	Rule init[Nrow];
	Unwinder uw;

	if(ignoreg)
		goto nohelp;
	memset(&uw, 0, sizeof(uw));
	*err = 0;
	if(!b->dw.frame && !b->dw.ehframe)
		goto nohelp;
	pc -= b->base;
	if(0 > getunwinder(b, pc, &uw, err))
		goto nohelp;

	/* initialize rules to Rnone */
	memset(init, 0, sizeof(init));

	/* evaluate initial cie rules */
	if(0 > evalrow(1, &uw, 0, b->mode, 0, init, err))
		goto fail;

	/* evaluate fde rules for this pc */
	memcpy(rule, init, sizeof(init));
	if(0 > evalrow(0, &uw, pc, b->mode, init, rule, err))
		goto fail;

	return 0;
nohelp:
	*err = Enone;
fail:
	return -1;
}

static void
parseallcu(Bin *b)
{
	char *p, *ep;
	CU *cu;
	u32 off;

	p = b->dw.info;
	ep = p+b->dw.ninfo;
	while(p < ep){
		off = p-b->dw.info;
		cu = parsecu(b, off, &p);
		if(cu == 0)
			continue;
		hput(b->dw.cu, off, cu);
	}
}

static Type*
die2anontyperec(Bin *b, CU *cu, u32 off, char **err)
{
	Type *t, *s;
	Die *d, *k;
	Attr *at;
	u32 i;
	Decl *m;

	t = 0;
	d = hget(cu->idx, off);
	if(d == 0)
		bug();
	switch(d->tag){
	case DW_TAG_base_type:
		t = mktype();
		t->kind = Tbase;
		if(0 > die2base(b, d, t))
			goto bad;
		return t;
	case DW_TAG_enumeration_type:
	case DW_TAG_class_type:
	case DW_TAG_structure_type:
	case DW_TAG_union_type:
		t = mktype();
		switch(d->tag){
		case DW_TAG_enumeration_type:
			t->kind = Tenum;
			break;
		case DW_TAG_class_type:
		case DW_TAG_structure_type:
			t->kind = Tstruct;
			break;
		case DW_TAG_union_type:
			t->kind = Tunion;
			break;
		}
		if(d->id){
			t->id = xstrdup(d->id);
			return t;
		}
		at = lookattrwithspec(d, DW_AT_name);
		if(at){
			t->id = xstrdup(cidentifier(at->str));
			return t;
		}
		goto bad;
	case DW_TAG_array_type:
		s = t = mktype();
		t->kind = Tarr;
		if(d->kid){
			k = d->kid;
			// FIXME: here we expect subrange_type children.
			// what is the meaning of DW_AT_type attribute of these
			// children?
			// spec also allows enumeration_type children.
			// what do these mean?
			while(k){
				if(k->tag != DW_TAG_subrange_type)
					goto bad;
				// FIXME: what about DW_AT_count?
				at = lookattr(k, DW_AT_upper_bound);
				if(at)
					// assume 0-based index
					s->n = at->x64+1;
				else
					s->n = 0;
				k = k->sib;
				if(k){
					s->link = mktype();
					s = s->link;
					s->kind = Tarr;
				}
			}
		}else
			s->n = 0;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			s->link = mkvoid();
		else
			s->link = die2anontyperec(b, cu, (u32)at->x64, err);
		if(s->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_pointer_type:
	case DW_TAG_reference_type:
		t = mktype();
		t->kind = Tptr;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2anontyperec(b, cu, (u32)at->x64, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_subprogram:
	case DW_TAG_subroutine_type:
		t = mktype();
		t->kind = Tfun;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2anontyperec(b, cu, (u32)at->x64, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag == DW_TAG_formal_parameter)
				t->n++;
			k = k->sib;
		}
		if(t->n == 0){
			/* like dwarf2cqct, say that params are "void" */
			t->n = 1;
			t->ds = xmalloc(t->n*sizeof(Decl*));
			m = t->ds[0] = mkdecl();
			m->type = mkvoid();
			return t;
		}
		k = d->kid;
		t->ds = xmalloc(t->n*sizeof(Decl*));
		i = 0;
		while(k){
			if(k->tag != DW_TAG_formal_parameter){
				k = k->sib;
				continue;
			}
			m = t->ds[i++] = mkdecl();
			at = lookattrwithspec(k, DW_AT_name);
			if(at)
				m->id = xstrdup(cidentifier(at->str));
			at = lookattr(k, DW_AT_type);
			if(at == 0)
				m->type = mkvoid();
			else
				m->type = die2anontyperec(b, cu, (u32)at->x64,
							  err);
			if(m->type == 0){
				freetype(t);
				return 0;
			}
			k = k->sib;
		}
		return t;
	case DW_TAG_typedef:
		t = mktype();
		t->kind = Ttypedef;
		at = lookattrwithspec(d, DW_AT_name);
		if(at == 0)
			goto bad;
		t->id = xstrdup(cidentifier(at->str));
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2anontyperec(b, cu, (u32)at->x64, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_const_type:
	case DW_TAG_packed_type:
	case DW_TAG_volatile_type:
	case DW_TAG_restrict_type:
	case DW_TAG_shared_type:
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			return mkvoid();
		else
			return die2anontyperec(b, cu, (u32)at->x64, err);
	case DW_TAG_string_type:
	case DW_TAG_file_type:
	case DW_TAG_thrown_type:
	case DW_TAG_unspecified_type:
	case DW_TAG_upc_shared_type:
	case DW_TAG_upc_strict_type:
	case DW_TAG_upc_relaxed_type:
	case DW_TAG_ptr_to_member_type:
	case DW_TAG_set_type:
		xfatal("die2anontyperec: unimplemented type %s",
		       dwarftagname[d->tag]);
	default:
		xfatal("die2anontyperec: unexpected type %s",
		       dwarftagname[d->tag]);
	}
bad:
	*err = Ebin;
	freetype(t);
	return 0;
}

/* FIXME: unify with die2type */
static Type*
die2anontype(Bin *b, Diet *et, Die *d, char **err)
{
	Type *t, *s;
	Die *k;
	Attr *at;
	u32 i;
	Decl *m;
	Enum *e;

	t = 0;
	switch(d->tag){
	case DW_TAG_enumeration_type:
		t = mktype();
		t->kind = Tenum;
		t->id = 0;  /* supplied by unifyanon if needed */
		at = lookattr(d, DW_AT_byte_size);
		if(at == 0)
			goto bad;
		switch(at->x64){
		case 1:
			if(b->enc == Mle)
				t->rep = Rs08le;
			else
				t->rep = Rs08be;
			break;
		case 4:
			if(b->enc == Mle)
				t->rep = Rs32le;
			else
				t->rep = Rs32be;
			break;
		case 8:
			if(b->enc == Mle)
				t->rep = Rs64le;
			else
				t->rep = Rs64be;
			break;
		default:
			xlog("unexpected enum size: %u for %s", at->x64, t->id);
			goto bad;
		}
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag != DW_TAG_enumerator){
//				xlog("unexpected %s in enum",
//				     dwarftagname[k->tag]);
				k = k->sib;
				continue;
			}
			t->n++;
			k = k->sib;
		}
		k = d->kid;
		t->en = xmalloc(t->n*sizeof(Enum*));
		for(i = 0; i < t->n; i++){
			e = t->en[i] = mkenumconst();
			e->type = t;
			at = lookattrwithspec(k, DW_AT_name);
			if(at == 0)
				goto bad;
			e->id = xstrdup(cidentifier(at->str));
			at = lookattr(k, DW_AT_const_value);
			if(at == 0){
//				goto bad;
				printf("warning: no const_value for %s in anonymous enum\n",
				       e->id);
				e->val = 0;
			}else
				e->val = at->sx64;
			k = k->sib;
		}
		return t;
	case DW_TAG_class_type:
	case DW_TAG_structure_type:
	case DW_TAG_union_type:
		t = mktype();
		switch(d->tag){
		case DW_TAG_class_type:
		case DW_TAG_structure_type:
			t->kind = Tstruct;
			break;
		case DW_TAG_union_type:
			t->kind = Tunion;
			break;
		}
		t->id = 0;  /* supplied by unifyanon if needed */
		at = lookattr(d, DW_AT_byte_size);
		if(at == 0)
			// might be an undefined type; treat as empty
			t->sz = 0;
		else
			t->sz = at->x64;
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag != DW_TAG_member){
				if(k->tag == DW_TAG_class_type
				   || k->tag == DW_TAG_structure_type
				   || k->tag == DW_TAG_union_type){
					k = k->sib;
					continue;
				}
//				xlog("unexpected %s in anonymous su",
//				     dwarftagname[k->tag]);
				k = k->sib;
				continue;
			}
			t->n++;
			k = k->sib;
		}
		k = d->kid;
		t->ds = xmalloc(t->n*sizeof(Decl*));
		for(i = 0; i < t->n; i++){
			while(k->tag != DW_TAG_member)
				k = k->sib;
			m = t->ds[i] = mkdecl();
			at = lookattrwithspec(k, DW_AT_name);
			if(at)
				m->id = xstrdup(cidentifier(at->str));
			else
				m->id = mkanonid(et->cuoff, et->dieoff);
			at = lookattr(k, DW_AT_type);
			if(at == 0)
				m->type = mkvoid();
			else
				m->type = die2anontyperec(b, d->cu,
							  (u32)at->x64,
							  err);
			if(m->type == 0){
				freetype(t);
				return 0;
			}
			if(d->tag == DW_TAG_union_type)
				m->off = 0;
			else{
				at = lookattr(k, DW_AT_data_member_location);
				if(at == 0)
					goto bad;
				m->off = at2u64(b, at);
			}
			at = lookattr(k, DW_AT_bit_size);
			if(at){
				/* this is a bitfield */
				s = mktype();
				s->kind = Tbitfield;
				s->link = m->type;
				m->type = s;
				s->w = at->x64;
				at = lookattr(k, DW_AT_bit_offset);
				if(at == 0)
					goto bad;
				s->pos = at->x64;
				if(b->enc == Mle){
					/* math inherited from dwarf2cqct */
					at = lookattr(k, DW_AT_byte_size);
					if(at == 0)
						goto bad;
					s->pos = 8*at->x64-s->pos-s->w;
				}
			}
			k = k->sib;
		}
		return t;
	default:
		xfatal("die2anontype: unexpected tag %s", dwarftagname[d->tag]);
	}

bad:
	*err = Ebin;
	freetype(t);
	return 0;
}

static int
anoneqsub(Type *t1, Type *t2)
{
	u32 i;

	if(t1->kind != t2->kind)
		return 0;
	switch(t1->kind){
	case Tbase:
		return t1->base == t2->base;
	case Tstruct:
	case Tunion:
	case Tenum:
	case Ttypedef:
		/* these must have ids */
		return !strcmp(t1->id, t2->id);
	case Tptr:
		return anoneqsub(t1->link, t2->link);
	case Tarr:
		return t1->n == t2->n && anoneqsub(t1->link, t2->link);
	case Tfun:
		if(!anoneqsub(t1->link, t2->link))
			return 0;
		for(i = 0; i < t1->n; i++){
			if((t1->id && !t2->id) || (!t1->id && t2->id))
				return 0;
			if(t1->id && strcmp(t1->id, t2->id))
				return 0;
			if(!anoneqsub(t1->link, t2->link))
				return 0;
		}
		return 1;
	case Tbitfield:
		if(t1->pos != t2->pos || t1->w != t2->w)
			return 0;
		return anoneqsub(t1->link, t2->link);
	default:
		bug();
		return 0;
	}
}

static int
anoneq(void *v1, void *v2)
{
	Type *t1, *t2;
	Decl *d1, *d2;
	Enum *e1, *e2;
	u32 i;

	t1 = v1;
	t2 = v2;

	if(t1->kind != t2->kind)
		return 0;
	switch(t1->kind){
	case Tstruct:
	case Tunion:
		if(t1->n != t2->n || t1->sz != t2->sz)
			return 0;
		for(i = 0; i < t1->n; i++){
			d1 = t1->ds[i];
			d2 = t2->ds[i];
			if(d1->off != d2->off)
				return 0;
			/* anonymous ids in equivalent types may disagree */
			if(strncmp(d1->id, ANONID, sizeof(ANONID)-1)
			   && strncmp(d2->id, ANONID, sizeof(ANONID)-1)
			   && strcmp(d1->id, d2->id))
				return 0;
			if(!anoneqsub(d1->type, d2->type))
				return 0;
		}
		return 1;
	case Tenum:
		if(t1->n != t2->n)
			return 0;
		for(i = 0; i < t1->n; i++){
			e1 = t1->en[i];
			e2 = t2->en[i];
			if(e1->val != e2->val || strcmp(e1->id, e2->id))
				return 0;
		}
		return 1;
	default:
		bug();
		return 0;
	}
}

static void
add1diet(void *u, uptr k, void *v)
{
	adddiet(v, u);
}

static HT*
unifyanon(Bin *b, char **err)
{
	Diets *ants, *tmp;
	Diet *e;
	Die *d;
	CU *cu;
	Alist *s;
	HT *x;
	u32 i;
	Type *t;
	char *id;
	u32 idn;

	ants = mkdiets();
	hforeach(b->dw.ant, add1diet, ants);
	s = mkalist(anoneq);
	x = mksht();
	idn = 0;

	while(ants->n){
		tmp = mkdiets();
		for(i = 0; i < ants->n; i++){
			e = ants->diet[i];
			cu = lookcu(b, e->cuoff);
			if(cu == 0){
				*err = Ebin;
				goto fail;
			}
			d = hget(cu->idx, e->dieoff);
			if(d == 0){
				*err = Ebin;
				goto fail;
			}
			t = die2anontype(b, e, d, err);
			if(t == 0){
				adddiet(e, tmp);
				continue;
			}
			id = lookalist(t, s);
			if(id == 0){
				id = mkanon(idn++);
				t->id = xstrdup(id);
				insalist(t, id, s);
				hput(x, (uptr)id, e);
			}else{
				freetype(t);
				freediet(e);
			}
			d->id = xstrdup(id);
		}

		freediets(ants);
		ants = tmp;
	}

	freediets(ants);
	for(i = 0; i < s->n; i++)
		freetype(s->k[i]);
	freealist(s);

	return x;
fail:
	freediets(ants);
	freealist(s);
	return 0;
}


/* FIXME: here, if nowhere else, it may be dangerous
    for us to be assuming the encodings of the
    integer-valued attributes.  dwarf2cqct does a
    better job (via evalblock). */
static Type*
die2type(Bin *b, CU *cu, u32 off, u32 complete, char **err)
{
	Type *t, *s;
	Die *d, *k;
	Attr *at;
	u32 i;
	Decl *m;
	Enum *e;

	t = 0;
	d = hget(cu->idx, off);
	if(d == 0)
		bug();
	switch(d->tag){
	case DW_TAG_base_type:
		t = mktype();
		t->kind = Tbase;
		if(0 > die2base(b, d, t))
			goto bad;
		return t;
	case DW_TAG_enumeration_type:
		t = mktype();
		t->kind = Tenum;
		at = lookattrwithspec(d, DW_AT_name);
		if(at)
			t->id = xstrdup(cidentifier(at->str));
		else
			t->id = xstrdup(d->id);
		if(!complete)
			return t;
		at = lookattr(d, DW_AT_byte_size);
		if(at == 0)
			goto bad;
		switch(at->x64){
		case 1:
			if(b->enc == Mle)
				t->rep = Rs08le;
			else
				t->rep = Rs08be;
			break;
		case 4:
			if(b->enc == Mle)
				t->rep = Rs32le;
			else
				t->rep = Rs32be;
			break;
		case 8:
			if(b->enc == Mle)
				t->rep = Rs64le;
			else
				t->rep = Rs64be;
			break;
		default:
			xlog("unexpected enum size: %u for %s", at->x64, t->id);
			goto bad;
		}
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag != DW_TAG_enumerator){
//				xlog("unexpected %s in enum",
//				     dwarftagname[k->tag]);
				k = k->sib;
				continue;
			}
			t->n++;
			k = k->sib;
		}
		k = d->kid;
		t->en = xmalloc(t->n*sizeof(Enum*));
		for(i = 0; i < t->n; i++){
			e = t->en[i] = mkenumconst();
			e->type = t;
			at = lookattrwithspec(k, DW_AT_name);
			if(at == 0)
				goto bad;
			e->id = xstrdup(cidentifier(at->str));
			at = lookattr(k, DW_AT_const_value);
			if(at == 0){
//				goto bad;
				printf("warning: no const_value for %s in enum %s\n",
				       e->id, t->id);
				e->val = 0;
			}else
				e->val = at->sx64;
			k = k->sib;
		}
		return t;
	case DW_TAG_class_type:
	case DW_TAG_structure_type:
	case DW_TAG_union_type:
		t = mktype();
		switch(d->tag){
		case DW_TAG_class_type:
		case DW_TAG_structure_type:
			t->kind = Tstruct;
			break;
		case DW_TAG_union_type:
			t->kind = Tunion;
			break;
		}
		at = lookattrwithspec(d, DW_AT_name);
		if(at)
			t->id = xstrdup(cidentifier(at->str));
		else
			t->id = xstrdup(d->id);
		if(!complete)
			return t;
		at = lookattr(d, DW_AT_byte_size);
		if(at == 0)
			// might be an undefined type; treat as empty
			t->sz = 0;
		else
			t->sz = at->x64;
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag != DW_TAG_member){
				if(k->tag == DW_TAG_class_type
				   || k->tag == DW_TAG_structure_type
				   || k->tag == DW_TAG_union_type){
					k = k->sib;
					continue;
				}
//				xlog("unexpected %s in su %s",
//				     dwarftagname[k->tag], t->id);
				k = k->sib;
				continue;
			}
			t->n++;
			k = k->sib;
		}
		k = d->kid;
		t->ds = xmalloc(t->n*sizeof(Decl*));
		for(i = 0; i < t->n; i++){
			while(k->tag != DW_TAG_member)
				k = k->sib;
			m = t->ds[i] = mkdecl();
			at = lookattrwithspec(k, DW_AT_name);
			if(at)
				m->id = xstrdup(cidentifier(at->str));
			else
				m->id = mkanonid(cu->off, off);
			at = lookattr(k, DW_AT_type);
			if(at == 0)
				m->type = mkvoid();
			else
				m->type = die2type(b, cu, (u32)at->x64, 0, err);
			if(m->type == 0){
				freetype(t);
				return 0;
			}
			if(d->tag == DW_TAG_union_type)
				m->off = 0;
			else{
				at = lookattr(k, DW_AT_data_member_location);
				if(at == 0)
					goto bad;
				m->off = at2u64(b, at);
			}
			at = lookattr(k, DW_AT_bit_size);
			if(at){
				/* this is a bitfield */
				s = mktype();
				s->kind = Tbitfield;
				s->link = m->type;
				m->type = s;
				s->w = at->x64;
				at = lookattr(k, DW_AT_bit_offset);
				if(at == 0)
					goto bad;
				s->pos = at->x64;
				if(b->enc == Mle){
					/* math inherited from dwarf2cqct */
					at = lookattr(k, DW_AT_byte_size);
					if(at == 0)
						goto bad;
					s->pos = 8*at->x64-s->pos-s->w;
				}
			}
			k = k->sib;
		}
		return t;
	case DW_TAG_array_type:
		s = t = mktype();
		t->kind = Tarr;
		if(d->kid){
			k = d->kid;
			// FIXME: here we expect subrange_type children.
			// what is the meaning of DW_AT_type attribute of these
			// children?
			// spec also allows enumeration_type children.
			// what do these mean?
			while(k){
				if(k->tag != DW_TAG_subrange_type)
					goto bad;
				// FIXME: what about DW_AT_count?
				at = lookattr(k, DW_AT_upper_bound);
				if(at)
					// assume 0-based index
					s->n = at->x64+1;
				else
					s->n = 0;
				k = k->sib;
				if(k){
					s->link = mktype();
					s = s->link;
					s->kind = Tarr;
				}
			}
		}else
			s->n = 0;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			s->link = mkvoid();
		else
			s->link = die2type(b, cu, (u32)at->x64, 0, err);
		if(s->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_pointer_type:
	case DW_TAG_reference_type:
		t = mktype();
		t->kind = Tptr;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2type(b, cu, (u32)at->x64, 0, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_subprogram:
	case DW_TAG_subroutine_type:
		t = mktype();
		t->kind = Tfun;
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2type(b, cu, (u32)at->x64, 0, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		t->n = 0;
		k = d->kid;
		while(k){
			if(k->tag == DW_TAG_formal_parameter)
				t->n++;
			k = k->sib;
		}
		if(t->n == 0){
			/* like dwarf2cqct, say that params are "void" */
			t->n = 1;
			t->ds = xmalloc(t->n*sizeof(Decl*));
			m = t->ds[0] = mkdecl();
			m->type = mkvoid();
			return t;
		}
		k = d->kid;
		t->ds = xmalloc(t->n*sizeof(Decl*));
		i = 0;
		while(k){
			if(k->tag != DW_TAG_formal_parameter){
				k = k->sib;
				continue;
			}
			m = t->ds[i++] = mkdecl();
			at = lookattrwithspec(k, DW_AT_name);
			if(at)
				m->id = xstrdup(cidentifier(at->str));
			at = lookattr(k, DW_AT_type);
			if(at == 0)
				m->type = mkvoid();
			else
				m->type = die2type(b, cu, (u32)at->x64, 0, err);
			if(m->type == 0){
				freetype(t);
				return 0;
			}
			k = k->sib;
		}
		return t;
	case DW_TAG_typedef:
		t = mktype();
		t->kind = Ttypedef;
		at = lookattrwithspec(d, DW_AT_name);
		if(at == 0)
			goto bad;
		t->id = xstrdup(cidentifier(at->str));
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			t->link = mkvoid();
		else
			t->link = die2type(b, cu, (u32)at->x64, 0, err);
		if(t->link == 0){
			freetype(t);
			return 0;
		}
		return t;
	case DW_TAG_const_type:
	case DW_TAG_packed_type:
	case DW_TAG_volatile_type:
	case DW_TAG_restrict_type:
	case DW_TAG_shared_type:
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			return mkvoid();
		else
			return die2type(b, cu, (u32)at->x64, complete, err);
	case DW_TAG_string_type:
	case DW_TAG_file_type:
	case DW_TAG_thrown_type:
	case DW_TAG_unspecified_type:
	case DW_TAG_upc_shared_type:
	case DW_TAG_upc_strict_type:
	case DW_TAG_upc_relaxed_type:
	case DW_TAG_ptr_to_member_type:
	case DW_TAG_set_type:
		xfatal("die2type: unimplemented type %s", dwarftagname[d->tag]);
	default:
		xfatal("die2type: unexpected type %s", dwarftagname[d->tag]);
	}
bad:
	*err = Ebin;
	freetype(t);
	return 0;
}

Type*
dwlooktype(Bin *b, Type *t, char **err)
{
	Type *rv;
	Diet *e;
	CU *cu;

	rv = 0;
	switch(t->kind){
	case Tbase:
		rv = mktype();
		rv->kind = Tbase;
		rv->base = t->base;
		rv->rep = b->dw.base[t->base].rep;
		break;
	case Tstruct:
	case Tunion:
	case Tenum:
		e = hget(b->dw.sue, (uptr)t->id);
		if(e == 0)
			e = hget(b->dw.ant, (uptr)t->id);
		if(e == 0){
			*err = Enone;
			return 0;
		}
		cu = lookcu(b, e->cuoff);
		if(cu == 0){
			*err = Ebin;
			return 0;
		}
		rv = die2type(b, cu, e->dieoff, 1, err);
		break;
	case Ttypedef:
		rv = hget(b->dw.xt, (uptr)t->id);
		if(rv){
			rv = copytype(rv);
			break;
		}
		e = hget(b->dw.typ, (uptr)t->id);
		if(e == 0){
			*err = Enone;
			return 0;
		}
		cu = lookcu(b, e->cuoff);
		if(cu == 0){
			*err = Ebin;
			return 0;
		}
		rv = die2type(b, cu, e->dieoff, 1, err);
		break;
	default:
		*err = Eproto;
		break;
	}
	return rv;
}

typedef
struct ET
{
	Bin *b;
	Types *ts;
	char **err;
} ET;

static void
do1xt(void *u, uptr id, void *v)
{
	ET *et;
	et = u;
	addtype(et->ts, copytype(v));
}

static void
do1diet(void *u, uptr id, void *v)
{
	ET *et;
	Type *t;
	Diet *e;
	CU *cu;

	et = u;
	e = v;
	cu = lookcu(et->b, e->cuoff);
	if(cu == 0){
		*et->err = Ebin;
		return;
	}
	t = die2type(et->b, cu, e->dieoff, 1, et->err);
	if(t == 0){
		*et->err = Ebin;
		return;
	}
	addtype(et->ts, t);
}

Types*
dwenumtype(Bin *b, char **err)
{
	Cbase cb;
	ET et;
	Types *ts;
	Type *t;

	ts = mktypes();

	// base types
	for(cb = Vlo; cb < Vnallbase; cb++){
		t = mktype();
		t->kind = Tbase;
		t->base = b->dw.base[cb].base;
		t->rep = b->dw.base[cb].rep;
		addtype(ts, t);
	}

	// tagged and typedef types
	et.b = b;
	et.ts = ts;
	et.err = err;
	hforeach(b->dw.xt, do1xt, &et);
	hforeach(b->dw.sue, do1diet, &et);
	hforeach(b->dw.ant, do1diet, &et);
	hforeach(b->dw.typ, do1diet, &et);

	return ts;
}

static void
basefill(Bin *b, Rootns *r)
{
	Cbase cb;
	Xtype xt;
	Type *t;

	for(cb = Vlo; cb <= Vnbase; cb++)
		if(b->dw.base[cb].rep == Rundef)
			b->dw.base[cb].rep = r->base[cb];
	for(xt = Xlo; xt < Xn; xt++)
		if(!hget(b->dw.typ, (uptr)xtypename[xt])){
			t = mktype();
			t->kind = Ttypedef;
			t->id = xstrdup(xtypename[xt]);
			t->link = copytype(&b->dw.base[r->xtype[xt]]);
			hput(b->dw.xt, (uptr)xtypename[xt], t);
		}
}

/* match a rootns to base types defined in dwarf and
   define any omitted types */
static void
basetypes(Bin *b)
{
	u32 i;
	Cbase cb;
	Rootns *r;
	unsigned ptrguess;

	ptrguess = b->dw.base[Vptr].rep == Rundef;

	for(i = 0; i < Rnr; i++){
		r = &rootns[i];
		for(cb = Vlo; cb <= Vnbase; cb++)
			if(r->base[cb] != Rundef && b->dw.base[cb].rep != Rundef
			   && r->base[cb] != b->dw.base[cb].rep)
				break;
		if(cb > Vnbase){
			basefill(b, r);
			if(ptrguess)
				xlog("warning: guessed pointer size in %s: %s",
				     b->path,
				     r->name);
			return;
		}
	}
	xlog("warning: cannot match rootns to base types of %s", b->path);
}

static void
do1istag(void *u, uptr xid, void *v)
{
	Bin *b;
	Diet *e;
	CU *cu;
	Type *t, *w, *x;
	Die *d;
	char *id, *err;
	u32 i;
	Sym *s;

	b = u;
	e = v;
	id = (char*)xid;
	cu = lookcu(b, e->cuoff);
	if(cu == 0)
		bug();
	d = hget(cu->idx, e->dieoff);
	if(d == 0)
		bug();
	if(d->tag != DW_TAG_enumeration_type)
		return;
	t = die2type(b, cu, e->dieoff, 1, &err);
	if(t == 0)
		bug();
	for(i = 0; i < t->n; i++){
		if(hget(b->sym, (uptr)t->en[i]->id))
			continue;
		w = mktype();
		w->kind = Tenum;
		w->id = xstrdup(t->id);
		x = mktype();
		x->kind = Tconst;
		x->link = w;
		s = mksym();
		s->id = xstrdup(t->en[i]->id);
		s->val = t->en[i]->val;
		s->flags |= Senum|Sro;
		s->type = x;
		hput(b->sym, (uptr)s->id, s);
	}
	freetype(t);
}

static void
do1issym(void *u, uptr xid, void *v)
{
	Bin *b;
	Diet *e;
	CU *cu;
	Sym *s;
	Attr *at;
	Die *d;
	char *id, *err;

	b = u;
	e = v;
	id = (char*)xid;

	if(hget(b->sym, (uptr)id))
		return;

	s = mksym();
	s->id = xstrdup(id);
	cu = lookcu(b, e->cuoff);
	if(cu == 0)
		bug();
	d = hget(cu->idx, e->dieoff);
	if(d == 0)
		bug();
	if(d->tag == DW_TAG_subprogram){
		s->type = die2type(b, cu, e->dieoff, 0, &err);
		at = lookattr(d, DW_AT_inline);
		if(at && (at->x64 == DW_INL_inlined
			  || at->x64 == DW_INL_declared_inlined))
			s->flags |= Sinline;
	}else{
		at = lookattr(d, DW_AT_type);
		if(at == 0)
			s->type = mkvoid();
		else
			s->type = die2type(b, cu, (u32)at->x64, 0, &err);
	}
	if(s->type == 0)
		xfatal("do1issym: die2type on %s: %s", id, err);
	hput(b->sym, (uptr)s->id, s);
}

static void
dwidxsym(Bin *b)
{
	/* index enum constants */
	hforeach(b->dw.sue, do1istag, b);
	hforeach(b->dw.ant, do1istag, b);

	/* index symbols */
	hforeach(b->dw.sym, do1issym, b);
}

static void
free1str(void *u, uptr k, void *v)
{
	free((char*)k);
}

static void
free1cu(void *u, uptr k, void *v)
{
	freecu(v);
}

static void
free1diet(void *u, uptr k, void *v)
{
	free((char*)k);
	freediet(v);
}

static void
free1antdiet(void *u, uptr k, void *v)
{
	free((char*)k);
	freediet(v);
}

static void
free1type(void *u, uptr k, void *v)
{
	freetype(v);
}

void
initdwarf(Bin *b)
{
	Cbase cb;
	char *err;
	HT *ant;

	b->dw.cu = mkht();
	b->dw.sue = mksht();
	b->dw.ant = mksht();
	b->dw.sym = mksht();
	b->dw.typ = mksht();
	b->dw.xt = mksht();
	for(cb = 0; cb < Vnallbase; cb++){
		b->dw.base[cb].kind = Tbase;
		b->dw.base[cb].base = cb;
		b->dw.base[cb].rep = Rundef;
	}
	parseallcu(b);
	basetypes(b);
	err = 0;
	ant = unifyanon(b, &err);
	if(ant == 0)
		xlog("unifyanon: %s", err);
	else{
		hforeach(b->dw.ant, free1str, 0);
		freeht(b->dw.ant);
		b->dw.ant = ant;
	}
	dwidxsym(b);
}

void
freedwarf(Bin *b)
{
	if(b->dw.cu == 0)
		return;
	hforeach(b->dw.cu, free1cu, 0);
	freeht(b->dw.cu);
	hforeach(b->dw.sue, free1diet, 0);
	freeht(b->dw.sue);
	hforeach(b->dw.ant, free1antdiet, 0);
	freeht(b->dw.ant);
	hforeach(b->dw.typ, free1diet, 0);
	freeht(b->dw.typ);
	hforeach(b->dw.sym, free1diet, 0);
	freeht(b->dw.sym);
	hforeach(b->dw.xt, free1type, 0);
	freeht(b->dw.xt);
}

void
dwarfinit()
{
}

void
dwarffini()
{
}
