#include "../inc/sctl.h"
#include "../inc/err.h"
#include "dwarf.h"
#include "dw.h"
#include "unix.h"

/* line number state machine */
typedef
struct SM
{
	u64 addr;
	u32 file;
	u32 line;
	u32 col;
	u8  stmt;
	u8  bb;
	u8  eseq;
	u8  epr;
	u8  bep;
	u8  isa;
} SM;

static Src*
mksrc(char **file, u32 nf)
{
	Src *src;
	src = xmalloc(sizeof(Src));
	src->file = file;
	src->nfile = nf;
	src->max = 256;
	src->loc = xmalloc(src->max*sizeof(Srcloc));
	src->n = 0;
	return src;
}

void
freesrc(Src *src)
{
	if(src == 0)
		return;
	free(src->file);
	free(src->loc);
	free(src);
}

static void
row(Src *src, SM *sm)
{
	Srcloc *l;
	if(src->n >= src->max){
		src->loc = xrealloc(src->loc,
				    src->max*sizeof(Srcloc),
				    2*src->max*sizeof(Srcloc));
		src->max *= 2;
	}
	l = &src->loc[src->n++];
	l->addr = sm->addr;
	l->file = src->file[sm->file];
	l->line = sm->line;
	l->col = sm->col;
}

static char*
strnchr(char *p, int c, char *e)
{
	while(p < e){
		if(*p == c)
			return p;
		p++;
	}
	return 0;
}

static void
reset(SM *sm, u8 dis)
{
	memset(sm, 0, sizeof(SM));
	sm->file = 1;
	sm->line = 1;
	sm->stmt = dis;
}

/* internalize line number data for CU */
static int
dwsrc(Bin *b, CU *cu, char **err)
{
	char *buf, *f, *l, *o, *p, *q, *e;
	u64 u, x64;
	s64 s;
	u32 i, m, ul, hl, nd, nf, il, x32;
	u16 v, fa;
	u8 mil, dis, lr, ob; 
	s8 lb;
	u8 *nrand;
	u64 di, lm, fl;
	u8 op;
	SM sm;
	char **dir, **file;
	Src *src;

	dir = file = 0;
	src = 0;

	if(b->dw.line == 0)
		bug();
	if(cu->soff == -1){
		xlog("srcinfo is corrupt");
		goto corrupt;
	}

	p = b->dw.line;
	e = p+b->dw.nline;
	p += cu->soff;
	if(p >= e)
		goto corrupt;

	/* read header */
	XRD(&ul, p, sizeof(u32));
	if(ul >= 0xffffff00){
		xlog("dwsrc: unsupported 64-bit dwarf format");
		return -1;
	}
	q = p+ul;
	if(q > e)
		goto corrupt;
	XRD(&v, p, sizeof(u16));
	XRD(&hl, p, sizeof(u32));
	o = p+hl;
	XRD(&mil, p, sizeof(u8));
	XRD(&dis, p, sizeof(u8));
	XRD(&lb, p, sizeof(s8));
	XRD(&lr, p, sizeof(u8));
	XRD(&ob, p, sizeof(u8));
	nrand = xmalloc(ob*sizeof(u8));
	for(i = 1; i < ob; i++)
		XRD(nrand+i, p, sizeof(u8));

	/* include directories */
	l = p;
	nd = 0;
	while(*p){
		nd++;
		p = strnchr(p, '\0', q);
		if(p == 0)
			goto corrupt;
		p++;
	}
	dir = xmalloc((nd+1)*sizeof(char*));
	nd = 0;
	p = l;
	dir[0] = xstrdup(cu->dir);
	while(*p){
		dir[++nd] = xstrdup(p);
		p = strnchr(p, '\0', q);
		if(p == 0)
			goto corrupt;
		p++;
	}
	p++;
	
	/* file names */
	l = p;
	nf = 0;
	while(*p){
		nf++;
		p = strnchr(p, '\0', q);
		if(p == 0)
			goto corrupt;
		p++;
		di = uleb(&p);
		lm = uleb(&p);
		fl = uleb(&p);
	}
	file = xmalloc((nf+1)*sizeof(char*));
	nf = 0;
	p = l;
	while(*p){
		f = p;
		p = strnchr(p, '\0', q);
		if(p == 0)
			goto corrupt;
		p++;
		di = uleb(&p);
		lm = uleb(&p);
		fl = uleb(&p);
		if(f[0] == '/')
			file[++nf] = xstrdup(f);
		else{
			m = strlen(dir[di])+1+strlen(f)+1;
			buf = xmalloc(m);
			snprintf(buf, m, "%s/%s", dir[di], f);
			file[++nf] = buf;
		}
	}
	p++;

	src = mksrc(file, nf);
	file = 0; /* now owned by src */

	reset(&sm, dis);
	while(o < q){
		op = *o++;
		if(op >= ob){
			/* special opcode */
			op -= ob;
			sm.addr += (op/lr)*mil;
			sm.line += lb+op%lr;
			row(src, &sm);
			sm.bb = sm.epr = sm.bep = 0;
		}else if(op == 0){
			/* extended opcodes */
			il = uleb(&o);
			op = *o++;
//			xlog("extended: %s", dwarflnsname[op]);
			switch(op){
			case DW_LNE_end_sequence:
				sm.eseq = 1;
				row(src, &sm);
				reset(&sm, dis);
				break;
			case DW_LNE_set_address:
				if(b->mode == M32){
					XRD(&x32, o, sizeof(u32));
					sm.addr = x32;
				}else{
					XRD(&x64, o, sizeof(u64));
					sm.addr = x64;
				}
				break;
			case DW_LNE_define_file:
				xlog("DW_LNE_define_file is unimplemented");
				goto unimpl;
				break;
			default:
				xlog("unrecognized extended opcode: %u", *o);
				goto unimpl;
			}
		}else{
			switch(op){
			case DW_LNS_copy:
				row(src, &sm);
				sm.bb = sm.epr = sm.bep = 0;
				break;
			case DW_LNS_advance_pc:
				u = uleb(&o);
				sm.addr += u*mil;
				break;
			case DW_LNS_advance_line:
				s = sleb32(&o);
				sm.line += s;
				break;
			case DW_LNS_set_file:
				u = uleb(&o);
				sm.file = u;
				break;
			case DW_LNS_set_column:
				u = uleb(&o);
				sm.col = u;
				break;
			case DW_LNS_negate_stmt:
				sm.stmt = !sm.stmt;
				break;
			case DW_LNS_set_basic_block:
				sm.bb = 1;
				break;
			case DW_LNS_const_add_pc:
				sm.addr += ((255-ob)/lr)*mil;
				break;
			case DW_LNS_fixed_advance_pc:
				XRD(&fa, o, sizeof(u16));
				sm.addr += fa;
				break;
			case DW_LNS_set_prologue_end:
				sm.epr = 1;
				break;
			case DW_LNS_set_epilogue_begin:
				sm.bep = 1;
				break;
			case DW_LNS_set_isa:
				u = uleb(&o);
				sm.isa = u;
				break;
			default:
				xlog("unrecognized standard opcode: %u", *o);
				goto unimpl;
			}
		}
	}

	free(dir);
	cu->src = src;
	return 0;
unimpl:
	xlog("source info for compile unit %p of %s has unimplemented feature",
	     cu->off, b->path);
	*err = Enone;
	goto lose;
corrupt:
	xlog("source info for compile unit %p of %s appears corrupt",
	     cu->off, b->path);
	*err = Enone;
	goto lose;
lose:
	free(dir);
	free(file);
	freesrc(src);
	return -1;
}

Srcloc*
looksrc(Bin *b, u64 pc, char **err)
{
	CU *cu;
	u32 i, cuoff;
	Src *src;
	Srcloc *l;

	if(b->dw.line == 0)
		goto lose;
	pc -= b->base;
	if(0 > lookarange(b, pc, &cuoff))
		goto lose;
	cu = lookcu(b, cuoff);
	if(cu == 0){
		xlog("cannot read dwarf unit %x for pc %p", cuoff, pc);
		*err = Ecorrupt;
		return 0;
	}
	if(cu->src == 0)
		if(0 > dwsrc(b, cu, err))
			return 0;
	src = cu->src;
	if(src->n == 0)
		goto lose;
	l = &src->loc[0];
	if(pc < l->addr)
		goto lose;
	for(i = 1; i < src->n; i++){
		if(src->loc[i].addr > pc)
			break;
		l = &src->loc[i];
	}
	return l;
lose:
	*err = Enone;
	return 0;
}

typedef
struct LL
{
	char *file;
	CU *cu;
} LL;

static void
find1cu(void *u, uptr k, void *v)
{
	LL *ll;
	CU *cu;
	u32 m;
	char *buf;

	cu = v;
	ll = u;
	if(!strcmp(ll->file, cu->file)){
		ll->cu = cu;
		return;
	}
	m = strlen(cu->dir)+1+strlen(cu->file)+1;
	buf = xmalloc(m);
	snprintf(buf, m, "%s/%s", cu->dir, cu->file);
	if(!strcmp(ll->file, buf))
		ll->cu = cu;
	free(buf);
}

int
lookpc(Bin *b, char *file, u32 line, u64 *addr, char **err)
{
	LL ll;
	CU *cu;
	u32 i;
	Src *src;
	Srcloc *l;

	ll.file = file;
	ll.cu = 0;
	hforeach(b->dw.cu, find1cu, &ll);
	if(ll.cu == 0)
		goto lose;
	cu = ll.cu;
	if(cu->src == 0)
		if(0 > dwsrc(b, cu, err))
			return -1;
	src = cu->src;
	for(i = 0; i < src->n; i++){
		l = &src->loc[i];
		if(l->line == line){
			*addr = b->base+l->addr;
			return 0;
		}
	}
lose:	   
	*err = Enone;
	return -1;
}

void
bufpushsrc(Buf *x, Srcloc *l)
{
	bufpushs(x, l->file);
	bufpushu32(x, l->line);
	bufpushu32(x, l->col);
}
