#include "../inc/sctl.h"
#include "../inc/err.h"
#include "../inc/elf.h"
#include "dwarf.h"
#include "dw.h"
#include "unix.h"
#include "ns.h"

static HT *bintab;
static u64 nextid;

static Syms*
mksyms()
{
	Syms *ss;
	ss = xmalloc(sizeof(Syms));
	ss->max = 1024;
	ss->sym = xmalloc(ss->max*sizeof(Sym*));
	return ss;
}

static void
addsym(Syms *ss, Sym *s)
{
	if(ss->n >= ss->max){
		ss->sym = xrealloc(ss->sym, ss->max*sizeof(Sym*),
				   2*ss->max*sizeof(Sym*));
		ss->max *= 2;
	}
	ss->sym[ss->n++] = s;
}

void
freesyms(Syms *ss)
{
	free(ss->sym);
	free(ss);
}

static Segs*
mksegs()
{
	Segs *gs;
	gs = xmalloc(sizeof(Segs));
	gs->max = 1024;
	gs->seg = xmalloc(gs->max*sizeof(Seg));
	return gs;
}

static Seg*
nextseg(Segs *gs)
{
	if(gs->n >= gs->max){
		gs->seg = xrealloc(gs->seg, gs->max*sizeof(Seg),
				   2*gs->max*sizeof(Seg));
		gs->max *= 2;
	}
	return &gs->seg[gs->n++];
}

static void
freesegs(Segs *gs)
{
	u32 i;
	for(i = 0; i < gs->n; i++)
		free(gs->seg[i].name);
	free(gs->seg);
	free(gs);
}


static char*
mapfile(char *file, u64 *len, char **err)
{
	int fd;
	struct stat st;
	char *p;

	fd = open(file, O_RDONLY);
	if(0 > fd){
		*err = Enone;
		return 0;
	}
	if(0 > fstat(fd, &st)){
		close(fd);
		return 0;
	}
	p = mmap(0, st.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
	close(fd);
	if(p == MAP_FAILED)
		return 0;
	*len = st.st_size;
	return p;
}

char*
cidentifier(char *id)
{
	char *p, *q, c;
	static char s[4096]; /* sic */

	p = id;
	q = s;
	c = id[0];
	if(!isalpha(c) && c != '_')
		*q++ = '_';
	while((c = *p)){
		if(!isalpha(c) && !isdigit(c) && c != '_')
			*q = '_';
		else
			*q = c;
		p++;
		q++;
		if(q >= s+sizeof(s))
			bug();	/* we're static to avoid mallocs */
	}
	*q = 0;
	return s;
}

/* 
   mangled names begin _Z<N><SYM>...
   where N is an integer >= 0 in ascii decimal and SYM is an N-byte
   identifier.  we don't care about the rest.

   we assume a name beginning _Z is mangled because C and C++
   reserve all identifiers beginning with _ and a captial letter.

   see http://sourcery.mentor.com/public/cxx-abi/abi.html
   and cp-demangle.c in the GNU liberty library
*/
static char*
demangle(char *id)
{
	char *p, *e;
	u32 m;
	if(strncmp(id, "_Z", 2) != 0)
		return id;
	p = id;
	e = p+strlen(id);
	p += 2; /* _Z */
	if(p >= e)
		return id;
	if(!isdigit(*p))
		/* malformed */
		return id;
	m = 0;
	while(p < e && isdigit(*p)){
		m *= 10;
		m += *p-'0';
		p++;
	}
	if(p >= e)
		/* malformed */
		return id;
	if(e-p < m)
		return id;
	memmove(id, p, m);
	id[m] = 0;
	return id;
}

Sym*
mksym()
{
	Sym *s;
	s = xmalloc(sizeof(Sym));
	return s;
}

void
freesym(Sym *s)
{
	free(s->id);
	freetype(s->type);
	free(s);
}

static char*
interp32(char *p)
{
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;
	unsigned i;

	ehdr = (Elf32_Ehdr*)p;
	phdr = (Elf32_Phdr*)(p+ehdr->e_phoff);
	for(i = 0; i < ehdr->e_phnum; i++){
		if(phdr[i].p_type != PT_INTERP)
			continue;
		return p+phdr[i].p_offset;
	}
	return 0;
}

static char*
interp64(char *p)
{
	Elf64_Ehdr *ehdr;
	Elf64_Phdr *phdr;
	unsigned i;

	ehdr = (Elf64_Ehdr*)p;
	phdr = (Elf64_Phdr*)(p+ehdr->e_phoff);
	for(i = 0; i < ehdr->e_phnum; i++){
		if(phdr[i].p_type != PT_INTERP)
			continue;
		return p+phdr[i].p_offset;
	}
	return 0;
}

char*
interp(char *exe, u8 *m, char **err)
{
	char *p, *ip;
	u64 len;

	ip = 0;
	p = mapfile(exe, &len, err);
	if(p == 0)
		return 0;
	if(strncmp(p, ELFMAG, SELFMAG)){
		*err = Ebin;
		goto out;
	}
	switch(p[EI_CLASS]){
	case ELFCLASS32:
		ip = interp32(p);
		*m = M32;
		break;
	case ELFCLASS64:
		ip = interp64(p);
		*m = M64;
		break;
	default:
		xlog("%s: unknown ELF class %d", exe, p[EI_CLASS]);
		*err = Ebin;
		break;
	}
	if(ip)
		ip = xstrdup(ip);
out:
	munmap(p, len);
	return ip;
}

static u64
loadaddr32(char *p)
{
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;
	unsigned i;

	ehdr = (Elf32_Ehdr*)p;
	phdr = (Elf32_Phdr*)(p+ehdr->e_phoff);
	for(i = 0; i < ehdr->e_phnum; i++){
		if(phdr[i].p_type != PT_LOAD)
			continue;
		return (u64)(uptr)(phdr[i].p_vaddr);
	}
	return 0;
}


static u64
loadaddr64(char *p)
{
	Elf64_Ehdr *ehdr;
	Elf64_Phdr *phdr;
	unsigned i;

	ehdr = (Elf64_Ehdr*)p;
	phdr = (Elf64_Phdr*)(p+ehdr->e_phoff);
	for(i = 0; i < ehdr->e_phnum; i++){
		if(phdr[i].p_type != PT_LOAD)
			continue;
		return (u64)(uptr)(phdr[i].p_vaddr);
	}
	return 0;
}

u64
loadaddr(char *exe, char **err)
{
	char *p, *ip;
	u64 la, len;

	ip = 0;
	p = mapfile(exe, &len, err);
	if(p == 0)
		return 0;
	if(strncmp(p, ELFMAG, SELFMAG)){
		*err = Ebin;
		goto out;
	}
	switch(p[EI_CLASS]){
	case ELFCLASS32:
		la = loadaddr32(p);
		break;
	case ELFCLASS64:
		la = loadaddr64(p);
		break;
	default:
		xlog("%s: unknown ELF class %d", exe, p[EI_CLASS]);
		*err = Ebin;
		break;
	}
out:
	munmap(p, len);
	return la;

}

typedef
struct Dwreloc
{
	char *name;
	u32 secoff; /* offset within Dwarf structure of section pointer */
	u32 lenoff; /* offset within Dwarf structure of section length */
} Dwreloc;

/* Yikes.  These turn Bin B and Dwreloc DR into the
   values of the corresponding Dwarf section pointer
   and length. */
#define BSEC(b, dr) (*(char**)((char*)&(b)->dw+(dr)->secoff))
#define BOFF(b, dr) (*(u32*)((char*)&(b)->dw+(dr)->lenoff))

static Dwreloc dwreloc[] = {
	{ ".rela.debug_info",
	  offsetof(Dwarf, info), offsetof(Dwarf, ninfo) },
	{ ".rela.debug_abbrev",
	  offsetof(Dwarf, abbrev), offsetof(Dwarf, nabbrev) },
	{ ".rela.debug_aranges",
	  offsetof(Dwarf, aranges), offsetof(Dwarf, naranges) },
	{ ".rela.debug_frame",
	  offsetof(Dwarf, frame), offsetof(Dwarf, nframe) },
	{ ".rela.eh_frame",
	  offsetof(Dwarf, ehframe), offsetof(Dwarf, nehframe) },
	{ ".rela.debug_line",
	  offsetof(Dwarf, line), offsetof(Dwarf, nline) },
	{ ".rela.debug_loc",
	  offsetof(Dwarf, loc), offsetof(Dwarf, nloc) },
	{ ".rela.debug_pubnames",
	  offsetof(Dwarf, pubnames), offsetof(Dwarf, npubnames) },
	{ ".rela.debug_ranges",
	  offsetof(Dwarf, ranges), offsetof(Dwarf, nranges) },
	{ ".rela.debug_str",
	  offsetof(Dwarf, str), offsetof(Dwarf, nstr) },
	{ 0, 0, 0 }
};

static int
onereloc32(char *p, u32 len, void *q, u32 na, char **err)
{
	char *ep;
	u32 i;
	Elf32_Rela *a;
	u32 *xu32;

	ep = p+len;
	for(i = 0, a = q; i < na; i++, a++)
		switch(ELF32_R_TYPE(a->r_info)){
		case R_386_32:
		case R_386_PC32: /* FIXME: is this right? */
			if(p+a->r_offset+sizeof(u32) > ep)
				xfatal("relocation out of bounds");
			xu32 = (u32*)(p+a->r_offset);
			*xu32 += (u32)a->r_addend;
			break;
		default:
			xfatal("binary has unimplemented relocation form: %d\n",
			       ELF32_R_TYPE(a->r_info));
			break;
		}
	return 0;
}

static int
onereloc64(char *p, u32 len, void *q, u32 na, char **err)
{
	char *ep;
	u32 i;
	Elf64_Rela *a;
	u32 *xu32;
	u64 *xu64;

	ep = p+len;
	for(i = 0, a = q; i < na; i++, a++)
		switch(ELF64_R_TYPE(a->r_info)){
		case R_X86_64_32S:
			if(p+a->r_offset+sizeof(u32) > ep)
				xfatal("relocation out of bounds");
			xu32 = (u32*)(p+a->r_offset);
			*xu32 += (s32)a->r_addend;
			break;
		case R_X86_64_32:
		case R_X86_64_PC32: /* FIXME: is this right? */
			if(p+a->r_offset+sizeof(u32) > ep)
				xfatal("relocation out of bounds");
			xu32 = (u32*)(p+a->r_offset);
			*xu32 += (u32)a->r_addend;
			break;
		case R_X86_64_64:
			if(p+a->r_offset+sizeof(u64) > ep)
				xfatal("relocation out of bounds");
			xu64 = (u64*)(p+a->r_offset);
			*xu64 += (u64)a->r_addend;
			break;
		default:
			xfatal("binary has unimplemented relocation form: %d\n",
			       ELF64_R_TYPE(a->r_info));
			break;
		}
	return 0;
}

static int
reloc32(Bin *b, char **err)
{
	Elf32_Ehdr *ehdr;
	Elf32_Shdr *shdr, *s;
	char *p, *str;
	unsigned i;
	Dwreloc *dr;

	p = b->b;
	ehdr = (Elf32_Ehdr*)p;
	shdr = (Elf32_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_type != SHT_RELA)
			continue;
		dr = dwreloc;
		while(dr->name){
			if(!strcmp(dr->name, str+s->sh_name) && BSEC(b, dr))
				if(0 > onereloc32(BSEC(b, dr), BOFF(b, dr),
						  b->b+s->sh_offset,
						  s->sh_size/s->sh_entsize,
						  err))
					return -1;
			dr++;
		}
	}
	return 0;
}

static int
reloc64(Bin *b, char **err)
{
	Elf64_Ehdr *ehdr;
	Elf64_Shdr *shdr, *s;
	char *p, *str;
	unsigned i;
	Dwreloc *dr;

	p = b->b;
	ehdr = (Elf64_Ehdr*)p;
	shdr = (Elf64_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_type != SHT_RELA)
			continue;
		dr = dwreloc;
		while(dr->name){
			if(!strcmp(dr->name, str+s->sh_name) && BSEC(b, dr))
				if(0 > onereloc64(BSEC(b, dr), BOFF(b, dr),
						  b->b+s->sh_offset,
						  s->sh_size/s->sh_entsize,
						  err))
					return -1;
			dr++;
		}
	}
	return 0;
}

static u64
lookrelplt32(Bin *b, u32 ndx)
{
	Elf32_Rel *r;
	Elf32_Rela *a;
	unsigned i;

	if(b->relplt)
		for(i = 0, r = (Elf32_Rel*)b->relplt; i < b->nrel; i++, r++)
			if(ELF32_R_SYM(r->r_info) == ndx)
				return r->r_offset;
	if(b->relaplt)
		for(i = 0, a = (Elf32_Rela*)b->relaplt; i < b->nrela; i++, a++)
			if(ELF32_R_SYM(a->r_info) == ndx)
				return a->r_offset;
	return 0;
}

static u64
lookrelplt64(Bin *b, u32 ndx)
{
	Elf64_Rel *r;
	Elf64_Rela *a;
	unsigned i;

	if(b->relplt)
		for(i = 0, r = (Elf64_Rel*)b->relplt; i < b->nrel; i++, r++)
			if(ELF64_R_SYM(r->r_info) == ndx)
				return r->r_offset;
	if(b->relaplt)
		for(i = 0, a = (Elf64_Rela*)b->relaplt; i < b->nrela; i++, a++)
			if(ELF64_R_SYM(a->r_info) == ndx)
				return a->r_offset;
	return 0;
}

static void
patchsym32(Bin *b)
{
	Elf32_Sym *s;
	u32 i, a;

	for(i = 0, s = (Elf32_Sym*)b->dynsym; i < b->ndyn; i++, s++){
		if(s->st_value == 0
		   && ELF32_ST_TYPE(s->st_info) == STT_FUNC
		   && s->st_shndx == STN_UNDEF){
			/* set value to offset in PLT */
			a = lookrelplt32(b, i);
			if(a){
				a = *(u32*)(b->got+(a-b->gotaddr));
				a -= 6; // sizeof jmp *<mem>
				s->st_value = a; // relative to load address
			}
		}
	}
}

static void
patchsym64(Bin *b)
{
	Elf64_Sym *s;
	u64 i, a;

	for(i = 0, s = (Elf64_Sym*)b->dynsym; i < b->ndyn; i++, s++){
		if(s->st_value == 0
		   && ELF64_ST_TYPE(s->st_info) == STT_FUNC
		   && s->st_shndx == STN_UNDEF){
			/* set value to offset in PLT */
			a = lookrelplt64(b, i);
			if(a){
				a = *(u64*)(b->got+(a-b->gotaddr));
				a -= 6; // sizeof jmp *<mem>
				s->st_value = a; // relative to load address
			}else
				;
				// xlog("no relocation for %s",
				//     b->dynstr+s->st_name);
		}
	}
}

static int
initsym32(Bin *b, char **err)
{
	Elf32_Ehdr *ehdr;
	Elf32_Shdr *shdr, *s;
	char *p, *str;
	unsigned i;

	p = b->b;
	ehdr = (Elf32_Ehdr*)p;
	shdr = (Elf32_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_type == SHT_SYMTAB){
			b->symtab = p+s->sh_offset;
			b->strtab = p+shdr[s->sh_link].sh_offset;
			b->nsym = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_DYNSYM){
			b->dynsym = p+s->sh_offset;
			b->dynstr = p+shdr[s->sh_link].sh_offset;
			b->ndyn = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_REL
		   && !strcmp(".rel.plt", str+s->sh_name)){
			b->relplt = p+s->sh_offset;
			b->nrel = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_RELA
		   && !strcmp(".rela.plt", str+s->sh_name)){
			b->relaplt = p+s->sh_offset;
			b->nrela = s->sh_size/s->sh_entsize;
		}
		if(!strcmp(".data", str+s->sh_name))
			b->ndxdata = i;
		if(!strcmp(".rodata", str+s->sh_name))
			b->ndxrodata = i;
		if(!strcmp(".bss", str+s->sh_name))
			b->ndxbss = i;
		if(!strcmp(".text", str+s->sh_name)){
			b->ndxtext = i;
			b->vtext = s->sh_addr;
			b->ntext = s->sh_size;
		}
		if(!strcmp(".got", str+s->sh_name)){
			b->got = p+s->sh_offset;
			b->gotaddr = s->sh_addr;
		}
		if(!strcmp(".debug_abbrev", str+s->sh_name)){
			b->dw.abbrev = p+s->sh_offset;
			b->dw.nabbrev = s->sh_size;
		}
		if(!strcmp(".debug_aranges", str+s->sh_name)){
			b->dw.aranges = p+s->sh_offset;
			b->dw.naranges = s->sh_size;
		}
		if(!strcmp(".debug_frame", str+s->sh_name)){
			b->dw.frame = p+s->sh_offset;
			b->dw.nframe = s->sh_size;
		}
		if(!strcmp(".eh_frame", str+s->sh_name)){
			b->dw.ehframe = p+s->sh_offset;
			b->dw.nehframe = s->sh_size;
			b->dw.vehframe = s->sh_addr;
		}
		if(!strcmp(".debug_info", str+s->sh_name)){
			b->dw.info = p+s->sh_offset;
			b->dw.ninfo = s->sh_size;
		}
		if(!strcmp(".debug_line", str+s->sh_name)){
			b->dw.line = p+s->sh_offset;
			b->dw.nline = s->sh_size;
		}
		if(!strcmp(".debug_loc", str+s->sh_name)){
			b->dw.loc = p+s->sh_offset;
			b->dw.nloc = s->sh_size;
		}
		if(!strcmp(".debug_pubnames", str+s->sh_name)){
			b->dw.pubnames = p+s->sh_offset;
			b->dw.npubnames = s->sh_size;
		}
		if(!strcmp(".debug_ranges", str+s->sh_name)){
			b->dw.ranges = p+s->sh_offset;
			b->dw.nranges = s->sh_size;
		}
		if(!strcmp(".debug_str", str+s->sh_name)){
			b->dw.str = p+s->sh_offset;
			b->dw.nstr = s->sh_size;
		}
	}
	return 0;
}

static int
initsym64(Bin *b, char **err)
{
	Elf64_Ehdr *ehdr;
	Elf64_Shdr *shdr, *s;
	char *p, *str;
	unsigned i;

	p = b->b;
	ehdr = (Elf64_Ehdr*)p;
	shdr = (Elf64_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_type == SHT_SYMTAB){
			b->symtab = p+s->sh_offset;
			b->strtab = p+shdr[s->sh_link].sh_offset;
			b->nsym = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_DYNSYM){
			b->dynsym = p+s->sh_offset;
			b->dynstr = p+shdr[s->sh_link].sh_offset;
			b->ndyn = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_REL
		   && !strcmp(".rel.plt", str+s->sh_name)){
			b->relplt = p+s->sh_offset;
			b->nrel = s->sh_size/s->sh_entsize;
		}
		if(s->sh_type == SHT_RELA
		   && !strcmp(".rela.plt", str+s->sh_name)){
			b->relaplt = p+s->sh_offset;
			b->nrela = s->sh_size/s->sh_entsize;
		}
		if(!strcmp(".data", str+s->sh_name))
			b->ndxdata = i;
		if(!strcmp(".rodata", str+s->sh_name))
			b->ndxrodata = i;
		if(!strcmp(".bss", str+s->sh_name))
			b->ndxbss = i;
		if(!strcmp(".text", str+s->sh_name)){
			b->ndxtext = i;
			b->vtext = s->sh_addr;
			b->ntext = s->sh_size;
		}
		if(!strcmp(".got", str+s->sh_name)){
			b->got = p+s->sh_offset;
			b->gotaddr = s->sh_addr;
		}
		if(!strcmp(".debug_abbrev", str+s->sh_name)){
			b->dw.abbrev = p+s->sh_offset;
			b->dw.nabbrev = s->sh_size;
		}
		if(!strcmp(".debug_aranges", str+s->sh_name)){
			b->dw.aranges = p+s->sh_offset;
			b->dw.naranges = s->sh_size;
		}
		if(!strcmp(".debug_frame", str+s->sh_name)){
			b->dw.frame = p+s->sh_offset;
			b->dw.nframe = s->sh_size;
		}
		if(!strcmp(".eh_frame", str+s->sh_name)){
			b->dw.ehframe = p+s->sh_offset;
			b->dw.nehframe = s->sh_size;
			b->dw.vehframe = s->sh_addr;
		}
		if(!strcmp(".debug_info", str+s->sh_name)){
			b->dw.info = p+s->sh_offset;
			b->dw.ninfo = s->sh_size;
		}
		if(!strcmp(".debug_line", str+s->sh_name)){
			b->dw.line = p+s->sh_offset;
			b->dw.nline = s->sh_size;
		}
		if(!strcmp(".debug_loc", str+s->sh_name)){
			b->dw.loc = p+s->sh_offset;
			b->dw.nloc = s->sh_size;
		}
		if(!strcmp(".debug_pubnames", str+s->sh_name)){
			b->dw.pubnames = p+s->sh_offset;
			b->dw.npubnames = s->sh_size;
		}
		if(!strcmp(".debug_ranges", str+s->sh_name)){
			b->dw.ranges = p+s->sh_offset;
			b->dw.nranges = s->sh_size;
		}
		if(!strcmp(".debug_str", str+s->sh_name)){
			b->dw.str = p+s->sh_offset;
			b->dw.nstr = s->sh_size;
		}
	}
	return 0;
}

static int
initsymtab(Bin *b, char **err)
{
	switch(b->mode){
	case M32:
		if(0 > initsym32(b, err))
			return -1;
		patchsym32(b);
		if(0 > reloc32(b, err))
			return -1;
		break;
	case M64:
		if(0 > initsym64(b, err))
			return -1;
		patchsym64(b);
		if(0 > reloc64(b, err))
			return -1;
		break;
	default:
		bug();
	}
	return 0;
}

static int
ispgmtype(Bin *b, u8 type, u32 ndx)
{
	if(type == STT_OBJECT || type == STT_FUNC)
		return 1;
	if(ndx == b->ndxtext)
		return 1;
	return 0;
}

static u8
elfsymflags(Bin *b, u8 type, u32 ndx)
{
	if(ndx == SHN_UNDEF){
		switch(type){
		case STT_FUNC:
			return Stext|Sundef;
		case STT_OBJECT:
			return Sdata|Sundef;
		default:
			bug();
			return 0;
		}
	}else if(ndx == b->ndxrodata)
		return Sro|Sdata;
	else if(ndx == b->ndxdata || ndx == b->ndxbss)
		return Sdata;
	else if(ndx == b->ndxtext)
		return Stext;
	else
		return 0; /* we don't know */
}

Sym*
lookaddr(Bin *b, u64 a)
{
	Sym *s;
	u32 n, i, z, m;
	u64 v;

	n = b->ss->n;
	if(n == 0)
		return 0;

	s = b->ss->sym[0];
	if(s->val > a)
		return 0;
	s = b->ss->sym[n-1];
	if(s->val < a)
		return 0;
	z = 0;
	while(1){
		i = n/2;
		m = z+i;
		if(i == 0)
			break;
		s = b->ss->sym[m];
		if(a < s->val)
			n = i;
		else{
			z = m;
			n = n-i;
		}
	}
	s = b->ss->sym[m];

	if((s->flags&Senum) == 0)
		return s;

	/* scan forward for any other non-enum symbol with same value */
	v = s->val;
	i = m;
	while(i < b->ss->n-1){
		i++;
		s = b->ss->sym[i];
		if(s->val > v)
			break;
		if(s->val == v && (s->flags&Senum) == 0)
			return s;
	}

	/* scan backward for first non-enum symbol */
	i = m;
	while(i > 0){
		i--;
		s = b->ss->sym[i];
		if((s->flags&Senum) == 0)
			return s;
	}

	return 0;
}

Sym*
looksym(Bin *b, char *id)
{
	return hget(b->sym, (uptr)id);
}

Syms*
enumsym(Bin *b)
{
	return b->ss;
}

static void
_idxsym32(Bin *b, char *tab, char *str, u32 n)
{
	unsigned i;
	Elf32_Sym *s;
	Sym *y;
	char *id;

	for(i = 0, s = (Elf32_Sym*)tab; i < n; i++, s++){
		id = str+s->st_name;
		if(*id == 0)
			continue;
		if(!ispgmtype(b, ELF32_ST_TYPE(s->st_info), s->st_shndx))
			continue;
		id = demangle(cidentifier(id));
		y = hget(b->sym, (uptr)id);
		if(y && (y->flags&Senum))
			continue;
		if(y == 0){
			y = mksym();
			y->id = xstrdup(id);
			y->type = mkvoid();
			hput(b->sym, (uptr)y->id, y);
		}
		y->val = b->base+s->st_value;
		y->flags |= elfsymflags(b, ELF32_ST_TYPE(s->st_info),
					s->st_shndx);
		y->sz = s->st_size;
	}
}

static void
idxsym32(Bin *b)
{
	_idxsym32(b, b->dynsym, b->dynstr, b->ndyn);
	_idxsym32(b, b->symtab, b->strtab, b->nsym);
}

static void
_idxsym64(Bin *b, char *tab, char *str, u32 n)
{
	unsigned i;
	Elf64_Sym *s;
	Sym *y;
	char *id;

	for(i = 0, s = (Elf64_Sym*)tab; i < n; i++, s++){
		id = str+s->st_name;
		if(*id == 0)
			continue;
		if(!ispgmtype(b, ELF64_ST_TYPE(s->st_info), s->st_shndx))
			continue;
		id = demangle(cidentifier(id));
		y = hget(b->sym, (uptr)id);
		if(y && (y->flags&Senum))
			continue;
		if(y == 0){
			y = mksym();
			y->id = xstrdup(id);
			y->type = mkvoid();
			hput(b->sym, (uptr)y->id, y);
		}
		y->val = b->base+s->st_value;
		y->flags |= elfsymflags(b, ELF64_ST_TYPE(s->st_info),
					s->st_shndx);
		y->sz = s->st_size;
	}
}

static void
idxsym64(Bin *b)
{
	_idxsym64(b, b->dynsym, b->dynstr, b->ndyn);
	_idxsym64(b, b->symtab, b->strtab, b->nsym);
}

static void
enum1sym(void *u, uptr id, void *v)
{
	addsym(u, v);
}

static int
symcmp(const void *v1, const void *v2)
{
	Sym *s1, *s2;
	s1 = *(Sym**)v1;
	s2 = *(Sym**)v2;
	if(s1->val < s2->val)
		return -1;
	if(s1->val > s2->val)
		return 1;
	return 0;
}

static void
sortsyms(Syms *ss)
{
	qsort(ss->sym, ss->n, sizeof(ss->sym[0]), symcmp);
}

static void
idxsym(Bin *b)
{
	switch(b->mode){
	case M32:
		idxsym32(b);
		break;
	case M64:
		idxsym64(b);
		break;
	default:
		bug();
	}
	hforeach(b->sym, enum1sym, b->ss);
	sortsyms(b->ss);
}

static void
initnodwarf(Bin *b)
{
	Cbase cb;
	Xtype xt;
	Type *t;
	Rootns *r;

	b->nodw.xt = mksht();
	for(cb = 0; cb < Vnallbase; cb++){
		b->nodw.base[cb].kind = Tbase;
		b->nodw.base[cb].base = cb;
		b->nodw.base[cb].rep = Rundef;
	}

	/* FIXME: don't assume that 64-bit is long pointer */
	if(b->enc == Mle && b->mode == M32)
		r = &rootns[Rc32le];
	else if(b->enc == Mle && b->mode == M64)
		r = &rootns[Rclp64le];
	else if(b->enc == Mbe && b->mode == M32)
		r = &rootns[Rc32be];
	else if(b->enc == Mbe && b->mode == M64)
		r = &rootns[Rclp64be];
	else
		bug();
	for(cb = Vlo; cb <= Vnbase; cb++)
		b->nodw.base[cb].rep = r->base[cb];
	for(xt = Xlo; xt < Xn; xt++){
		t = mktype();
		t->kind = Ttypedef;
		t->id = xstrdup(xtypename[xt]);
		t->link = copytype(&b->nodw.base[r->xtype[xt]]);
		hput(b->nodw.xt, (uptr)xtypename[xt], t);
	}
}

static void
free1type(void *u, uptr k, void *v)
{
	freetype(v);
}

static void
freenodwarf(Bin *b)
{
	hforeach(b->nodw.xt, free1type, 0);
	freeht(b->nodw.xt);
}

static void
do1xt(void *u, uptr id, void *v)
{
	Types *ts;
	ts = u;
	addtype(ts, copytype(v));
}

static Types*
nodwenumtype(Bin *b, char **err)
{
	Cbase cb;
	Types *ts;
	Type *t;

	ts = mktypes();

	// base types
	for(cb = Vlo; cb < Vnallbase; cb++){
		t = mktype();
		t->kind = Tbase;
		t->base = b->nodw.base[cb].base;
		t->rep = b->nodw.base[cb].rep;
		addtype(ts, t);
	}

	// typedef types
	hforeach(b->nodw.xt, do1xt, ts);

	return ts;
}

static Type*
nodwlooktype(Bin *b, Type *t, char **err)
{
	Type *rv;

	rv = 0;
	switch(t->kind){
	case Tbase:
		rv = mktype();
		rv->kind = Tbase;
		rv->base = t->base;
		rv->rep = b->nodw.base[t->base].rep;
		return rv;
	case Tstruct:
	case Tunion:
	case Tenum:
		*err = Enone;
		break;
	case Ttypedef:
		rv = hget(b->nodw.xt, (uptr)t->id);
		if(rv)
			return copytype(rv);
		*err = Enone;
		break;
	default:
		*err = Eproto;
		break;
	}
	return 0;
}


Bin*
mkbin(char *file, u64 base, char **err)
{
	Bin *b;
	char *p;
	u64 len;
	unsigned char m, e;

	p = mapfile(file, &len, err);
	if(p == 0)
		return 0;
	if(strncmp(p, ELFMAG, SELFMAG)){
		*err = Ebin;
		return 0;
	}

	switch(p[EI_CLASS]){
	case ELFCLASS32:
		m = M32;
		break;
	case ELFCLASS64:
		m = M64;
		break;
	default:
		xlog("%s: unknown ELF class %d", file, p[EI_CLASS]);
		break;
	}
	switch(p[EI_DATA]){
	case ELFDATA2LSB:
		e = Mle;
		break;
	case ELFDATA2MSB:
		e = Mbe;
		break;
	default:
		xlog("%s: unknown ELF data encoding %d", file, p[EI_DATA]);
		break;
	}
	b = xmalloc(sizeof(Bin));
	b->id = nextid++;
	b->path = xstrdup(file);
	b->b = p;
	b->len = len;
	b->base = base;
	b->mode = m;
	b->enc = e;
	b->sym = mksht();
	b->ss = mksyms();
	b->refcnt = 1;
	hput(bintab, b->id, b);
	if(0 > initsymtab(b, err)){
		freebin(b);
		return 0;
	}
	if(b->dw.info && !ignoreg)
		initdwarf(b);
	else
		initnodwarf(b);
	idxsym(b);
	return b;
}

static void
free1sym(void *u, uptr xid, void *v)
{
	freesym(v);
}

static void
refbin(Bin *b)
{
	b->refcnt++;
}

void
freebin(Bin *b)
{
	if(b->refcnt-- > 1)
		return;
	if(b->dw.info && !ignoreg)
		freedwarf(b);
	else
		freenodwarf(b);
	hdel(bintab, b->id);
	free(b->path);
	munmap(b->b, b->len);
	hforeach(b->sym, free1sym, 0);
	freeht(b->sym);
	freesyms(b->ss);
	free(b);
}

static void
enumseg32(Bin *b, Segs *gs)
{
	char *p, *str;
	Elf32_Ehdr *ehdr;
	Elf32_Shdr *shdr, *s;
	Seg *g;
	unsigned i;

	p = b->b;
	ehdr = (Elf32_Ehdr*)p;
	shdr = (Elf32_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_addr == 0)
			continue;
		g = nextseg(gs);
		g->addr = b->base+s->sh_addr;
		g->len = s->sh_size;
		g->name = xstrdup(str+s->sh_name);
	}
}

static void
enumseg64(Bin *b, Segs *gs)
{
	char *p, *str;
	Elf64_Ehdr *ehdr;
	Elf64_Shdr *shdr, *s;
	Seg *g;
	unsigned i;

	p = b->b;
	ehdr = (Elf64_Ehdr*)p;
	shdr = (Elf64_Shdr*)(p+ehdr->e_shoff);
	str = p+shdr[ehdr->e_shstrndx].sh_offset;
	for(i = 0; i < ehdr->e_shnum; i++){
		s = shdr+i;
		if(s->sh_addr == 0)
			continue;
		g = nextseg(gs);
		g->addr = b->base+s->sh_addr;
		g->len = s->sh_size;
		g->name = xstrdup(str+s->sh_name);
	}
}

static Segs*
enumseg(Bin *b)
{
	Segs *gs;
	gs = mksegs();
	switch(b->mode){
	case M32:
		enumseg32(b, gs);
		return gs;
	case M64:
		enumseg64(b, gs);
		return gs;
	default:
		bug();
		return 0;
	}
}

// append B to end of list
void
addbin(Ctl *ctl, Bin *b)
{
	Bin **q;

	b->link = 0;
	q = &ctl->as->x.bin;
	while(*q)
		q = &(*q)->link;
	*q = b;
}

Bin*
lookbinpath(Ctl *ctl, char *path)
{
	Bin *b;
	b = ctl->as->x.bin;
	while(b){
		if(!strncmp(path, b->path, strlen(b->path)))
			return b;
		b = b->link;
	}
	return 0;
}

Bin*
lookbinpc(Ctl *ctl, u64 pc)
{
	Bin *b;
	b = ctl->as->x.bin;
	while(b){
		if(b->vtext <= pc-b->base && pc-b->base < b->vtext+b->ntext)
			return b;
		b = b->link;
	}
	return 0;
}

void
symstat(Ctl *ctl, Stat *s)
{
	Bin *b;
	unsigned n;

	b = ctl->as->x.bin;
	n = 0;
	while(b){
		b = b->link;
		n++;
	}
	s->b = xmalloc(n*sizeof(Bin*));
	s->nb = n;

	b = ctl->as->x.bin;
	n = 0;
	while(b){
		s->b[n++] = b;
		b = b->link;
	}
}

Bin*
names(char *path, char **err)
{
	Bin *b;
	b = mkbin(path, 0, err);
	if(!b)
		return 0;
	return b;
}

void
symdupas(As *as, As *nas)
{
	Bin *b;
	b = as->x.bin;
	while(b){
		refbin(b);
		b = b->link;
	}
	nas->x.bin = as->x.bin;
}

void
symfreeas(As *as)
{
	Bin *b, *n;
	free(as->x.aux);
	b = as->x.bin;
	while(b){
		n = b->link;
		freebin(b);
		b = n;
	}
	as->x.aux = 0;
	as->x.bin = 0;
}

void
syminit()
{
	bintab = mkht();
	dwarfinit();
}

void
symfini()
{
	freeht(bintab);
	dwarffini();
}

struct Nsreq
{
	u64 id;
	u64 addr;
	char *sym;
	char *path;
};

#undef XRD
#define XRD(to,fr,sz) { 		\
	if(len < (sz)) goto fail;	\
	memcpy((to),(fr),(sz));		\
	(fr) += (sz);			\
	len -= (sz);			\
}

/* sctl does not yet serve attributes, but
   the protocol has come to require them. */
static void
bufpushnullattr(Buf *x)
{
	bufpushu64(x, 0); /* no attributes */
}

static void
bufpushtname(Buf *x, Type *t)
{
	u32 i;
	Decl *d;

	switch(t->kind){
	case Tbase:
		bufpushu8(x, Tbase);
		bufpushu8(x, t->base);
		break;
	case Tstruct:
	case Tunion:
	case Tenum:
		bufpushu8(x, t->kind);
		bufpushs(x, t->id);
		break;
	case Tptr:
		bufpushu8(x, Tptr);
		bufpushtname(x, t->link);
		break;
	case Tconst:
		bufpushu8(x, Tconst);
		bufpushtname(x, t->link);
		break;
	case Tarr:
		bufpushu8(x, Tarr);
		bufpushu64(x, t->n);
		bufpushtname(x, t->link);
		break;
	case Tfun:
		bufpushu8(x, Tfun);
		bufpushtname(x, t->link);
		bufpushu64(x, t->n);
		for(i = 0; i < t->n; i++){
			d = t->ds[i];
			bufpushnullattr(x);
			if(d->id)
				bufpushs(x, d->id);
			else
				bufpushs(x, "");
			bufpushtname(x, d->type);
		}
		break;
	case Ttypedef:
		bufpushu8(x, Ttypedef);
		bufpushs(x, t->id);
		break;
	case Tbitfield:
		bufpushu8(x, Tbitfield);
		bufpushu8(x, t->w);
		bufpushu8(x, t->pos);
		bufpushtname(x, t->link);
		break;
	default:
		xfatal("bufpushtname: unimplemented case %d", t->kind);
	}
}

static void
bufpushtdef(Buf *x, Type *t)
{
	u32 i;
	Decl *d;
	Enum *e;

	switch(t->kind){
	case Tbase:
		bufpushu8(x, Tbase);
		bufpushu8(x, t->base);
		bufpushu8(x, t->rep);
		break;
	case Tstruct:
	case Tunion:
		bufpushu8(x, t->kind);
		bufpushs(x, t->id);
		bufpushu64(x, (u64)t->sz);
		bufpushnullattr(x);
		bufpushu64(x, (u64)t->n);
		for(i = 0; i < t->n; i++){
			d = t->ds[i];
			bufpushu64(x, d->off);
			bufpushnullattr(x);
			bufpushs(x, d->id);
			bufpushtname(x, d->type);
		}
		break;
	case Tenum:
		bufpushu8(x, Tenum);
		bufpushs(x, t->id);
		bufpushu8(x, t->rep);
		bufpushu64(x, (u64)t->n);
		for(i = 0; i < t->n; i++){
			e = t->en[i];
			bufpushs(x, e->id);
			bufpushu64(x, e->val);
		}
		break;
	case Tptr:
		bufpushu8(x, Tptr);
		bufpushtname(x, t->link);
		break;
	case Tconst:
		bufpushu8(x, Tconst);
		bufpushtname(x, t->link);
		break;
	case Tarr:
		bufpushu8(x, Tarr);
		bufpushu64(x, t->n);
		bufpushtname(x, t->link);
		break;
	case Tfun:
		bufpushu8(x, Tfun);
		bufpushtname(x, t->link);
		bufpushu64(x, t->n);
		for(i = 0; i < t->n; i++){
			d = t->ds[i];
			bufpushnullattr(x);
			if(d->id)
				bufpushs(x, d->id);
			else
				bufpushs(x, "");
			bufpushtname(x, d->type);
		}
		break;
	case Ttypedef:
		bufpushu8(x, Ttypedef);
		bufpushs(x, t->id);
		bufpushtname(x, t->link);
		break;
	default:
		xfatal("bufpushtdef: unimplemented case %d", t->kind);
	}
}

static void
bufpushsym(Buf *x, Sym *s)
{
	bufpushs(x, s->id);
	bufpushu8(x, s->flags);
	bufpushu64(x, s->val);
	bufpushnullattr(x);
	bufpushu64(x, s->sz);
	bufpushtname(x, s->type);
}

static void
bufpushrule(Buf *x, Rule *r, u32 n)
{
	u32 i;
	Rulekind k;
	bufpushu32(x, n);
	for(i = 0; i < n; i++){
		k = r[i].kind;
		if(k == Rexp || k == Rvexp || k == Rcfae)
			xfatal("bufpushrule: unimplemented rule %d", k); 
		bufpushu8(x, k);
		bufpushu64(x, r[i].r);
		bufpushu64(x, r[i].n); /* FIXME: this is signed */
	}
}

static void
bufpushlexpr(Buf *x, Lexpr *le)
{
	if(le == 0)
		xfatal("unexpected null lexpr");
		
	bufpushu8(x, le->lkind);
	switch(le->lkind){
	case Lreg:
		bufpushu8(x, le->no);
		break;
	case Lderef:
		bufpushlexpr(x, le->op1);
		break;
	case Ladd:
	case Lsub:
		bufpushlexpr(x, le->op1);
		bufpushlexpr(x, le->op2);
		break;
	case Lulit:
		bufpushu64(x, le->u);
		break;
	case Lslit:
		bufpushu64(x, le->s);
		break;
	default:
		xfatal("unexpected lexpr kind");
	}
}

Types*
enumtype(Bin *b, char **err)
{
	if(b->dw.info && !ignoreg)
		return dwenumtype(b, err);
	else
		return nodwenumtype(b, err);
}

Type*
looktype(Bin *b, Type *t, char **err)
{
	if(b->dw.info && !ignoreg)
		return dwlooktype(b, t, err);
	else
		return nodwlooktype(b, t, err);
}

int
ctlns(u8 op, void *buf, u64 len, void *xbuf, char **err)
{
	Bin *b;
	Buf *x;
	char *p;
	int rv;
	u64 i, j, m, pc;
	struct Nsreq q;
	Sym *s;
	Syms *ss;
	Seg *g;
	Segs *gs;
	Locs *ls;
	Loc *loc;
	L *l;
	Type ti, *to;
	Types *ts;
	u8 tr;
	Rule rule[Nrow];
	Srcloc *srcloc;
	char *file;
	u32 line;

	*err = Ebug;
	x = xbuf;
	p = buf;
	memset(&q, 0, sizeof(q));
	switch(op){
	case Tlooksym:
		XRD(&q.id, p, sizeof(q.id));
		XRD(&m, p, sizeof(m));
		q.sym = xmalloc(m+1);
		XRD(q.sym, p, m);
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		s = looksym(b, q.sym);
		if(s == 0){
			*err = Enone;
			goto fail;
		}
		bufpushsym(x, s);
		free(q.sym);
		break;
	case Tenumsym:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		ss = enumsym(b);
		bufpushu64(x, ss->n);
		for(i = 0; i < ss->n; i++)
			bufpushsym(x, ss->sym[i]);
		break;
	case Tlookaddr:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&q.addr, p, sizeof(q.addr));
		s = lookaddr(b, q.addr);
		if(s == 0){
			*err = Enone;
			goto fail;
		}
		bufpushsym(x, s);
		break;
	case Tlooktype:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&tr, p, sizeof(tr));
		memset(&ti, 0, sizeof(ti));
		switch(tr){
		case Tbase:
			ti.kind = Tbase;
			XRD(&tr, p, sizeof(tr));
			ti.base = tr;
			if(ti.base > Vnbase){
				*err = Eproto;
				goto fail;
			}
			to = looktype(b, &ti, err);
			break;
		case Tstruct:
		case Tunion:
		case Tenum:
		case Ttypedef:
			ti.kind = tr;
			XRD(&m, p, sizeof(m));
			ti.id = xmalloc(m+1);
			XRD(ti.id, p, m);
			to = looktype(b, &ti, err);
			free(ti.id);
			break;
		case Tptr:
		case Tarr:
		case Tfun:
		default:
			*err = Eproto;
			goto fail;
		}
		if(to == 0)
			goto fail;
		bufpushtdef(x, to);
		freetype(to);
		break;
	case Tenumtype:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		ts = enumtype(b, err);
		if(ts == 0)
			goto fail;
		bufpushu64(x, (u64)ts->n);
		for(i = 0; i < ts->n; i++)
			bufpushtdef(x, ts->t[i]);
		freetypes(ts);
		break;
	case Tenumloc:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&pc, p, sizeof(pc));
		ls = enumloc(b, pc, err);
		if(ls == 0)
			goto fail;
		/* FIXME: two messes:
		   - locals data structure is too complex at this level
		   - enumloc returns relative addresses but receives absolute
		*/
		pc -= b->base;
		m = 0;
		for(i = 0; i < ls->n; i++){
			loc = &ls->loc[i];
			for(j = 0; j < loc->ls->n; j++){
				l = &loc->ls->l[j];
				if(l->lo > pc || l->hi <= pc)
					continue;
				m++;
			}
		}
		bufpushu64(x, m);
		for(i = 0; i < ls->n; i++){
			loc = &ls->loc[i];
			for(j = 0; j < loc->ls->n; j++){
				l = &loc->ls->l[j];
				if(l->lo > pc || l->hi <= pc)
					continue;
				bufpushs(x, loc->id);
				bufpushtname(x, loc->type);
				bufpushu8(x, loc->vkind);
				bufpushlexpr(x, l->le);
			}
		}
		freelocs(ls);
		break;
	case Tenumseg:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		gs = enumseg(b);
		bufpushu64(x, gs->n);
		for(i = 0; i < gs->n; i++){
			g = &gs->seg[i];
			bufpushu64(x, g->addr);
			bufpushu64(x, g->len);
			bufpushs(x, g->name);
		}
		freesegs(gs);
		break;
	case Tunwind1:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&pc, p, sizeof(pc));
		rv = unwind1(b, pc, rule, err);
		if(0 > rv)
			goto fail;
		bufpushrule(x, rule, Nrow);
		break;
	case Tlooksrc:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&pc, p, sizeof(pc));
		srcloc = looksrc(b, pc, err);
		if(srcloc == 0)
			goto fail;
		bufpushsrc(x, srcloc);
		break;
	case Tlookpc:
		XRD(&q.id, p, sizeof(q.id));
		b = hget(bintab, q.id);
		if(b == 0){
			*err = Eid;
			goto fail;
		}
		XRD(&m, p, sizeof(m));
		file = xmalloc(m+1);
		XRD(file, p, m);
		XRD(&line, p, sizeof(line));
		rv = lookpc(b, file, line, &pc, err);
		free(file);
		if(0 > rv)
			goto fail;
		bufpushu64(x, pc);
		break;
	case Tnames:
		XRD(&m, p, sizeof(m));
		q.path = xmalloc(m+1);
		XRD(q.path, p, m);
		b = names(q.path, err);
		if(b == 0)
			goto fail;
		bufpushu64(x, b->id);
		free(q.path);
		break;
	default:
		goto fail;
	}
	return 0;
fail:
	if(q.sym)
		free(q.sym);
	if(q.path)
		free(q.path);
	return -1;
}
