#include "../inc/sctl.h"
#include "../inc/err.h"
#include "win.h"

char osversion[] = "x86-win-2012";

enum
{
	int3 = 0xcc,		/* breakpoint */
	FL_TF = 0x100,		/* eflags single step bit */
	Wblock = 0,
	Wnoblock = 1,
};

static HT *pid2ctl;
static Event *evqfromsys;  /* pending events, newest first */
static Event *evqtouser;   /* pending events, oldest first */

char*
dbg2str(int c)
{
	switch(c){
	case EXCEPTION_DEBUG_EVENT:
		return "EXCEPTION_DEBUG_EVENT";
	case OUTPUT_DEBUG_STRING_EVENT:
		return "OUTPUT_DEBUG_STRING_EVENT";
	case UNLOAD_DLL_DEBUG_EVENT:
		return "UNLOAD_DLL_DEBUG_EVENT";
	case LOAD_DLL_DEBUG_EVENT:
		return "LOAD_DLL_DEBUG_EVENT";
	case EXIT_PROCESS_DEBUG_EVENT:
		return "EXIT_PROCESS_DEBUG_EVENT";
	case EXIT_THREAD_DEBUG_EVENT:
		return "EXIT_THREAD_DEBUG_EVENT";
	case RIP_EVENT:
		return "RIP_EVENT";
	case CREATE_PROCESS_DEBUG_EVENT:
		return "CREATE_PROCESS_DEBUG_EVENT";
	case CREATE_THREAD_DEBUG_EVENT:
		return "CREATE_THREAD_DEBUG_EVENT";
	default:
		return "(unknown debug event)";
	}
}

static char*
dbgex2str(int c)
{
	static char buf[32];
	switch(c){
	case EXCEPTION_ACCESS_VIOLATION:
		return "EXCEPTION_ACCESS_VIOLATION";
	case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
		return "EXCEPTION_ARRAY_BOUNDS_EXCEEDED";
	case EXCEPTION_BREAKPOINT:
		return "EXCEPTION_BREAKPOINT";
	case EXCEPTION_DATATYPE_MISALIGNMENT:
		return "EXCEPTION_DATATYPE_MISALIGNMENT";
	case EXCEPTION_FLT_DENORMAL_OPERAND:
		return "EXCEPTION_FLT_DENORMAL_OPERAND";
	case EXCEPTION_FLT_DIVIDE_BY_ZERO:
		return "EXCEPTION_FLT_DIVIDE_BY_ZERO";
	case EXCEPTION_FLT_INEXACT_RESULT:
		return "EXCEPTION_FLT_INEXACT_RESULT";
	case EXCEPTION_FLT_INVALID_OPERATION:
		return "EXCEPTION_FLT_INVALID_OPERATION";
	case EXCEPTION_FLT_OVERFLOW:
		return "EXCEPTION_FLT_OVERFLOW";
	case EXCEPTION_FLT_STACK_CHECK:
		return "EXCEPTION_FLT_STACK_CHECK";
	case EXCEPTION_FLT_UNDERFLOW:
		return "EXCEPTION_FLT_UNDERFLOW";
	case EXCEPTION_ILLEGAL_INSTRUCTION:
		return "EXCEPTION_ILLEGAL_INSTRUCTION";
	case EXCEPTION_IN_PAGE_ERROR:
		return "EXCEPTION_IN_PAGE_ERROR";
	case EXCEPTION_INT_DIVIDE_BY_ZERO:
		return "EXCEPTION_INT_DIVIDE_BY_ZERO";
	case EXCEPTION_INT_OVERFLOW:
		return "EXCEPTION_INT_OVERFLOW";
	case EXCEPTION_INVALID_DISPOSITION:
		return "EXCEPTION_INVALID_DISPOSITION";
	case EXCEPTION_NONCONTINUABLE_EXCEPTION:
		return "EXCEPTION_NONCONTINUABLE_EXCEPTION";
	case EXCEPTION_PRIV_INSTRUCTION:
		return "EXCEPTION_PRIV_INSTRUCTION";
	case EXCEPTION_SINGLE_STEP:
		return "EXCEPTION_SINGLE_STEP";
	case EXCEPTION_STACK_OVERFLOW:
		return "EXCEPTION_STACK_OVERFLOW";
	case EXCEPTION_GUARD_PAGE:
		return "EXCEPTION_GUARD_PAGE";
	default:
		snprintf(buf, sizeof(buf), "unknown exception: 0x%x", c);
		return buf;
	}
}

static void
showdbg(DEBUG_EVENT *dbg)
{
	xlog("dbg: %d %d %s",
	     dbg->dwProcessId,
	     dbg->dwThreadId,
	     dbg2str(dbg->dwDebugEventCode));
}

static void
showexception(DEBUG_EVENT *dbg)
{
	xlog("%s: %d %d %s",
	     dbg2str(dbg->dwDebugEventCode),
	     dbg->dwProcessId,
	     dbg->dwThreadId,
	     dbgex2str(dbg->u.Exception.ExceptionRecord.ExceptionCode));
}

static char*
wdork(char *buf)
{
	char *p, *q;
	p = q = buf;
	printf("wdork: ");
	while(q[0] != 0 && q[1] != 0){
		printf(".%c.", *q);
		*p = *q;
		p++;
		q += 2;
	}
	printf("\n");
	*p = 0;
	return buf;
}

static int
atunload(Ctl *ctl, DEBUG_EVENT *dbg, char **dll, u64 *base, char **err)
{
	Region *r;
	u64 i, n, b;

	ctl->state = Sstop;
	b = (u64)dbg->u.UnloadDll.lpBaseOfDll;
	r = dlls(ctl->x.process, &n, err);
	if(r == 0)
		return -1;
	for(i = 0; i < n; i++){
		if(r[i].base == b){
			*dll = xstrdup(r[i].path);
			*base = b;
		}
	}
	if(i >= n){
		*dll = xstrdup("<unknown-dll>");
		*base = b;
	}
	freeregions(r, n);
	return 0;
}

// reading LoadDll.lpImageName would be simpler but it is not reliable.
// this wizardry is from msdn.
static int
dllname(HANDLE dll, char *buff, int max)
{
	HANDLE map;
	LPVOID ptr;
	char szTemp[MAX_PATH], szName[MAX_PATH], szTempFile[MAX_PATH];
	char szDrive[3] = TEXT(" :");
	char *p;
	BOOL bFound;
	UINT uNameLen;

	map = CreateFileMapping(dll, 0, PAGE_READONLY, 0, 1, 0);
	if(map == NULL)
		return -1;
	ptr = MapViewOfFile(map, FILE_MAP_READ, 0, 0, 1);
	if(ptr == NULL){
		CloseHandle(map);
		return -1;
	}
	if(!GetMappedFileName(GetCurrentProcess(), ptr, buff, max)){
		UnmapViewOfFile(ptr);
		CloseHandle(map);
		return -1;
	}

	// translate device name to drive letter
	szTemp[0] = 0;
	if(GetLogicalDriveStrings(MAX_PATH-1, szTemp)){
		bFound = FALSE;
		p = szTemp;
		do{
			*szDrive = *p;
			if(QueryDosDevice(szDrive, szName, MAX_PATH)){
				uNameLen = _tcslen(szName);
				if(uNameLen < MAX_PATH){
					bFound = _tcsnicmp(buff, szName,
							   uNameLen) == 0;
					if(bFound){
						sprintf(szTempFile,
							TEXT("%s%s"),
							szDrive,
							buff+uNameLen);
						strncpy(buff, szTempFile,
							MAX_PATH);
						buff[MAX_PATH-1] = 0;
					}
				}
			}
			while(*p++);
		}while(!bFound && *p);
	}
	UnmapViewOfFile(ptr);
	CloseHandle(map);
	return 0;
}

static int
atload(Ctl *ctl, DEBUG_EVENT *dbg, char **dll, u64 *base, char **err)
{
	char ubuf[MAX_PATH*2];

	ctl->state = Sstop;
	if(0 > dllname(dbg->u.LoadDll.hFile, ubuf, MAX_PATH)){
		if(err)
			*err = lasterror();
		return -1;
	}
	*dll = xstrdup(ubuf);
	*base = (u64)dbg->u.LoadDll.lpBaseOfDll;
	return 0;
}

As*
mkas(Ctl *ctl)
{
	As *as;
	as = _mkas(ctl);
	return as;
}

void
freeas(As *as)
{
	_freeas(as);
}

static u64
ctlid(u32 pid, u32 tid)
{
	return ((u64)pid<<32)|tid;
}

Ctl*
mkctl(u32 pid, u32 tid)
{
	Ctl *ctl;
	ctl = _mkctl(ctlid(pid, tid));
	ctl->tracemode = Etrap|Eclone|Efork|Eexit|Eload|Eunload|Esignal;
	ctl->x.pid = pid;
	ctl->x.tid = tid;
	ctl->x.process = INVALID_HANDLE_VALUE;
	ctl->x.thread = INVALID_HANDLE_VALUE;
	return ctl;
}

void
freectl(Ctl *ctl)
{
	_freectl(ctl);
}

void
freeevent(Event *e)
{
	if(e->kind == Eload)
		free(e->u.load.path);
	free(e);
}

static void
evqpush(Event *p)
{
	p->link = evqfromsys;
	evqfromsys = p;
}

static Event*
evqreverse(Event *p)
{
	Event *t, *q;

	q = 0;
	while(p){
		t = p->link;
		p->link = q;
		q = p;
		p = t;
	}
	return q;
}

static Event*
evqappend(Event *p, Event *q)
{
	Event *t;
	if(p == 0)
		return q;
	if(q == 0)
		return p;
	t = p;
	while(t->link)
		t = t->link;
	t->link = q;
	return p;
}

static void
evqfix()
{
	evqappend(evqtouser, evqreverse(evqfromsys));
	evqfromsys = 0;
}

Event*
evqpop()
{
	Event *p;
	evqfix();
	p = evqtouser;
	if(p)
		evqtouser = p->link;
	return p;
}

/* find first event matching ctl on the event queue */
static Event*
evqmatch(Ctl *ctl)
{
	Event *p, **q;
	evqfix();
	q = &evqtouser;
	p = evqtouser;
	while(p){
		if(p->id == ctl->id){
			*q = p->link;
			p->link = 0;
			return p;
		}
		q = &p->link;
		p = p->link;
	}
	return 0;
}

static Event*
eventtrap(Ctl *ctl, Brkref *br)
{
	Event *e;
	if(br == 0)
		return 0;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Etrap;
	e->u.trap.id = br->id;
	e->u.trap.ctx = ctl->ctx;
	e->link = eventtrap(ctl, br->link);
	return e;
}

static Event*
eventexit(Ctl *ctl)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eexit;
	return e;
}

static Event*
eventspawn(Evkind kind, Ctl *ctl,  Ctl *spwn)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = kind;
	e->u.spawn.cid = spwn->id;
	e->u.spawn.ctx = ctl->ctx;
	e->u.spawn.cctx = spwn->ctx;
	return e;
}

static Event*
eventclone(Ctl *ctl, Ctl *spwn)
{
	return eventspawn(Eclone, ctl, spwn);
}

static Event*
eventfork(Ctl *ctl, Ctl *spwn)
{
	return eventspawn(Efork, ctl, spwn);
}

static Event*
eventload(Ctl *ctl, char *path, u64 base)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eload;
	e->u.load.path = xstrdup(path);
	e->u.load.base = base;
	e->u.load.ctx = ctl->ctx;
	return e;
}

static Event*
eventunload(Ctl *ctl, char *path, u64 base)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Eunload;
	e->u.load.path = xstrdup(path);
	e->u.load.base = base;
	e->u.load.ctx = ctl->ctx;
	return e;
}

static Event*
eventsignal(Ctl *ctl)
{
	Event *e;
	e = xmalloc(sizeof(Event));
	e->id = ctl->id;
	e->kind = Esignal;
	e->u.signal.ctx = ctl->ctx;
	return e;
}

void
packevent(Buf *b, Event *e)
{
	bufpushu64(b, e->id);
	bufpushu64(b, e->kind);
	switch(e->kind){
	case Etrap:
		bufpushu64(b, e->u.trap.id);
		packctx(b, &e->u.trap.ctx);
		break;
	case Eclone:
	case Efork:
		bufpushu64(b, e->u.spawn.cid);
		packctx(b, &e->u.spawn.ctx);
		packctx(b, &e->u.spawn.cctx);
		break;
	case Eexit:
		break;
	case Eload:
	case Eunload:
		bufpushs(b, e->u.load.path);
		bufpushu64(b, e->u.load.base);
		packctx(b, &e->u.load.ctx);
		break;
	case Esignal:
		packctx(b, &e->u.signal.ctx);
		break;
	default:
		bug();
	}
}

void
ctlcont(Ctl *ctl)
{
	if(ctl->ctx.curexn){
		if(!ContinueDebugEvent(ctl->x.pid, ctl->x.tid,
				       DBG_EXCEPTION_NOT_HANDLED))
			xfatalno("ContinueDebugEvent pid: %d tid: %d",
				 ctl->x.pid, ctl->x.tid);
	}else{
		if(!ContinueDebugEvent(ctl->x.pid, ctl->x.tid, DBG_CONTINUE))
			xfatalno("ContinueDebugEvent pid: %d tid: %d",
				 ctl->x.pid, ctl->x.tid);
	}
	ctl->ctx.curexn = 0;
	ctl->state = Srun;
}

int
ctltrace(Ctl *ctl, Trace *t, char **err)
{
	return _ctltrace(ctl, t, err);
}

static void
xgetregs(Ctl *ctl, Reg *regs)
{
	regs->ContextFlags = CONTEXT_FULL;
	if(!GetThreadContext(ctl->x.thread, regs))
		xfatalno("GetThreadContext");
}

void
xsetregs(Ctl *ctl, Reg *regs)
{
	regs->ContextFlags = CONTEXT_FULL;
	if(!SetThreadContext(ctl->x.thread, regs))
		xfatalno("SetThreadContext");
}

static void
atbrkx86(Ctl *ctl)
{
	xgetregs(ctl, &ctl->ctx.regs);
	--ctl->ctx.regs.PC;    /* pc follows 1-byte trap instruction */
	xsetregs(ctl, &ctl->ctx.regs);
}

static void
notatbrkx86(Ctl *ctl)
{
	++ctl->ctx.regs.PC;    /* pc follows 1-byte trap instruction */
	xsetregs(ctl, &ctl->ctx.regs);
}

static void
_wait1(u32 pid, u32 tid, DEBUG_EVENT *dbg, int mode)
{
	Event *e;
	char *err;
	while(1){
		if(!WaitForDebugEvent(dbg, mode == Wblock ? INFINITE : 0))
			/* FIXME: return value on Wnoblock? */
			xfatalno("WaitForDebugEvent");
		if(pid == dbg->dwProcessId
		   && (tid == 0 || tid == dbg->dwThreadId))
			break;
		err = 0;
		e = doevent(dbg, &err);
		if(!e && err)
			xfatal("doevent: %s", err);
		if(e)
			evqpush(e);
	}
}

static void
wait1(Ctl *ctl, DEBUG_EVENT *dbg, int mode)
{
	_wait1(ctl->x.pid, ctl->x.tid, dbg, mode);
}

Event*
doevent(DEBUG_EVENT *dbg, char **err)
{
	Ctl *ctl, *spwn;
	Event *e;
	u32 pid, tid, c;
	Brkref *br;
	char *dll;
	u64 base;

	e = 0;
	pid = dbg->dwProcessId;
	tid = dbg->dwThreadId;
	ctl = lookctl(ctlid(pid, tid));

//	showdbg(dbg);

	if(ctl == 0){
		switch(dbg->dwDebugEventCode){
		case CREATE_PROCESS_DEBUG_EVENT:
			ctl = mkctl(dbg->dwProcessId, dbg->dwThreadId);
			ctl->x.pid = dbg->dwProcessId;
			ctl->x.tid =  dbg->dwThreadId;
			ctl->x.process = dbg->u.CreateProcessInfo.hProcess;
			ctl->x.thread = dbg->u.CreateProcessInfo.hThread;
			ctl->state = Sstop;
			ctl->as = mkas(ctl);
			/* FIXME:  ctl->whence = ???; */
			hput(pid2ctl, ctl->x.pid, ctl);
			/* FIXME: Efork requires a parent ctl; we don't have
			   one, so we can't report this event. */
			ctlcont(ctl);
			break;
		case CREATE_THREAD_DEBUG_EVENT:
			ctl = hget(pid2ctl, pid);
			if(ctl == 0)
				xfatal("no parent ctl for new thread %d:%d",
				       pid, tid);
			spwn = mkctl(pid, tid);
			spwn->x.process = ctl->x.process;
			spwn->x.thread = dbg->u.CreateThread.hThread;
			spwn->state = Sstop;
			spwn->whence = ctl->whence;
			linkspawn(ctl, spwn, 0);
			if(ctl->tracemode&Eclone)
				e = eventclone(ctl, spwn);
				/* FIXME: parent is running, contrary to Eclone
				   semantics */
			else
				ctlcont(spwn);
			break;
		default:
			xlog("unexpected %s", dbg2str(dbg->dwDebugEventCode));
			ContinueDebugEvent(dbg->dwProcessId,
					   dbg->dwThreadId,
					   DBG_CONTINUE);
		}
		return e;
	}

	switch(dbg->dwDebugEventCode){
	case EXCEPTION_DEBUG_EVENT:
		c = dbg->u.Exception.ExceptionRecord.ExceptionCode;
		if(c != EXCEPTION_BREAKPOINT){
			ctl->state = Sstop;
			if(ctl->ctx.curexn)
				/* FIXME: can this happen? */
				xfatal("prctl: nested exception");
			ctl->ctx.curexn = c;
			if(ctl->tracemode&Esignal)
				// FIXME: get current context?
				e = eventsignal(ctl);
			else
				ctlcont(ctl);
			break;
		}
		atbrkx86(ctl);
		br = atbrk(ctl);
		if(br)
			e = eventtrap(ctl, br);
		else if(ctl->state == Strap){
			if(0 > steptrap(ctl, err))
				return 0;
			ctlcont(ctl);
		}else{
			notatbrkx86(ctl);
			ctlcont(ctl);
		}
		break;
	case OUTPUT_DEBUG_STRING_EVENT:
		ctlcont(ctl);
		break;
	case UNLOAD_DLL_DEBUG_EVENT:
		if(ctl->tracemode&Eunload){
			if(0 > atunload(ctl, dbg, &dll, &base, err)){
				xlog("undecodable UNLOAD_DLL_DEBUG_EVENT: %s",
				     *err);
				*err = 0;
				ctlcont(ctl);
			}else{
				e = eventunload(ctl, dll, base);
				free(dll);
			}
		}else
			ctlcont(ctl);
		break;
	case LOAD_DLL_DEBUG_EVENT:
		if(ctl->tracemode&Eload){
			if(0 > atload(ctl, dbg, &dll, &base, err)){
				xlog("undecodable LOAD_DLL_DEBUG_EVENT: %s",
				     *err);
				*err = 0;
				ctlcont(ctl);
			}else{
				e = eventload(ctl, dll, base);
				free(dll);
			}
		}else
			ctlcont(ctl);
		if(dbg->u.LoadDll.hFile)
			CloseHandle(dbg->u.LoadDll.hFile);
		break;
	case EXIT_PROCESS_DEBUG_EVENT:
		if(ctl->tracemode&Eexit)
			e = eventexit(ctl);
		ctlcont(ctl);
		dead(ctl);
		break;
	case EXIT_THREAD_DEBUG_EVENT:
		if(ctl->tracemode&Eexit)
			e = eventexit(ctl);
		ctlcont(ctl);
		dead(ctl);
		break;
	case RIP_EVENT:
		xfatal("unexpected RIP_EVENT");
		break;
	case CREATE_PROCESS_DEBUG_EVENT:
		xfatal("unexpected CREATE_PROCESS_DEBUG_EVENT");
		break;
	case CREATE_THREAD_DEBUG_EVENT:
		xfatal("unexpected CREATE_THREAD_DEBUG_EVENT");
		break;
	default:
		xfatal("unknown debug event %d\n", dbg->dwDebugEventCode);
		break;
	}
	return e;
}

void
sctlinit()
{
	_sctlinit();
	pid2ctl = mkht();
}

void
sctlfini()
{
	_sctlfini();
}

void
dead(Ctl *ctl)
{
	_dead(ctl);
}

void
freelaunch(Launch *l)
{
	free(l->cmd);
	free(l);
}

Launch*
unpacklaunch(void *buf, u64 len)
{
	Launch *l;
	char *p, *q;
	u64 i, narg, m;

	l = xmalloc(sizeof(Launch));
	p = buf;
	XRD(&l->flags, p, sizeof(l->flags));
	XRD(&narg, p, sizeof(narg));
	l->cmd = xmalloc(len+1);
	q = l->cmd;
	for(i = 0; i < narg; i++){
		XRD(&m, p, sizeof(m));
		XRD(q, p, m);
		q += m;
		if(i < narg-1)
			*q++ = ' ';
	}
	return l;
fail:
	freelaunch(l);
	return 0;
}

static Ctl*
finishnewctl(Ctl *ctl, char **err)
{
	DEBUG_EVENT dbg;

	wait1(ctl, &dbg, Wblock);
	if(dbg.dwDebugEventCode != CREATE_PROCESS_DEBUG_EVENT)
		xfatalno("unexpected debug event on launch");
	ctl->x.process = dbg.u.CreateProcessInfo.hProcess;
	ctl->x.thread = dbg.u.CreateProcessInfo.hThread;
	if(!ctl->x.process || !ctl->x.thread){
		if(err)
			*err = "cannot debug target";
		goto fail;
	}
	while(1){
		ctlcont(ctl);
		wait1(ctl, &dbg, Wblock);
		if(dbg.dwDebugEventCode == EXCEPTION_DEBUG_EVENT)
			break;
	}
	ctl->as = mkas(ctl);
	xgetregs(ctl, &ctl->ctx.regs);
	ctl->state = Sstop;
	return ctl;
fail:
	if(err)
		*err = lasterror();
	return 0;
}

Ctl*
launch(Launch *l, char **err)
{
	u64 id;
	int i, k, ret;
	DEBUG_EVENT dbg;
	HANDLE hIN[2], hOUT[2], hERR[2];
  	SECURITY_ATTRIBUTES saAttr;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	Ctl *ctl;

	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.bInheritHandle = TRUE;    /* inherit pipe handles */
	saAttr.lpSecurityDescriptor = NULL;
   	memset(&pi, 0, sizeof(PROCESS_INFORMATION));
   	memset(&si, 0, sizeof(STARTUPINFO));
	si.hStdInput = INVALID_HANDLE_VALUE;
	si.hStdOutput = INVALID_HANDLE_VALUE;
	si.hStdError = INVALID_HANDLE_VALUE;

	if(l->flags&Fstdin){
		if(!CreatePipe(&hIN[0], &hIN[1], &saAttr, 0))
			goto fail;
		if(!SetHandleInformation(hIN[1], HANDLE_FLAG_INHERIT, 0))
			goto fail;
		si.hStdInput = hIN[0];
	}
	if(l->flags&Fstdout){
		if(!CreatePipe(&hOUT[0], &hOUT[1], &saAttr, 0))
			goto fail;
		if(!SetHandleInformation(hOUT[0], HANDLE_FLAG_INHERIT, 0))
			goto fail;
		si.hStdOutput = hOUT[1];
	}
	if(l->flags&Fstderr){
		if(!CreatePipe(&hERR[0], &hERR[1], &saAttr, 0))
			goto fail;
		if(!SetHandleInformation(hERR[0], HANDLE_FLAG_INHERIT, 0))
			goto fail;
		si.hStdError = hERR[1];
	}
	si.cb = sizeof(STARTUPINFO);
	si.dwFlags |= STARTF_USESTDHANDLES;

	ret = CreateProcess(NULL,		// target file name
			    l->cmd,		// command line
			    NULL,		// process attributes
			    NULL,		// thread attributes
			    TRUE,		// handles are not inherited
			    DEBUG_PROCESS,	// debug process and children
			    NULL,		// use current environment
			    NULL,		// use cwd
			    &si,
			    &pi);
	if(!ret)
		goto fail;

	ctl = mkctl(pi.dwProcessId, pi.dwThreadId);
	ctl->x.pid = pi.dwProcessId;
	ctl->x.tid = pi.dwThreadId;
	ctl->state = Sstop;
	ctl->whence = Wlaunched;
	hput(pid2ctl, ctl->x.pid, ctl);

	if(l->flags&Fstdin)
		ctl->fd[0] = hIN[1];
	if(l->flags&Fstdout)
		ctl->fd[1] = hOUT[0];
	if(l->flags&Fstderr)
		ctl->fd[2] = hERR[0];

	return finishnewctl(ctl, err);
fail:
	if(err)
		*err = lasterror();
	return 0;
}

Ctl*
attach(int apid, char **err)
{
	Ctl *ctl, *spwn;
	DEBUG_EVENT dbg;
	char *serr;
	Event *e;
	int c, pid, tid;

	if(!DebugActiveProcess(apid)){
		if(err){
			xlog("error 1 %d\n", apid);
			*err = lasterror();
		}
		return 0;
	}

	_wait1(apid, 0, &dbg, Wblock);
	if(dbg.dwDebugEventCode != CREATE_PROCESS_DEBUG_EVENT)
		xfatalno("attach: unexpected debug event %s",
			 dbg2str(dbg.dwDebugEventCode));
	ctl = mkctl(dbg.dwProcessId, dbg.dwThreadId);
	ctl->x.pid = dbg.dwProcessId;
	ctl->x.tid = dbg.dwThreadId;
	ctl->x.process = dbg.u.CreateProcessInfo.hProcess;
	ctl->x.thread = dbg.u.CreateProcessInfo.hThread;
	ctl->state = Sstop;
	ctl->whence = Wlaunched;
	ctl->as = mkas(ctl);
	hput(pid2ctl, ctl->x.pid, ctl);

	ctlcont(ctl);
	while(1){
		_wait1(apid, 0, &dbg, Wblock);
		c = dbg.dwDebugEventCode;
		pid = dbg.dwProcessId;
		tid = dbg.dwThreadId;
		switch(c){
		case CREATE_THREAD_DEBUG_EVENT:
			ctl = hget(pid2ctl, pid);
			if(ctl == 0)
				xfatal("no parent ctl for new thread %d:%d",
				       pid, tid);
			spwn = mkctl(pid, tid);
			spwn->x.process = ctl->x.process;
			spwn->x.thread = dbg.u.CreateThread.hThread;
			spwn->state = Sstop;
			spwn->whence = ctl->whence;
			linkspawn(ctl, spwn, 0);
			ctlcont(spwn);
			break;
		case LOAD_DLL_DEBUG_EVENT:
			ctl = lookctl(ctlid(pid, tid));
			if(ctl == 0)
				xfatal("no ctl for dll event on %d:%d",
				       pid, tid);
			if(dbg.u.LoadDll.hFile)
				CloseHandle(dbg.u.LoadDll.hFile);
			ctlcont(ctl);
			break;
		default:
			xlog("unexpected %s", dbg2str(c));
			break;
		case EXCEPTION_DEBUG_EVENT:
			ctl = lookctl(ctlid(pid, tid));
			if(ctl == 0)
				xfatal("no ctl for debug event on %d:%d",
				       pid, tid);
			xgetregs(ctl, &ctl->ctx.regs);
			ctl->state = Sstop;
			return ctl;
		}
	}
}

static int
editbrk(Ctl *ctl, Brk *bp, int set, char **err)
{
	u_char text;
	DWORD oldperm;

	if(!VirtualProtectEx(ctl->x.process,
			     (void*)bp->addr,
			     1,
			     PAGE_EXECUTE_READWRITE,
			     &oldperm))
		goto fail;
	if(set){
		/* Windows debug API allows one-byte
 		 * reads and writes */
		if(!ReadProcessMemory(ctl->x.process,
				      (LPCVOID)bp->addr,
				      &text,
				      1,
				      NULL))
			goto fail;
		bp->text = text;
		text = int3;
	} else
		text = bp->text;
	if(!WriteProcessMemory(ctl->x.process,
				(void*)bp->addr,
				&text,
				1,
				NULL))
		goto fail;
	if(!FlushInstructionCache(ctl->x.process,
				  (LPCVOID)bp->addr,
				  1))
		goto fail;
	if(!VirtualProtectEx(ctl->x.process,
				(void*)bp->addr,
				1,
				oldperm,
				&oldperm))
		goto fail;
	return 0;
fail:
	if(err)
		*err = lasterror();
	return -1;
}

int
clrbrk(Ctl *ctl, Brk *bp, char **err)
{
	/* FIXME: some thread may be stopped in this BP */
	return editbrk(ctl, bp, 0, err);
}

int
setbrk(Ctl *ctl, Brk *bp, char **err)
{
	return editbrk(ctl, bp, 1, err);
}

int
ctlread(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	SIZE_T m;
	if(!ReadProcessMemory(ctl->x.process, (void*)addr, buf, cnt, &m)){
		if(err)
			*err = lasterror();
		return -1;
	}
	return 0;
}

int
ctlwrite(Ctl *ctl, unsigned long addr, void *buf, unsigned long cnt, char **err)
{
	/* FIXME: Does it really return -1? */
	if(0 > WriteProcessMemory(ctl->x.process, (void*)addr, buf, cnt, NULL))
		goto fail;
	if(!FlushInstructionCache(ctl->x.process, (LPCVOID)addr, cnt))
		goto fail;
	return 0;
fail:
	if(err)
		*err = lasterror();
	return -1;
}

/* FIXME: this detaches all threads associated with the process */
int
ctldetach(Ctl *ctl, char **err)
{
	u32 pid;
	pid = ctl->x.pid;
	switch(ctl->state){
	case Srun:
	case Strap:
	case Sstop:
		dead(ctl);
		if(!DebugActiveProcessStop(pid))
			xfatalno("DebugActiveProcessStop");
		break;
	case Sdead:
		return -1;
	default:
		bug();
	}
	return 0;
}

/* FIXME: this kills all threads associated with the process */
int
ctlkill(Ctl *ctl, char **err)
{
	switch(ctl->state){
	case Strap:
		ctlcont(ctl);
		/* fall through */
	case Srun:
	case Sstop:
		if(!TerminateProcess(ctl->x.process, 0))
			xfatalno("TerminateProcess");
		break;
	case Sdead:
		return -1;
	default:
		bug();
	}
	dead(ctl);
	return 0;
}

Event*
ctlstop(Ctl *ctl, char **err)
{
	DEBUG_EVENT dbg;
	int c;

	xgetregs(ctl, &ctl->ctx.regs);
	ctl->ctx.regs.EFlags |= FL_TF;
	xsetregs(ctl, &ctl->ctx.regs);
	wait1(ctl, &dbg, Wblock);
	if(dbg.dwDebugEventCode != EXCEPTION_DEBUG_EVENT)
		xfatal("ctlstep: unexpected %s", dbg2str(dbg.dwDebugEventCode));
	c = dbg.u.Exception.ExceptionRecord.ExceptionCode;
	if(c != EXCEPTION_SINGLE_STEP)
		xfatal("ctlstep: EXCEPTION_DEBUG_EVENT is %d", c);
	ctl->state = Sstop;
	return 0;
}

int
ctlstep(Ctl *ctl, char **err)
{
	DEBUG_EVENT dbg;
	int c;

	// previous: standalone getregs
	// with ctl->ctx.regs.ContextFlags = CONTEXT_CONTROL;

	xgetregs(ctl, &ctl->ctx.regs);
	ctl->ctx.regs.EFlags |= FL_TF;
	xsetregs(ctl, &ctl->ctx.regs);
	ctlcont(ctl);
	wait1(ctl, &dbg, Wblock);
	if(dbg.dwDebugEventCode != EXCEPTION_DEBUG_EVENT)
		xfatal("ctlstep: unexpected %s", dbg2str(dbg.dwDebugEventCode));
	c = dbg.u.Exception.ExceptionRecord.ExceptionCode;
	if(c != EXCEPTION_SINGLE_STEP)
		xfatal("ctlstep: EXCEPTION_DEBUG_EVENT is %d", c);
	ctl->state = Sstop;
	return 0;
}

int
steptrap(Ctl *ctl, char **err)
{
	if(0 > clrbrk(ctl, &ctl->curbrk, err) ||
	   0 > ctlstep(ctl, err) ||
	   0 > setbrk(ctl, &ctl->curbrk, err))
		return -1;
	return 0;
}

Ctx*
ctlgetctx(Ctl *ctl, char **err)
{
	Ctx *c;
	c = xmalloc(sizeof(Ctx));
	*c = ctl->ctx;
	return c;
}

int
ctlsetctx(Ctl *ctl, Ctx *c, char **err)
{
	ctl->ctx = *c;
	xsetregs(ctl, &ctl->ctx.regs);
	return 0;
}

Event*
ctlwait(Ctl *maybectl)
{
	Event *e;
	char *err;
	DEBUG_EVENT dbg;

	e = 0;
	if(maybectl){
		e = evqmatch(maybectl);
		while(!e){
			wait1(maybectl, &dbg, Wblock);
			err = 0;
			e = doevent(&dbg, &err);
			if(!e && err)
				xfatal("doevent: %s", err);
		}
	}else{
		/* wait for any ctl */
		e = evqpop();
		while(!e){
			if(!WaitForDebugEvent(&dbg, INFINITE))
				xfatalno("WaitForDebugEvent");
			err = 0;
			e = doevent(&dbg, &err);
			if(!e && err)
				xfatal("doevent: %s", err);
		}
	}
	return e;
}

Event*
ctlpoll(Ctl *maybectl)
{
	Ctl *ctl;
	Event *e;
	char *err;
	DEBUG_EVENT dbg;

	e = 0;
	if(maybectl){
		e = evqmatch(maybectl);
		if(!e){
			wait1(maybectl, &dbg, Wblock);
			err = 0;
			e = doevent(&dbg, &err);
			if(!e && err)
				xfatal("doevent: %s", err);
			return e;
		}
	}else{
		/* poll for any ctl */
		e = evqpop();
		if(!e){
			if(!WaitForDebugEvent(&dbg, 0))
				/* FIXME: how does return value work? */
				return 0;
			err = 0;
			e = doevent(&dbg, &err);
			if(!e && err)
				xfatal("doevent: %s", err);
			return e;
		}
	}
	return 0;
}

Ctl*
ctlsnap(Ctl *ctl)
{
	xfatal("unimplemented");
	return 0;
}

int
ctlns(u8 op, void *buf, u64 len, void *xbuf, char **err)
{
	xfatal("unimplemented");
}

/* this would belong in win/sym.c */
void
symdupas(As *as, As *nas)
{
}
